Package: chipenrich
Type: Package
Title: Gene Set Enrichment For ChIP-seq Peak Data
Version: 2.35.0
Date: 2025-07-22
Authors@R: c(
	person(c("Ryan","P."),"Welch",role = c("aut","cph"),email = "welchr@umich.edu"),
	person("Chee","Lee",role = c("aut"),email = "cheelee@umich.edu"),
	person(c("Raymond","G."),"Cavalcante",role = c("aut"),email = "rcavalca@umich.edu"),
	person("Kai","Wang",role = c("cre"),email = "wangdaha@umich.edu"),
	person("Chris","Lee",role = c("aut"),email = "leetaiyi@umich.edu"),
	person(c("Laura","J."),"Scott",role = c("ths"), email = "ljst@umich.edu"),
	person(c("Maureen","A."),"Sartor",role = c("ths"),email = "sartorma@med.umich.edu"))
Description: ChIP-Enrich and Poly-Enrich perform gene set enrichment testing
	using peaks called from a ChIP-seq experiment. The method empirically
	corrects for confounding factors such as the length of genes,
	and the mappability of the sequence surrounding genes.
biocViews: ImmunoOncology, ChIPSeq, Epigenetics, FunctionalGenomics,
        GeneSetEnrichment, HistoneModification, Regression
License: GPL-3
Imports: AnnotationDbi, BiocGenerics, chipenrich.data, Seqinfo,
        GenomicRanges, grDevices, grid, IRanges, lattice, latticeExtra,
        MASS, methods, mgcv, org.Dm.eg.db, org.Dr.eg.db, org.Hs.eg.db,
        org.Mm.eg.db, org.Rn.eg.db, parallel, plyr, rms, rtracklayer,
        S4Vectors (>= 0.23.10), stats, stringr, utils
Suggests: BiocStyle, devtools, knitr, rmarkdown, roxygen2, testthat
Depends: R (>= 3.4.0)
LazyLoad: yes
Maintainer: Kai Wang <wangdaha@umich.edu>
RoxygenNote: 7.1.1
VignetteBuilder: knitr
Collate: 'assign_peaks.R' 'test_broadenrich.R' 'peaks_per_gene.R'
        'read.R' 'randomize.R' 'setup.R' 'supported.R' 'utils.R'
        'constants.R' 'plot_gene_coverage.R' 'broadenrich.R'
        'test_chipapprox.R' 'test_chipenrich.R'
        'test_chipenrich_slow.R' 'test_fisher.R' 'test_binomial.R'
        'test_approx.R' 'plot_chipenrich_spline.R' 'plot_dist_to_tss.R'
        'chipenrich.R' 'chipenrich_package_doc.R' 'test_polyapprox.R'
        'test_polyenrich_weighted.R' 'test_polyenrich_slow.R'
        'test_polyenrich.R' 'plot_polyenrich_spline.R' 'polyenrich.R'
        'hybrid.R' 'peak_weights.R' 'peaks2genes.R' 'test_proxReg.R'
        'proxReg.R'
git_url: https://git.bioconductor.org/packages/chipenrich
git_branch: devel
git_last_commit: 93dd65a
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-11-06
NeedsCompilation: no
Packaged: 2025-11-06 21:46:26 UTC; biocbuild
Author: Ryan P. Welch [aut, cph],
  Chee Lee [aut],
  Raymond G. Cavalcante [aut],
  Kai Wang [cre],
  Chris Lee [aut],
  Laura J. Scott [ths],
  Maureen A. Sartor [ths]

# Generated by roxygen2: do not edit by hand

export(assign_peak_segments)
export(assign_peaks)
export(broadenrich)
export(calc_peak_gene_overlap)
export(chipenrich)
export(hybridenrich)
export(load_peaks)
export(num_peaks_per_gene)
export(peaks2genes)
export(plot_chipenrich_spline)
export(plot_dist_to_tss)
export(plot_gene_coverage)
export(plot_polyenrich_spline)
export(polyenrich)
export(proxReg)
export(read_bed)
export(supported_genesets)
export(supported_genomes)
export(supported_locusdefs)
export(supported_methods)
export(supported_read_lengths)
import(GenomicRanges)
import(chipenrich.data)
import(grid)
import(lattice)
import(latticeExtra)
import(methods)
import(mgcv)
import(org.Dm.eg.db)
import(org.Dr.eg.db)
import(org.Hs.eg.db)
import(org.Mm.eg.db)
import(org.Rn.eg.db)
import(parallel)
import(rms)
import(rtracklayer)
importFrom(AnnotationDbi,mappedkeys)
importFrom(BiocGenerics,unlist)
importFrom(IRanges,IRanges)
importFrom(S4Vectors,queryHits)
importFrom(S4Vectors,splitAsList)
importFrom(S4Vectors,subjectHits)
importFrom(Seqinfo,seqinfo)
importFrom(grDevices,dev.off)
importFrom(grDevices,pdf)
importFrom(plyr,rbind.fill)
importFrom(stats,as.formula)
importFrom(stats,binom.test)
importFrom(stats,coef)
importFrom(stats,complete.cases)
importFrom(stats,fisher.test)
importFrom(stats,fitted)
importFrom(stats,na.omit)
importFrom(stats,p.adjust)
importFrom(stats,pchisq)
importFrom(stats,predict)
importFrom(stats,qpois)
importFrom(stats,quantile)
importFrom(stats,resid)
importFrom(stringr,str_match)
importFrom(stringr,str_sub)
importFrom(utils,data)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,write.table)

# ChIP-Enrich

R package for gene set enrichment testing using ChIP-seq data.

# Description

ChIP-Enrich performs gene set enrichment testing using peaks called from a ChIP-seq experiment. The method empirically corrects for confounding factors such as the length of genes, and the mappability of the sequence surrounding genes.

# Used in ..plot_polyenrich_spline(...)
avg_binned_numpeaks = function(gpw) {
	bygroup = stats::aggregate(cbind(num_peaks, length) ~ group, gpw, mean)
	bygroup$log_avg_length = log10(bygroup$length)
	names(bygroup) = c("group", "num_peaks", "avg_length", "log_avg_length")

	return(bygroup)
}

# Used in ..plot_polyenrich_spline(...)
calc_weights_polyenrich = function(gpw) {
	# Create model.
	model = "num_peaks ~ s(log10_length, bs = 'cr')"

	# Compute binomial spline fit.
	fit = gam(as.formula(model), data = gpw, family = "nb")

	fitted_numpeaks = fitted(fit)

	# These are not used, likely part of development
	# Perhaps of use for Chris in weighted polyenrich
	w0 = 1 / (fitted_numpeaks / mean(gpw$num_peaks, na.rm = TRUE))
	w0 = w0 / mean(w0, na.rm = TRUE)
	gpw$weight = w0
	gpw$resid.dev = resid(fit, type="deviance")

	# This is the only part that gets used
	gpw$fitted_numpeaks = fitted_numpeaks

	return(gpw)
}

#' QC plot for Poly-Enrich
#'
#' Create a plot the relationship between number of peaks assigned to a gene and
#' locus length. The plot shows an empirical fit to the data using a binomial
#' smoothing spline.
#'
#' @param peaks Either a file path or a \code{data.frame} of peaks in BED-like
#' format. If a file path, the following formats are fully supported via their
#' file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
#' or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
#' without these extensions are supported under the conditions that the first 3
#' columns correspond to 'chr', 'start', and 'end' and that there is either no
#' header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
#' \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
#' acceptable column names.
#' @param locusdef One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
#' '1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
#' '10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
#' see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
#' a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
#' must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
#' example custom locus definition file, see the vignette.
#' @param genome One of the \code{supported_genomes()}.
#' @param mappability One of \code{NULL}, a file path to a custom mappability file,
#' or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
#' If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
#' Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
#' For an example custom mappability file, see the vignette. Default value is NULL.
#' @param legend If true, a legend will be drawn on the plot.
#' @param xlim Set the x-axis limit. NULL means select x-lim automatically.
#' @param ylim Set the y-axis limit. NULL means select y-lim automatically.
#'
#' @return A trellis plot object.
#'
#' @examples
#'
#' # Spline plot for E2F4 example peak dataset.
#' data(peaks_E2F4, package = 'chipenrich.data')
#'
#' # Create the plot for a different locus definition
#' # to compare the effect.
#' plot_polyenrich_spline(peaks_E2F4, locusdef = 'nearest_gene', genome = 'hg19')
#'
#' @export
#' @include constants.R utils.R supported.R setup.R randomize.R
#' @include read.R assign_peaks.R peaks_per_gene.R
plot_polyenrich_spline = function(peaks, locusdef = "nearest_tss", genome = supported_genomes(), mappability = NULL, legend = TRUE, xlim = NULL, ylim = NULL) {
	genome = match.arg(genome)

	ldef_list = setup_locusdef(locusdef, genome)
	ldef = ldef_list[['ldef']]
	tss = ldef_list[['tss']]

	mappa = setup_mappa(mappa_code = mappability, genome = genome, ldef_code = locusdef, ldef_obj = ldef)

	if (class(peaks) == "data.frame") {
		peakobj = load_peaks(peaks)
	} else if (class(peaks) == "character") {
		peakobj = read_bed(peaks)
	}
	num_peaks = length(peakobj)

	assigned_peaks = assign_peaks(peakobj, ldef, tss)
	ppg = num_peaks_per_gene(assigned_peaks, ldef, mappa)

	# Make plot.
	plotobj = ..plot_polyenrich_spline(gpw = ppg, mappability = mappability, num_peaks = num_peaks, legend = legend, xlim = xlim, ylim = ylim)
	return(plotobj)
}

..plot_polyenrich_spline = function(gpw, mappability, num_peaks, legend = TRUE, xlim = NULL, ylim = NULL) {
	############################################################################
	# Prepare the gpw table
	gpw = gpw[order(gpw$log10_length), ]
	gpw$group = ceiling((1:nrow(gpw)) / 25)

	############################################################################
	# Quantities for the plot

	# Scatterplot stuff
	avg_bins = avg_binned_numpeaks(gpw)

	# Spline stuff
	gpw = calc_weights_polyenrich(gpw) # gpw = genes, peaks, weights
	# genome_length = sum(as.numeric(gpw$length))
	# gpw$false_prob = 1 - (1 - (gpw$length / genome_length))^(num_peaks)

	############################################################################
	# Plotting parameters
	col_spline = "darkorange"

	panel_func = function(x,y,...) {
		lattice::panel.xyplot(x, y, ...)
	}

	custom_key = list(
		text = list(
		c(
			"Binomial Smoothing Spline Fit",
			"Avg Number of Peaks per Gene in Bins"
			)
		),
		lines = list(
			pch = c(17, 20),
			type = c("l", "p"),
			col = c(col_spline, "black"),
			lwd = 3
		),
		cex = 1.2
	)

	if (!legend) {
		custom_key = NULL
	}

	if (is.null(mappability)) {
		xlab = expression(paste(Log[10], " Locus Length"))
	} else {
		xlab = expression(paste(Log[10], " Mappable Locus Length"))
	}

	xmin_nopad = base::ifelse(is.null(xlim[1]), floor(min(gpw$log10_length)), floor(xlim[1]))
	xmax_nopad = base::ifelse(is.null(xlim[2]), ceiling(max(gpw$log10_length)), ceiling(xlim[2]))

	ymin_nopad = base::ifelse(is.null(ylim[1]), floor(min(avg_bins$num_peaks)), floor(ylim[1]))
	ymax_nopad = base::ifelse(is.null(ylim[2]), ceiling(max(avg_bins$num_peaks)), ceiling(ylim[2]))

	scales = list(
		x = list(
			axs = 'i',
			at = seq(xmin_nopad, xmax_nopad, 1)
		),
		y = list(
			axs = 'i',
			at = seq(ymin_nopad, ymax_nopad, 0.2)
		)
	)

	############################################################################
	# Plot
	plotobj = lattice::xyplot(
		fitted_numpeaks ~ log10_length,
		gpw,
		xlab = list(label = xlab, cex = 1.4),
		ylab = list(label = "Avg Number of Peaks per Gene in Bins", cex = 1.4),
		ylim = c(ymin_nopad - 0.5, ymax_nopad + 0.5),
		xlim = c(xmin_nopad - 0.5, xmax_nopad + 0.5),
		panel = panel_func,
		type = "l",
		lwd = 3,
		key = custom_key,
		scales = list(cex = 1.4),
		par.settings = lattice::simpleTheme(pch = c(17), col = c(col_spline))
	) + latticeExtra::as.layer(lattice::xyplot(num_peaks ~ log_avg_length, avg_bins, pch = 20, cex = 0.4, col = "black"))

	return(plotobj)
}

context('Test QC plot functions')

data(peaks_E2F4, package = 'chipenrich.data')
data(peaks_H3K4me3_GM12878, package = 'chipenrich.data')

test_that('Distance to TSS plot works', {
	plot = suppressWarnings(plot_dist_to_tss(peaks = peaks_E2F4, genome = 'hg19'))

	expect_equal(class(plot), 'trellis')
})

test_that('ChIP-Enrich spline plot works', {
	plot = suppressWarnings(plot_chipenrich_spline(peaks = peaks_E2F4, locusdef = 'nearest_tss', genome = 'hg19'))

	expect_equal(class(plot), 'trellis')
})

test_that('ChIP-Enrich spline plot with mappability works', {
	plot = suppressWarnings(plot_chipenrich_spline(peaks = peaks_E2F4, locusdef = 'nearest_tss', genome = 'hg19', mappability = '50'))

	expect_equal(class(plot), 'trellis')
})

test_that('Poly-Enrich spline plot works', {
	plot = suppressWarnings(plot_polyenrich_spline(peaks = peaks_E2F4, locusdef = 'nearest_tss', genome = 'hg19'))

	expect_equal(class(plot), 'trellis')
})

test_that('Broad-Enrich coverage plot works', {
	plot = suppressWarnings(plot_gene_coverage(peaks = peaks_H3K4me3_GM12878, locusdef = 'nearest_tss',  genome = 'hg19'))

	expect_equal(class(plot), 'trellis')
})

#' Assign peak midpoints to defined gene loci.
#'
#' Determine the midpoints of a set of input regions \code{peaks} and the overlap of the midpoints with a given locus definition \code{locusdef}. Also report the TSS that is nearest each region (peak) overlapping a defined locus and its distance.
#'
#' Typically, this function will not be used alone, but inside \code{chipenrich()}.
#'
#' @param peaks A \code{GRanges} object representing regions to be used for enrichment.
#' @param locusdef A locus definition object from \code{chipenrich.data}.
#' @param tss A \code{GRanges} object representing the TSSs for the genome build. Includes \code{mcols} for Entrez Gene ID \code{gene_id} and gene symbol \code{symbol}.
#' @param weighting A string defining what weighting option they want. Current options are 'multiAssign', 'signalValue', and 'logSignal Value'. Default is NULL.
#'
#' @return A \code{data.frame} with columns for \code{peak_id, chr, peak_start, peak_end, gene_locus_start, gene_locus_end, gene_id, nearest_tss, nearest_tss_gene, dist_to_tss, nearest_tss_gene_strand}. The result is used in \code{num_peaks_per_gene()}.
#'
#' @examples
#'
#' data('locusdef.hg19.nearest_tss', package = 'chipenrich.data')
#' data('tss.hg19', package = 'chipenrich.data')
#'
#' file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' assigned_peaks = assign_peaks(
#' 	peaks = peaks,
#' 	locusdef = locusdef.hg19.nearest_tss,
#' 	tss = tss.hg19)
#'
#' @export
assign_peaks = function(peaks, locusdef, tss, weighting=NULL) {
	# Extract GRanges of locusdef
	# NOTE: locusdef is an environment, so uses @
	ldef_gr = locusdef@granges

	# Determine midpoints of peaks and construct a GRanges object
	# on that basis. Include the peak name for later merging
	peak_mids = IRanges::mid(GenomicRanges::ranges(peaks))
	mids_gr = GenomicRanges::GRanges(
		seqnames = Seqinfo::seqnames(peaks),
		ranges = IRanges::IRanges(start = peak_mids, end = peak_mids),
		name = GenomicRanges::mcols(peaks)$name
	)
    if (all(c("signalValue", "logsignalValue") %in% weighting)) {
        stop("You can only choose one of signalValue and logsignalValue!")
    }

    if (any(c("signalValue", "logsignalValue") %in% weighting)) {
        if (!("signalValue" %in% colnames(GenomicRanges::mcols(peaks)))) {
            stop("No signalValue column!")
        }
        mids_gr$signalValue = GenomicRanges::mcols(peaks)$signalValue
    }


	# Determine overlaps of peak midpoints with locus definition
	mid_ldef_overlaps = GenomicRanges::findOverlaps(mids_gr, ldef_gr)

	if(length(mid_ldef_overlaps) == 0) {
		stop('Intersection between peak midpoints and locus definition is empty. Try selecting a different locus definition.')
	}

	# Determine distance of peak midpoints to nearest TSSs
	mid_dist_to_ntss = GenomicRanges::distanceToNearest(mids_gr, tss)

	# Sign the distances by up (-) or down (+) stream
	mid_starts = GenomicRanges::start(mids_gr)[S4Vectors::queryHits(mid_dist_to_ntss)]
	tss_starts = GenomicRanges::start(tss)[S4Vectors::subjectHits(mid_dist_to_ntss)]
	tss_pos_strand = GenomicRanges::strand(tss)[S4Vectors::subjectHits(mid_dist_to_ntss)] == "+"
	tss_neg_strand = GenomicRanges::strand(tss)[S4Vectors::subjectHits(mid_dist_to_ntss)] == "-"

	neg1 = (mid_starts < tss_starts) & (tss_pos_strand)
	neg2 = (mid_starts > tss_starts) & (tss_neg_strand)
	neg_dists = as.logical(neg1 | neg2)

	# Reassign the signed distances
	dists = GenomicRanges::mcols(mid_dist_to_ntss)$distance
	dists[neg_dists] = dists[neg_dists] * -1
	GenomicRanges::mcols(mid_dist_to_ntss)$distance = dists

	# Construct data.frames for overlaps and distances
	mid_indices = S4Vectors::queryHits(mid_ldef_overlaps)
	ldef_indices = S4Vectors::subjectHits(mid_ldef_overlaps)
	mid_ldef_df = data.frame(
		peak_id = GenomicRanges::mcols(mids_gr)$name[mid_indices],
		chr = Seqinfo::seqnames(mids_gr)[mid_indices],
		peak_start = GenomicRanges::start(peaks)[mid_indices],
		peak_end = GenomicRanges::end(peaks)[mid_indices],
		gene_locus_start = GenomicRanges::start(ldef_gr)[ldef_indices],
		gene_locus_end = GenomicRanges::end(ldef_gr)[ldef_indices],
		gene_id = GenomicRanges::mcols(ldef_gr)$gene_id[ldef_indices],
		gene_symbol = GenomicRanges::mcols(ldef_gr)$symbol[ldef_indices],
		stringsAsFactors = FALSE
	)
    
    # Add the signalValue column if method == "polyenrich_weighted"
    # Stops if there isn't one
    if (any(c("signalValue","logsignalValue") %in% weighting)) {
        mid_ldef_df$signalValue = GenomicRanges::mcols(peaks)$signalValue[mid_indices]
    }

	mid_indices = S4Vectors::queryHits(mid_dist_to_ntss)
	tss_indices = S4Vectors::subjectHits(mid_dist_to_ntss)
	mid_dist_df = data.frame(
		peak_id = GenomicRanges::mcols(mids_gr)$name[mid_indices],
		nearest_tss = GenomicRanges::start(tss)[tss_indices],
		nearest_tss_gene_id = GenomicRanges::mcols(tss)$gene_id[tss_indices],
		nearest_tss_symbol = GenomicRanges::mcols(tss)$symbol[tss_indices],
		dist_to_tss = GenomicRanges::mcols(mid_dist_to_ntss)$distance,
		nearest_tss_gene_strand = GenomicRanges::strand(tss)[tss_indices],
		stringsAsFactors = FALSE
	)

	# Merge the overlap and distance data.frames on the peak names
	d = base::merge(
		x = mid_ldef_df,
		y = mid_dist_df,
		by = 'peak_id',
		all.x = TRUE,
		all.y = FALSE,
		sort = FALSE
	)

	# Reorder the columns
	column_order = c(
		"peak_id",
		"chr",
		"peak_start",
		"peak_end",
		"gene_id",
		"gene_symbol",
		"gene_locus_start",
		"gene_locus_end",
		"nearest_tss",
		"dist_to_tss",
		"nearest_tss_gene_id",
		"nearest_tss_symbol",
		"nearest_tss_gene_strand"
	)
    if (any(c("signalValue","logsignalValue") %in% weighting)) {
        column_order = c(column_order, "signalValue")
    }
	d = d[, column_order]

	return(d)
}

#' Assign whole peaks to all overlapping defined gene loci.
#'
#' Determine all overlaps between the set of input regions \code{peaks} and the given locus definition \code{locusdef}. In addition, report where each overlap begins and ends, as well as the length of the overlap.
#'
#' Typically, this function will not be used alone, but inside \code{chipenrich()} with \code{method = 'broadenrich'}.
#'
#' @param peaks A \code{GRanges} object representing regions to be used for enrichment.
#' @param locusdef A locus definition object from \code{chipenrich.data}.
#'
#' @return A \code{data.frame} with columns for \code{peak_id, chr, peak_start, peak_end, gene_locus_start, gene_locus_end, gene_id, overlap_start, overlap_end, peak_overlap}. The result is used in \code{num_peaks_per_gene()}.
#'
#' @examples
#'
#' data('locusdef.hg19.nearest_tss', package = 'chipenrich.data')
#' data('tss.hg19', package = 'chipenrich.data')
#'
#' file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' assigned_peaks = assign_peak_segments(
#' 	peaks = peaks,
#' 	locusdef = locusdef.hg19.nearest_tss)
#'
#' @export
assign_peak_segments = function(peaks, locusdef) {
	# Extract GRanges of locusdef
	# NOTE: locusdef is a class, so uses @
	ldef_gr = locusdef@granges

	# Determine overlaps of peaks with locus definition
	peak_ldef_overlaps = GenomicRanges::findOverlaps(peaks, ldef_gr)

	if(length(peak_ldef_overlaps) == 0) {
		stop('Intersection between peaks and locus definition is empty. Try selecting a different locus definition.')
	}

	# Determine start and ends for peaks and genes that overlap
	peak_indices = S4Vectors::queryHits(peak_ldef_overlaps)
	ldef_indices = S4Vectors::subjectHits(peak_ldef_overlaps)

	peak_start = GenomicRanges::start(peaks)[peak_indices]
	peak_end = GenomicRanges::end(peaks)[peak_indices]
	gene_start = GenomicRanges::start(ldef_gr)[ldef_indices]
	gene_end = GenomicRanges::end(ldef_gr)[ldef_indices]

	# Use pmin() and pmax() to determine componentwise mins and maxes
	# which are also the starts and ends of the overlaps
	overlap_start = pmax(peak_start, gene_start)
	overlap_end = pmin(peak_end, gene_end)
	overlap_length = overlap_end - overlap_start + 1

	d = data.frame(
		peak_id = GenomicRanges::mcols(peaks)$name[peak_indices],
		chr = Seqinfo::seqnames(peaks)[peak_indices],
		peak_start = peak_start,
		peak_end = peak_end,
		gene_locus_start = gene_start,
		gene_locus_end = gene_end,
		gene_id = GenomicRanges::mcols(ldef_gr)$gene_id[ldef_indices],
		gene_symbol = GenomicRanges::mcols(ldef_gr)$symbol[ldef_indices],
		overlap_start = overlap_start,
		overlap_end = overlap_end,
		peak_overlap = overlap_length,
		stringsAsFactors = FALSE
	)

	return(d)
}

context('Test peaks per gene functions')

################################################################################
# Test human locus definitions

# Load the required data for the test
hg19_ldefs = grep('locusdef.hg19', data(package = 'chipenrich.data')$results[,3], value=T)
data(list = hg19_ldefs, package = 'chipenrich.data')
tss = get(data('tss.hg19', package = 'chipenrich.data'))

file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
peaks = suppressMessages(read_bed(file))

# Expected values for peak assignments
expected_numpeaks = list(
	'locusdef.hg19.10kb' = c(3),
	'locusdef.hg19.10kb_outside' = c(1),
	'locusdef.hg19.10kb_outside_upstream' = c(1),
	'locusdef.hg19.1kb' = c(2),
	'locusdef.hg19.1kb_outside' = c(1,1),
	'locusdef.hg19.1kb_outside_upstream' = c(1,1),
	'locusdef.hg19.5kb' = c(3),
	'locusdef.hg19.5kb_outside' = c(1),
	'locusdef.hg19.5kb_outside_upstream' = c(1),
	'locusdef.hg19.exon' = c(1),
	'locusdef.hg19.intron' = c(1),
	'locusdef.hg19.nearest_gene' = c(3, 1),
	'locusdef.hg19.nearest_tss' = c(3, 1)
)

# Do the tests in a loop for each locus definition of hg19
for(ldef in hg19_ldefs) {
	test_assign = assign_peaks(peaks = peaks, locusdef = get(ldef), tss = tss)

	test_ppg = num_peaks_per_gene(assigned_peaks = test_assign, locusdef = get(ldef), mappa = NULL)

	expect_true(all(subset(test_ppg, num_peaks != 0)$num_peaks == expected_numpeaks[[ldef]]),
		info = sprintf('Test num_peaks_per_gene(): %s', ldef))
}

expected_overlaps = list(
	'locusdef.hg19.10kb' = c(700,80),
	'locusdef.hg19.10kb_outside' = c(921),
	'locusdef.hg19.10kb_outside_upstream' = c(921),
	'locusdef.hg19.1kb' = c(327),
	'locusdef.hg19.1kb_outside' = c(1000,374),
	'locusdef.hg19.1kb_outside_upstream' = c(1000,374),
	'locusdef.hg19.5kb' = c(700),
	'locusdef.hg19.5kb_outside' = c(1000),
	'locusdef.hg19.5kb_outside_upstream' = c(1000),
	'locusdef.hg19.exon' = c(100),
	'locusdef.hg19.intron' = c(100),
	'locusdef.hg19.nearest_gene' = c(700,1000),
	'locusdef.hg19.nearest_tss' = c(700,1000)
)

# Do the tests in a loop for each locus definition of hg19
for(ldef in hg19_ldefs) {
	test_assign = assign_peak_segments(peaks = peaks, locusdef = get(ldef))

	test_ppg = num_peaks_per_gene(assigned_peaks = test_assign, locusdef = get(ldef), mappa = NULL)
	test_ppg = calc_peak_gene_overlap(assigned_peaks = test_assign, ppg = test_ppg)

	expect_true(all(subset(test_ppg, peak_overlap != 0)$peak_overlap == expected_overlaps[[ldef]]),
		info = sprintf('Test num_peaks_per_gene(): %s', ldef))
}

context('Test peak assignment functions')

################################################################################
# Test no assigned peaks

test_that('No peaks assigned', {

	ldef_file = system.file('extdata', 'test_ldef_symbol.txt', package = 'chipenrich')
	peak_file = system.file('extdata', 'test.broadPeak', package = 'chipenrich')

	peaks = suppressWarnings(read_bed(peak_file))
	ldef_list = setup_locusdef(ldef_file, genome = 'hg19')

	expect_error(assign_peaks(peaks, ldef_list$ldef, ldef_list$tss), 'Intersection between peak midpoints and locus definition is empty')
	expect_error(assign_peak_segments(peaks, ldef_list$ldef), 'Intersection between peaks and locus definition is empty')
})

################################################################################
# Test human locus definitions

test_chipenrich_slow = function(geneset, gpw, n_cores) {
	# Restrict our genes/weights/peaks to only those genes in the genesets.
	# Here, geneset is not all combined, but GOBP, GOCC, etc.
	gpw = subset(gpw, gpw$gene_id %in% geneset@all.genes)

	# Construct model formula.
	model = "peak ~ goterm + s(log10_length,bs='cr')"

	# Run tests. NOTE: If os == 'Windows', n_cores is reset to 1 for this to work
	results_list = parallel::mclapply(as.list(ls(geneset@set.gene)), function(go_id) {
		single_chipenrich_slow(go_id, geneset, gpw, 'chipenrich_slow', model)
	}, mc.cores = n_cores)

	# Collapse results into one table
	results = Reduce(rbind,results_list)

	# Correct for multiple testing
	results$FDR = stats::p.adjust(results$P.value, method = "BH")

	# Create enriched/depleted status column
	results$Status = ifelse(results$Effect > 0, 'enriched', 'depleted')

	results = results[order(results$P.value), ]

	return(results)
}

single_chipenrich_slow = function(go_id, geneset, gpw, method, model) {
	final_model = as.formula(model)

	# Genes in the geneset
	go_genes = geneset@set.gene[[go_id]]

	# Background genes and the background presence of a peak
	b_genes = gpw$gene_id %in% go_genes
	sg_go = gpw$peak[b_genes]

 	# Information about the geneset
	r_go_id = go_id
	r_go_genes_num = length(go_genes)
	r_go_genes_avg_length = mean(gpw$length[b_genes])

    # Information about peak genes
	go_genes_peak = gpw$gene_id[b_genes][sg_go == 1]
	r_go_genes_peak = paste(go_genes_peak, collapse = ", ")
	r_go_genes_peak_num = length(go_genes_peak)

	# Small correction for case where every gene in this geneset has a peak.
	if (all(as.logical(sg_go))) {
		cont_length = quantile(gpw$length, 0.0025)

		cont_gene = data.frame(
			gene_id = "continuity_correction",
			length = cont_length,
			log10_length = log10(cont_length),
			num_peaks = 0,
			peak = 0,
			stringsAsFactors = FALSE)

		if ("mappa" %in% names(gpw)) {
			cont_gene$mappa = 1
		}
		gpw = rbind(gpw,cont_gene)
		b_genes = c(b_genes,1)

		message(sprintf("Applying correction for geneset %s with %i genes...", go_id, length(go_genes)))
	}

    fit = mgcv::gam(final_model, data = cbind(gpw, goterm = as.numeric(b_genes)), family = "binomial")

	# Results from the logistic regression
    r_effect = coef(fit)[2]
    r_pval = summary(fit)$p.table[2, 4]

	out = data.frame(
		"P.value" = r_pval,
		"Geneset ID" = r_go_id,
		"N Geneset Genes" = r_go_genes_num,
		"Geneset Peak Genes" = r_go_genes_peak,
		"N Geneset Peak Genes" = r_go_genes_peak_num,
		"Effect" = r_effect,
		"Odds.Ratio" = exp(r_effect),
		"Geneset Avg Gene Length" = r_go_genes_avg_length,
		stringsAsFactors = FALSE)

	return(out)
}

#' A helper function to post-process peak GRanges
#'
#' Check for overlapping input regions, sort peaks, and force peak names
#'
#' @param gr A \code{GRanges} of input peaks.
#'
#' @return A \code{GRanges} that is sorted if the \code{seqinfo} is set, and has named peaks.
postprocess_peak_grs = function(gr) {
	# Check for overlapping peaks
	reduced_gr = GenomicRanges::reduce(gr)
	if(length(gr) != length(reduced_gr)) {
		warning('Some input regions overlap. It is recommended that input regions be disjoint.')
	}

	# Sort
	gr = GenomicRanges::sort(gr)

	# Create name column, or replace what's there
	GenomicRanges::mcols(gr)$name = paste0('peak:', seq_along(gr))

	return(gr)
}

#' Read files containing peaks or genomic regions
#'
#' The following formats are fully supported via their file extensions: .bed,
#' .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph or .bdg. BED3 through
#' BED6 files are supported under the .bed extension. Files without these extensions
#' are supported under the conditions that the first 3 columns correspond to
#' chr, start, and end and that there is either no header column, or it is
#' commented out. Files may be compressed with gzip, and so might end in .narrowPeak.gz,
#' for example. For files with extension support, the \code{rtracklayer::import()}
#' function is used to read peaks, so adherence to the mentioned file formats is
#' necessary.
#'
#' NOTE: Header rows must be commented with \code{#} to be ignored. Otherwise,
#' an error may result.
#'
#' NOTE: A warning is given if any input regions overlap. In the case of enrichment
#' testing with \code{method = 'broadenrich'}, regions should be disjoint.
#'
#' Typically, this function will not be used alone, but inside \code{chipenrich()}.
#'
#' @param file_path A path to a file with input peaks/regions. See extended
#' description above for details about file support.
#'
#' @return A \code{GRanges} with \code{mcols} matching any extra columns.
#'
#' @examples
#'
#' # Example of generic .txt file with peaks
#' file = system.file('extdata', 'test_header.txt', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' # Example of BED3
#' file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' # Example of narrowPeak
#' file = system.file('extdata', 'test.narrowPeak', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' # Example of gzipped broadPeak
#' file = system.file('extdata', 'test.broadPeak.gz', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' # Example of gzipped gff3 Fly peaks
#' file = system.file('extdata', 'test.gff3.gz', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' @export
read_bed = function(file_path) {
	message(sprintf('Reading peaks from %s', file_path))

	# Establish format and assign extraCols and format as needed
	# See https://genome.ucsc.edu/FAQ/FAQformat for format details
	if(grepl('.narrowPeak', file_path)) {
		extraCols = c('signalValue' = 'numeric', 'pValue' = 'numeric', 'qValue' = 'numeric', 'peak' = 'integer')
		format = 'bed'
	} else if (grepl('.broadPeak', file_path)) {
		extraCols = c('signalValue' = 'numeric', 'pValue' = 'numeric', 'qValue' = 'numeric')
		format = 'bed'
	} else if (grepl('.gff3', file_path) || grepl('.gff2', file_path) || grepl('.gff', file_path)) {
		extraCols = character()
		format = 'gff'
	} else if (grepl('.bedGraph', file_path) || grepl('.bdg', file_path)) {
		extraCols = character()
		format = 'bedGraph'
	} else if (grepl('.bed', file_path)) {
		# This works for BED3 through BED6
		extraCols = character()
		format = 'bed'
	} else {
		# Catch other formats
		format = 'txt'
	}

	# Read in the file using rtracklayer::import() for anything but format == 'txt'
	# Otherwise, use read.table + load_peaks() to get the GRanges we need
	if(format %in% c('gff3','gff2','gff','bedGraph')) {
		# Extension detection takes care of the above formats
		gr = rtracklayer::import(con = file_path)
		# Check for overlapping peaks and warn, sort, and name the peaks
		gr = postprocess_peak_grs(gr)
	} else if (format == 'bed') {
		# narrowPeak and broadPeak aren't supported with extension detection
		# so give the fixed extraCols and use the bed format.
		gr = rtracklayer::import(con = file_path, format = format, extraCols = extraCols)
		# Check for overlapping peaks and warn, sort, and name the peaks
		gr = postprocess_peak_grs(gr)
	} else {
		# This catches format == 'txt' (i.e. all other extensions)
		df = read.table(file_path, header = FALSE, sep = '\t', stringsAsFactors = FALSE)
		# Rename the first three columns
		colnames(df)[1:3] = c('chr','start','end')
		# Don't keep extra columns because we don't know what they are
		# NOTE: This includes postprocess_peak_grs()
		gr = load_peaks(dframe = df, keep.extra.columns = FALSE)
	}

	return(gr)
}

#' Convert a BEDX+Y data.frame and into GRanges
#'
#' Given a \code{data.frame} in BEDX+Y format, use the built-in function
#' \code{GenomicRanges::makeGRangesFromDataFrame()} to convert to \code{GRanges}.
#'
#' Typically, this function will not be used alone, but inside \code{chipenrich()}.
#'
#' @param dframe A BEDX+Y style \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame}
#' for acceptable column names for appropriate conversion to \code{GRanges}.
#' @param keep.extra.columns Keep extra columns parameter from \code{GenomicRanges::makeGRangesFromDataFrame()}.
#'
#' @return A \code{GRanges} that may or may not \code{keep.extra.columns}, and
#' that may or may not be stranded, depending on whether there is strand column
#' in the \code{dframe}.
#'
#' @examples
#'
#' # Example with just chr, start, end
#' peaks_df = data.frame(
#'	chr = c('chr1','chr2','chr3'),
#'	start = c(35,74,235),
#'	end = c(46,83,421),
#'	stringsAsFactors = FALSE)
#' peaks = load_peaks(peaks_df)
#'
#' # Example with extra columns
#' peaks_df = data.frame(
#'	chr = c('chr1','chr2','chr3'),
#'	start = c(35,74,235),
#'	end = c(46,83,421),
#'	strand = c('+','-','+'),
#'	score = c(36, 747, 13),
#'	stringsAsFactors = FALSE)
#' peaks = load_peaks(peaks_df, keep.extra.columns = TRUE)
#'
#' @export
load_peaks = function(dframe, keep.extra.columns = TRUE) {
	# Use built-in function
	gr = GenomicRanges::makeGRangesFromDataFrame(
		df = dframe,
		keep.extra.columns = keep.extra.columns)

	# Check for overlapping peaks and warn, sort, and name the peaks
	gr = postprocess_peak_grs(gr)

	return(gr)
}

#' Function to read custom locus definition from file
#'
#' This function reads a tab-delimited text (with a header) file that should have
#' columns 'chr', 'start', 'end', and a column named 'gene_id' (or 'geneid') with the
#' Entrez Gene ID. If a \code{supported_genomes()} is given, then a column
#' of gene symbols named 'symbol', will be added. If an unsupported genome is
#' used there are two options: 1) Have a column named 'symbols' with the gene
#' symbols in the custom locus definition, and leave \code{genome = NA}, or 2)
#' leave \code{genome = NA}, do not provide gene symbols, and NAs will be used.
#'
#' @param file_path A file path for the custom locus definition.
#' @param genome A genome from \code{supported_genomes()}, default \code{NA}.
#'
#' @return A \code{LocusDefinition} class object with slots \code{dframe},
#' \code{granges}, \code{genome.build}, and \code{organism}.
read_ldef = function(file_path, genome = NA) {

	message("Reading user-specified gene locus definitions: ", file_path)
	df = read.table(file_path, sep = "\t", header = TRUE, stringsAsFactors = FALSE)

	# Remove duplicated rows, not duplicated genes.
	# Some genes will exist on multiple rows because they are split by
	# other transcripts, or small nuclear RNAs, etc.
	df = unique(df)

	# Rename 'geneid' columns 'gene_id' for backwards compatibility
	if('geneid' %in% colnames(df)) {
		colnames(df)[which(colnames(df) == 'geneid')] = 'gene_id'
	}

	if(!('gene_id' %in% colnames(df))) {
		stop("Error: Custom locus definition must have column named 'gene_id'.")
	}

	if('symbol' %in% colnames(df)) {
		# Do nothing
		message('Using given symbol column...')
	} else if (genome %in% supported_genomes()) {
		message('Using orgDb package to get gene symbols...')
		eg2symbol = genome_to_orgdb(genome)
		df$symbol = eg2symbol[match(df$gene_id, eg2symbol$gene_id), 'symbol']
	} else {
		message('Setting gene symbols to NA...')
		df$symbol = NA
	}

	gr = GenomicRanges::makeGRangesFromDataFrame(
		df = df,
		keep.extra.columns = TRUE)

	# Create new locus definition object.
	object = new("LocusDefinition")

	object@dframe = df
	object@granges = gr

	return(object)
}

# Creates an object that mimics the GeneSet class
# from the chipenrich.data package, as in:
#    setClass("GeneSet",representation(
#      set.gene = "environment",
#      type = "character",
#      set.name = "environment",
#      all.genes = "character",
#      organism = "character",
#      dburl = "character"
#    ),prototype(
#      set.gene = new.env(parent=emptyenv()),
#      type = "",
#      set.name = new.env(parent=emptyenv()),
#      all.genes = "",
#      organism = "",
#      dburl = ""
#    ),
#      package = "chipenrich.data"
#    )

#' Function to read custom gene sets from file
#'
#' This function reads a two-columned tab-delimited text file (with header). Column
#' names are ignored, but the first column should be geneset names or IDs and the
#' second column should be Entrez Gene IDs.
#'
#' @param file_path A file path for the custom gene set.
#'
#' @return A \code{GeneSet} class object.
read_geneset = function(file_path) {

	# Read in flat file.
	message("Reading user-specified gene set definitions: ", file_path)
	d = read.table(file_path, sep = "\t", header = TRUE, stringsAsFactors = FALSE)

	# Split the gene_id column (second) by the geneset_id column (first)
	gs = split(d[,2], d[,1])

	# Create the names list which is identity
	gs.names = as.list(names(gs))
	names(gs.names) = names(gs)

	# Create new GeneSet object
	object = new('GeneSet')

	# Populate the GeneSet object
	object@type = 'user-supplied'
	object@organism = 'user-supplied'
	object@dburl = 'user-supplied'

	object@set.gene = as.environment(gs)
	object@all.genes = as.character(unlist(gs, use.names = FALSE))

	object@set.name = as.environment(gs.names)
	message('Done setting up user-specified geneset..')

	return(object)
}

#' Function to read custom mappability files
#'
#' This function reads a two-columned tab-delimited text file (with header). Expected
#' column names are 'mappa' and 'gene_id'. Each line is for a unique 'gene_id'
#' and contains the mappability (between 0 and 1) for that gene.
#'
#' @param file_path A file path for the custom mappability.
#'
#' @return A \code{data.frame} containing \code{gene_id} and \code{mappa} columns.
read_mappa = function(file_path) {

	d = read.table(file_path, sep = "\t", header = TRUE, stringsAsFactors = FALSE)

	# Check columns.
	if(!(all(colnames(d) %in% c('mappa','gene_id')))) {
		stop("Error: header must contain columns named 'mappa' and 'gene_id'")
	}

	# Genes in this file should not be duplicated.
	if (sum(duplicated(d$gene_id)) > 0) {
		stop("Error reading mappability data: duplicate gene_ids exist in file.")
	}

	# Mappability should be between 0 and 1.
	if (min(d$mappa) < 0) {
		stop("Error: mappability must be >= 0.")
	}

	if (max(d$mappa) > 1) {
		stop("Error: mappability must be <= 1.")
	}

	return(d)
}

context('Test read functions')

################################################################################
# Custom mappability definitions

test_that('Bad mappability header', {
	mappa_file = system.file('extdata', 'test_mappa_badhead.txt', package='chipenrich')
	expect_error(read_mappa(mappa_file), 'header must contain columns named')
})

test_that('Duplicate mappability gene_ids', {
	mappa_file = system.file('extdata', 'test_mappa_dup.txt', package='chipenrich')
	expect_error(read_mappa(mappa_file), 'duplicate gene_ids exist')
})

test_that('Negative mappability', {
	mappa_file = system.file('extdata', 'test_mappa_neg.txt', package='chipenrich')
	expect_error(read_mappa(mappa_file), 'mappability must be >= 0')
})

test_that('Too big mappability', {
	mappa_file = system.file('extdata', 'test_mappa_one.txt', package='chipenrich')
	expect_error(read_mappa(mappa_file), 'mappability must be <= 1')
})

test_that('Fine mappability', {
	mappa_file = system.file('extdata', 'test_mappa_good.txt', package='chipenrich')
	mappa = read_mappa(mappa_file)
	expect_true(all(mappa$gene_id == c(1,10,1000)))
})

################################################################################
# Custom locus definitions

ldef_noextras_file = system.file('extdata', 'test_ldef_noextras.txt', package='chipenrich')
ldef_geneid_file = system.file('extdata', 'test_ldef_geneid.txt', package='chipenrich')
ldef_symbol_file = system.file('extdata', 'test_ldef_symbol.txt', package='chipenrich')
ldef_nosymbol_file = system.file('extdata', 'test_ldef_nosymbol.txt', package='chipenrich')

test_that('Test no gene_id column throws error', {
	expect_error(read_ldef(ldef_noextras_file), 'Custom locus definition must have column')
})

test_that('Test renaming geneid to gene_id and NA genome', {
	ldef = read_ldef(ldef_geneid_file, genome = NA)

	expect_true(all(is.na(ldef@dframe$symbol)))
	expect_true(all(is.na(ldef@granges$symbol)))
})

test_that('Test renaming geneid to gene_id and hg19 genome', {
	ldef = read_ldef(ldef_geneid_file, genome = 'hg19')

	expect_true(all(ldef@dframe$symbol == c('SAMD11','SAMD11','HES4','SAMD11','SAMD11')))
	expect_true(all(ldef@granges$symbol == c('SAMD11','SAMD11','HES4','SAMD11','SAMD11')))
})

test_that('Test nosymbol and NA genome', {
	ldef = read_ldef(ldef_nosymbol_file, genome = NA)

	expect_true(all(is.na(ldef@dframe$symbol)))
	expect_true(all(is.na(ldef@granges$symbol)))
})

test_that('Test nosymbol and unsupported genome', {
	ldef = read_ldef(ldef_nosymbol_file, genome = 'ab5')

	expect_true(all(is.na(ldef@dframe$symbol)))
	expect_true(all(is.na(ldef@granges$symbol)))
})

test_that('Test nosymbol and hg19 genome', {
	ldef = read_ldef(ldef_nosymbol_file, genome = 'hg19')

	expect_true(all(ldef@dframe$symbol == c('A1BG','NAT2','CDH2')))
	expect_true(all(ldef@granges$symbol == c('A1BG','NAT2','CDH2')))
})

test_that('Test symbol and NA genome', {
	ldef = read_ldef(ldef_symbol_file, genome = NA)

	expect_true(all(ldef@dframe$symbol == c('HI','BYE','SURE')))
	expect_true(all(ldef@granges$symbol == c('HI','BYE','SURE')))
})

test_that('Test symbol and hg19 genome', {
	ldef = read_ldef(ldef_symbol_file, genome = 'hg19')

	expect_true(all(ldef@dframe$symbol == c('HI','BYE','SURE')))
	expect_true(all(ldef@granges$symbol == c('HI','BYE','SURE')))
})

test_that('Test symbol and unsupported genome', {
	ldef = read_ldef(ldef_symbol_file, genome = 'ab5')

	expect_true(all(ldef@dframe$symbol == c('HI','BYE','SURE')))
	expect_true(all(ldef@granges$symbol == c('HI','BYE','SURE')))
})

################################################################################
# Custom genesets

# Test building the geneset itself
geneset_file = system.file('extdata', 'test_geneset.txt', package='chipenrich')
test_geneset = suppressMessages(read_geneset(geneset_file))

test_that('Object of GeneSet class is returned',{
    expect_equal(class(test_geneset)[1], 'GeneSet')
})

test_that('Three genesets are created',{
    expect_equal(length(test_geneset@set.gene), 3)
})

test_that('Test names of genesets',{
    expect_equal(all(ls(test_geneset@set.gene) == c('GO:0035909','GO:0045822','GO:0045823')), TRUE)
})

################################################################################
# read_bed()

test_that('Uncompressed gff3, no genome', {
	peak_path = system.file('extdata', 'test.gff3', package='chipenrich')
	peaks = read_bed(peak_path)

	expect_equal(length(peaks), 19)

})

test_that('Compressed gff3, no genome', {
	peak_path = system.file('extdata', 'test.gff3.gz', package='chipenrich')
	peaks = read_bed(peak_path)

	expect_equal(length(peaks), 19)
})

test_that('BED3 with commented header', {
	peak_path = system.file('extdata', 'test_header.bed', package='chipenrich')
	peaks = suppressWarnings(read_bed(peak_path))

	expect_equal(length(peaks), 10)
	expect_warning(read_bed(peak_path), 'input regions overlap')
})

test_that('Uncompressed broadPeak', {
	peak_path = system.file('extdata', 'test.broadPeak', package='chipenrich')
	peaks = suppressWarnings(read_bed(peak_path))

	expect_equal(length(peaks), 27)
	expect_warning(read_bed(peak_path), 'input regions overlap')
})

test_that('Uncompressed narrowPeak', {
	peak_path = system.file('extdata', 'test.narrowPeak', package='chipenrich')
	peaks = read_bed(peak_path)

	expect_equal(length(peaks), 14)
})

test_that('Compressed broadPeak', {
	peak_path = system.file('extdata', 'test.broadPeak.gz', package='chipenrich')
	peaks = read_bed(peak_path)

	expect_equal(length(peaks), 17)
})

test_that('Compressed narrowPeak', {
	peak_path = system.file('extdata', 'test.narrowPeak.gz', package='chipenrich')
	peaks = read_bed(peak_path)

	expect_equal(length(peaks), 11)
})

test_that('Text input header', {
	peak_path = system.file('extdata', 'test_header.txt', package='chipenrich')
	peaks = suppressWarnings(read_bed(peak_path))

	expect_equal(length(peaks), 10)
	expect_warning(read_bed(peak_path), 'input regions overlap')
})

test_that('Text input no header', {
	peak_path = system.file('extdata', 'test_noheader.txt', package='chipenrich')
	peaks = suppressWarnings(read_bed(peak_path))

	expect_equal(length(peaks), 10)
	expect_warning(read_bed(peak_path), 'input regions overlap')
})

################################################################################
# load_peaks()

test_that('load_peaks() with and without genome', {
	peaks_df = data.frame(
		chr = c('chr1','chr2','chr3'),
		start = c(35,74,235),
		end = c(46,83,421),
		stringsAsFactors=F)

	# With genome
	peaks = load_peaks(peaks_df)
	expect_equal(length(peaks), 3)

	# Without genome
	peaks = load_peaks(peaks_df)

	expect_equal(length(peaks), 3)
})

test_that('load_peaks() with extra columns', {
	peaks_df = data.frame(
		chr = c('chr1','chr2','chr3'),
		start = c(35,74,235),
		end = c(46,83,421),
		strand = c('+','-','+'),
		score = c(36, 747, 13),
		stringsAsFactors=F)
	peaks = load_peaks(peaks_df, keep.extra.columns = TRUE)

	expect_equal(length(peaks), 3)
})

test_proxReg = function(geneset, peaks, regloc, n_cores) {
	# Restrict our genes/weights/peaks to only those genes in the genesets.
	# Here, geneset is not all combined, but GOBP, GOCC, etc.
	# i.e. A specific one.
	peaks = subset(peaks, peaks$gene_id %in% geneset@all.genes)
	
	# Run tests. NOTE: If os == 'Windows', n_cores is reset to 1 for this to work
	results_list = parallel::mclapply(as.list(ls(geneset@set.gene)), function(go_id) {
		single_proxReg(go_id, geneset, peaks, regloc)}, mc.cores = n_cores)
	
	# Collapse results into one table
	results = Reduce(rbind,results_list)
	
	# Correct for multiple testing
	results$FDR = p.adjust(results$P.value, method="BH")
	
	# Create enriched/depleted status column
	results$Status = ifelse(results$Effect > 0, 'closer', 'farther')
	
	results = results[order(results$P.value),]
	
	return(results)
}

single_proxReg = function(go_id, geneset, peaks, regloc) {

	# Genes in the geneset
	go_genes = geneset@set.gene[[go_id]]
	
	# Background genes and the background presence of a peak
	b_genes = peaks$gene_id %in% go_genes
	if (regloc == "tss"){
		x = peaks$scaled_dtss[b_genes]
		y = peaks$scaled_dtss[!b_genes]
	}
	if (regloc == "enhancer"){
		x = peaks$scaled_denh[b_genes]
		y = peaks$scaled_denh[!b_genes]
	}
	x <- x[is.finite(x)]
	y <- y[is.finite(y)]
	
	# Information about the geneset
	r_go_id = go_id
	r_go_genes_num = length(go_genes)

	# Information about peak genes
	go_genes_peak = peaks$gene_id[b_genes]
	r_go_genes_peak = length(table(go_genes_peak))
	r_go_genes_peak_num = length(go_genes_peak)

	r_effect = NA
	r_pval = NA
	
	if (length(x)== 0L | length(y) == 0L) {
		sprintf("Geneset: %s has zero peaks in one group. NAs given", go_id)
	} else {
		tryCatch(
			{ #Code from wilcox.test.default
			r <- rank(c(x, y))
			n.x <- as.double(length(x))
			n.y <- as.double(length(y))
			STATISTIC <- c(W = sum(r[seq_along(x)]) - n.x * (n.x + 1)/2)
			NTIES <- table(r)
			z <- STATISTIC - n.x * n.y/2
			SIGMA <- sqrt((n.x * n.y/12) * ((n.x + n.y + 1) - 
							sum(NTIES^3 - NTIES)/((n.x + n.y) * (n.x + n.y - 
								1))))
			z = z/SIGMA
			
			# Results from the Wilcoxon test
			r_effect = -z; #We want closer to be positive effect
            r_pval = 2 * min(stats::pnorm(z),
                             stats::pnorm(z, lower.tail = FALSE))
			},
			error = {function(e) {warning(
				sprintf("Error in geneset: %s. NAs given", go_id))
			}}
		)
	}
	
	out = data.frame(
		"P.value"=r_pval,
		"Geneset ID"=r_go_id,
		"N Geneset Genes"=r_go_genes_num,
		"Geneset Peak Genes"=r_go_genes_peak,
		"N Geneset Peak Genes"=r_go_genes_peak_num,
		"Effect"=r_effect,
		stringsAsFactors = FALSE)
	
	return(out)
}

test_fisher_exact = function(geneset,gpw,alternative="two.sided") {
	warning("Fisher's exact test should only be used with the 1kb or 5kb locus definition.")

	# Restrict to only those genes in the genesets.
	gpw = subset(gpw, gpw$gene_id %in% geneset@all.genes)

	genes = gpw$gene_id
	peaks = gpw$peak

	results = lapply(ls(geneset@set.gene), function(go_term) {
		go_genes = geneset@set.gene[[go_term]]
		n_go_genes = length(go_genes)

		in_cat = as.numeric(genes %in% go_genes)
		n_go_peak_genes = sum((genes %in% go_genes) & (peaks == 1))
		xt = table(in_cat, peaks)

		pval = 1
		odds_ratio = 0
		try({
			fet_result = stats::fisher.test(xt, conf.int = FALSE, alternative = alternative)
			pval = fet_result$p.value
			odds_ratio = fet_result$estimate
		}, silent = TRUE)

		go_peak_genes = gpw[(gpw$gene_id %in% go_genes) & (gpw$peak == 1),]$gene_id
		go_peak_genes_str = paste(go_peak_genes, collapse = ", ")

		enr = NA

		if (odds_ratio > 1) {
			enr = "enriched"
		}

		if (odds_ratio < 1) {
			enr = "depleted"
		}

		if (odds_ratio == 0) {
			enr = "depleted"
		}

		if (is.infinite(odds_ratio)) {
			enr = "enriched"
		}

		data.frame(
			"Geneset ID" = go_term,
			"N Geneset Genes" = n_go_genes,
			"N Geneset Peak Genes" = n_go_peak_genes,
			"Geneset Peak Genes" = go_peak_genes_str,
			"Odds.Ratio" = odds_ratio,
			"Status" = enr,
			"P-value" = pval,
			stringsAsFactors = FALSE)
	})

	results = plyr::rbind.fill(results)

	results$FDR = stats::p.adjust(results$P.value, method = "BH")

	results = results[order(results$P.value), ]
	return(results)
}

# Randomize the locus definition @dframe and rebuild the @granges and @chrom2iranges
randomize_ldef_bylocation = function(ldef, resolution=50) {
	message('Randomizing locus definition in location bins...')
	# Extract GRanges representation
	ldef_gr = ldef@granges

	# Split by chromosome
	ldef_grl = S4Vectors::split(ldef_gr, Seqinfo::seqnames(ldef_gr))

	# Within each chromosome:
	# 1. Form groups based on the number of genes on that chromosome
	# 2. Split the chromosome GRanges by the groups (GRangesList)
	# 3. Within each group within each chromosome, scramble the mcols
	# 4. Collapse the groups so the result is a GRanges for the chromosome again (empty GRanges for chromosomes without genes).
	re_ldef_grl = S4Vectors::endoapply(ldef_grl, function(gr) {
		if(length(gr) > 0) {
			# Create groups within the chromosome
			group = floor( (seq_along(gr) + (resolution - 1)) / resolution )
			# Split by the group
			grl = S4Vectors::splitAsList(gr, group)
			# Rearrange mcols within each group
			re_grl = S4Vectors::endoapply(grl, function(grg){
				GenomicRanges::mcols(grg) = GenomicRanges::mcols(grg)[sample(seq_along(grg), length(grg)),]
				return(grg)
			})
			# Collapse the GRangesList into the scrambled GRanges for the chromosome
			re_gr = BiocGenerics::unlist(re_grl, use.names = FALSE)
		} else {
			# Need this for chromosomes without genes
			re_gr = GenomicRanges::GRanges()
		}

		return(re_gr)
	})
	# Collapse the GRangesList across the chromosomes
	re_ldef_gr = unlist(re_ldef_grl, use.names = FALSE)
	# Make sure it's sorted
	re_ldef_gr = sort(re_ldef_gr)

	# Construct data.frame for new locus definition
	re_ldef_df = data.frame(re_ldef_gr, stringsAsFactors = FALSE)
	re_ldef_df = re_ldef_df[, c('seqnames','start','end','gene_id','symbol')]
	colnames(re_ldef_df) = c('chr','start','end','gene_id','symbol')

	# Reassign the dframe and granges to the ldef
	ldef@dframe = re_ldef_df
	ldef@granges = re_ldef_gr

	return(ldef)
}

# Randomize ppg after all additions have been made across all genes
randomize_ppg_all = function(ppg) {
	ppg = ppg[order(ppg$length),]
	rownames(ppg) = 1:nrow(ppg)

	reordering = sample(1:nrow(ppg), nrow(ppg))
	ppg = data.frame('gene_id'=ppg$gene_id, ppg[reordering,2:ncol(ppg)], stringsAsFactors = FALSE)

	return(ppg)
}

# Randomize ppg after all additions have been made across all genes
randomize_ldef_complete = function(ldef) {
	message('Randomizing locus definition...')
	gr = ldef@granges
	df = ldef@dframe

	shuffle = base::sample(seq_along(gr), length(gr))

	GenomicRanges::mcols(gr) = GenomicRanges::mcols(gr)[shuffle, ]
	df[, c('gene_id', 'symbol')] = df[shuffle, c('gene_id', 'symbol')]

	ldef@granges = gr
	ldef@dframe = df

	return(ldef)
}

# Randomize ppg after all additions have been made within length bins
randomize_ppg_length = function(ppg) {
	ppg = ppg[sample(1:nrow(ppg),nrow(ppg)),]
	ppg = ppg[order(ppg$length),]
	rownames(ppg) = 1:nrow(ppg)

	group = floor(as.numeric(rownames(ppg))+99)/100
	group = floor(group)

	split_ppg = split(ppg, group)
	split_ppg = lapply(split_ppg, function(bin){
		reordering = sample(1:nrow(bin), nrow(bin))

		data.frame('gene_id'=bin$gene_id, bin[reordering,2:ncol(bin)], stringsAsFactors = FALSE)
	})
	ppg = Reduce(rbind, split_ppg)

	return(ppg)
}

randomize_ldef_bylength = function(ldef, resolution = 100) {
	message('Randomizing locus definition in locus length bins...')
	gr = ldef@granges
	df = ldef@dframe

	widths = data.frame(
		original_idx = seq_along(gr),
		width = GenomicRanges::width(gr),
		stringsAsFactors = FALSE)
	widths = widths[order(widths$width), ]
	rownames(widths) = seq_along(gr)

	group = floor(as.numeric(rownames(widths))+ (resolution - 1)) / resolution
	group = floor(group)

	split_widths = split(widths, group)
	split_widths = lapply(split_widths, function(bin){
		shuffle = base::sample(1:nrow(bin), nrow(bin))
		bin$shuffle_idx = bin$original_idx[shuffle]

		return(bin)
	})
	widths = Reduce(rbind, split_widths)

	GenomicRanges::mcols(gr)[widths$original_idx, ] = GenomicRanges::mcols(gr)[widths$shuffle_idx, ]
	df[widths$original_idx, c('gene_id', 'symbol')] = df[widths$shuffle_idx, c('gene_id', 'symbol')]

	ldef@granges = gr
	ldef@dframe = df

	return(ldef)
}

context('Test randomize functions')

################################################################################
# Test complete randomization
test_that('Test complete locusdef randomization', {
	data(locusdef.dm6.nearest_tss, package = 'chipenrich.data')

	randomized = randomize_ldef_complete(locusdef.dm6.nearest_tss)

	expect_equal(nrow(locusdef.dm6.nearest_tss@dframe), nrow(randomized@dframe))
	expect_equal(length(locusdef.dm6.nearest_tss@granges), length(randomized@granges))
	expect_equal(locusdef.dm6.nearest_tss@genome.build, randomized@genome.build)
	expect_equal(locusdef.dm6.nearest_tss@organism, randomized@organism)
	expect_true(all(locusdef.dm6.nearest_tss@granges$gene_id %in% randomized@granges$gene_id))
})

################################################################################
# Test bylength randomization by location
test_that('Test locusdef randomization by length', {
	data(locusdef.dm6.nearest_tss, package = 'chipenrich.data')

	randomized = randomize_ldef_bylength(locusdef.dm6.nearest_tss, resolution = 100)

	expect_equal(nrow(locusdef.dm6.nearest_tss@dframe), nrow(randomized@dframe))
	expect_equal(length(locusdef.dm6.nearest_tss@granges), length(randomized@granges))
	expect_equal(locusdef.dm6.nearest_tss@genome.build, randomized@genome.build)
	expect_equal(locusdef.dm6.nearest_tss@organism, randomized@organism)
	expect_true(all(locusdef.dm6.nearest_tss@granges$gene_id %in% randomized@granges$gene_id))
})

################################################################################
# Test locusdef randomization by location
test_that('Test locusdef randomization by location', {
	data(locusdef.dm6.nearest_tss, package = 'chipenrich.data')

	randomized = randomize_ldef_bylocation(locusdef.dm6.nearest_tss, resolution = 100)

	expect_equal(nrow(locusdef.dm6.nearest_tss@dframe), nrow(randomized@dframe))
	expect_equal(length(locusdef.dm6.nearest_tss@granges), length(randomized@granges))
	expect_equal(locusdef.dm6.nearest_tss@genome.build, randomized@genome.build)
	expect_equal(locusdef.dm6.nearest_tss@organism, randomized@organism)
	expect_true(all(locusdef.dm6.nearest_tss@granges$gene_id %in% randomized@granges$gene_id))
})

################################################################################
# Test randomizations without binning
test_that('Test ppg randomization without length bins', {
	ppg_file = system.file('extdata', 'test_ppg.txt', package = 'chipenrich')
	ppg = read.table(ppg_file, header = T, sep='\t', stringsAsFactors=F)

	randomized = randomize_ppg_all(ppg)

	expect_equal(mean(ppg$length), mean(randomized$length))
	expect_equal(mean(ppg$log10_length), mean(randomized$log10_length))
	expect_equal(mean(ppg$num_peaks), mean(randomized$num_peaks))
	expect_true(all(ppg$gene_id %in% randomized$gene_id))
})

################################################################################
# Test randomization in bins
test_that('Test ppg randomization within length bins', {
	ppg_file = system.file('extdata', 'test_ppg.txt', package = 'chipenrich')
	ppg = read.table(ppg_file, header = T, sep='\t', stringsAsFactors=F)

	randomized = randomize_ppg_length(ppg)

	expect_equal(mean(ppg$length), mean(randomized$length))
	expect_equal(mean(ppg$log10_length), mean(randomized$log10_length))
	expect_equal(mean(ppg$num_peaks), mean(randomized$num_peaks))
	expect_true(all(ppg$gene_id %in% randomized$gene_id))
})

#' Run Proximity Regulation test on a set of narrow genomic regions
#' 
#' This method is designed for a set of narrow genomic regions (e.g. TF peaks) and is used to test
#' whether the genomic regions assigned to genes in a gene set are closer to 
#' regulatory locations (i.e. promoters or enhancers) than by chance.
#' 
#' @section Regulatory locations:
#' Current supported regulatory locations are gene transcription
#' start sites (tss) or enhancer locations (hg19 only)
#' 
#' @section Method:
#' ProxReg first calculates the distance between each peak midpoint and 
#' regulatory location in base pairs. For gene transcription start sites, 
#' since parts of the chromosome are more sparse than others, there is an
#' association with gene locus length that needs to be adjusted for.
#' When using tss as the regulatory location, the peak distances are 
#' adjusted for this confounding variable based on an average of 90 ENCODE
#' ChIP-seq experiments (details in citation pending). Similarly, for enhancers, 
#' distances depend on the density of enhancers within a gene locus, so distance
#' to enhancer is adjusted using an empirical average of 90 ChIP-seq ENCODE
#' experiments.
#' 
#' For each gene set of interest, the genomic regions are divided into two groups indicating
#' the gene with the nearest tss is in the gene set or not. A Wilcoxon Rank-Sum test is 
#' then done to test for a difference in the adjusted distances (either to tss or enhancer).
#' 
#' @param peaks Either a file path or a \code{data.frame} of peaks in BED-like
#' format. If a file path, the following formats are fully supported via their
#' file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
#' or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
#' without these extensions are supported under the conditions that the first 3
#' columns correspond to 'chr', 'start', and 'end' and that there is either no
#' header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
#' \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
#' acceptable column names.
#' @param out_name Prefix string to use for naming output files. This should not
#' contain any characters that would be illegal for the system being used (Unix,
#' Windows, etc.) The default value is "proxReg", and a file "proxReg_results.tab"
#' is produced. If \code{qc_plots} is set, then a file "proxReg_qcplots.pdf"
#' is produced containing a number of quality control plots. If \code{out_name}
#' is set to NULL, no files are written, and results then must be retrieved from
#' the list returned by \code{proxReg}.
#' @param out_path Directory to which results files will be written out. Defaults
#' to the current working directory as returned by \code{\link{getwd}}.
#' @param genome One of the \code{supported_genomes()}. If reglocation = enhancer,
#' genome MUST be 'hg19'.
#' @param genesets A character vector of geneset databases to be tested for
#' enrichment. See \code{supported_genesets()}. Alternately, a file path to
#' a tab-delimited text file with header and first column being the geneset ID
#' or name, and the second column being Entrez Gene IDs. For an example custom
#' gene set file, see the vignette.
#' @param randomization One of: 'shuffle', 'unif', 'bylength', 'byenh'. These were used to
#' test for Type I error under the null hypothesis. A general user will never have to
#' use these.
#' @param reglocation One of: 'tss', 'enhancer'. Details in the "Regulatory locations" section
#' @param qc_plots A logical variable that enables the automatic generation of
#' plots for quality control.
#' @param min_geneset_size Sets the minimum number of genes a gene set may have
#' to be considered for testing.
#' @param max_geneset_size Sets the maximum number of genes a gene set may have
#' to be considered for testing.
#' @param n_cores The number of cores to use for testing. We recommend
#' using only up to the maximum number of \emph{physical} cores present, as
#' virtual cores do not significantly decrease runtime. Default number of cores
#' is set to 1. NOTE: Windows does not support multicore testing.
#' 
#' @return A list, containing the following items:
#'
#' \item{opts }{A data frame containing the arguments/values passed to \code{polyenrich}.}
#'
#' \item{peaks }{
#' A data frame containing peak assignments to genes. Peaks which do not overlap
#' a gene locus are not included. Each peak that was assigned to a gene is listed,
#' along with the peak midpoint or peak interval coordinates (depending on which
#' was used), the gene to which the peak was assigned, the locus start and end
#' position of the gene, and the distance from the peak to the TSS.
#'
#' The columns are:
#'
#' \describe{
#'   \item{peak_id}{an ID given to unique combinations of chromosome, peak start, and peak end. }
#'   \item{chr}{the chromosome the peak originated from. }
#'   \item{peak_start}{start position of the peak. }
#'   \item{peak_end}{end position of the peak. }
#'   \item{gene_id}{the Entrez ID of the gene to which the peak was assigned. }
#'   \item{gene_symbol}{the official gene symbol for the gene_id (above). }
#'   \item{gene_locus_start}{the start position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
#'   \item{gene_locus_end}{the end position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
#'   \item{nearest_tss}{the closest TSS to this peak (for any gene, not necessarily the gene this peak was assigned to.) }
#'   \item{dist_to_tss}{the distance in bp to the closest TSS to this peak. }
#'   \item{nearest_tss_gene}{the gene having the closest TSS to the peak (should be the same as gene_id when using the nearest TSS locus definition.) }
#'   \item{nearest_tss_gene_strand}{the strand of the gene with the closest TSS. }
#'   \item{log_dtss}{ log of dist_to_tss}
#'   \item{log_gene_ll}{the log of length of the gene locus in bp }
#'   \item{scaled_dtss}{the adjusted distance to TSS, used in the calculations. Shown if reglocation = "tss" }
#'   \item{dist_to_enh}{the distance to the nearest enhancer. Shown if reglocation = "enhancer" }
#'   \item{avg_denh}{the empirical average for distance to the nearest enhancer for the gene the peak is assigned to. Shown if reglocation = enhancer }
#'   \item{scaled_denh}{the adjusted distance to the nearest enhancer. Shown if reglocation = enhancer }

#' }}
#'
#' \item{results }{
#' A data frame of the results from performing the proxReg test on
#' each geneset that was requested (all genesets are merged into one final data
#' frame.) The columns are:
#'
#' \describe{
#'   \item{Geneset.ID}{the identifier for a given gene set from the selected database.  For example, GO:0000003. }
#'   \item{Geneset.Type}{ specifies from which database the Geneset.ID originates.  For example, "Gene Ontology Biological Process."}
#'   \item{Description}{ gives a definition of the geneset. For example, "reproduction."}
#'   \item{P.Value}{the probability of observing the proxmity of genomic regions in the gene set given the null hypothesis that peaks are not closer or farther in the gene set.}
#'   \item{FDR}{the false discovery rate proposed by Bejamini \& Hochberg for adjusting the p-value to control for family-wise error rate.}
#'   \item{Effect}{the signed Wilcoxon statistic, with positive values meaning the gene set has closer genomic regions than expected by chance.}
#'   \item{Status}{specifies if the peaks in the gene set tend to be closer or farther than those not in the gene set.}
#'   \item{Odds.Ratio}{the estimated odds that peaks are associated with a given gene set compared to the odds that peaks are associated with other gene sets, after controlling for locus length and/or mappability.  An odds ratio greater than 1 indicates enrichment, and less than 1 indicates depletion.}
#'   \item{N.Geneset.Genes}{the number of genes in the gene set.}
#'   \item{N.Geneset.Peak.Genes}{the number of genes in the genes set that were assigned at least one peak.}
#'   \item{Geneset.Peak.Genes}{the list of genes from the gene set that had at least one peak assigned.}
#' }}
#' 
#' @examples 
#' # Run proxReg using an example dataset, assigning peaks to the nearest TSS,
#' # and on a small custom geneset
#' data(peaks_E2F4, package = 'chipenrich.data')
#' peaks_E2F4 = subset(peaks_E2F4, peaks_E2F4$chrom == 'chr1')
#' gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
#' results = proxReg(peaks_E2F4, reglocation = 'tss',
#' 			genome = 'hg19', genesets=gs_path, out_name=NULL)
#'
#' # Get the list of peaks that were assigned to genes and their distances to 
#' # regulatory regions.
#' assigned_peaks = results$peaks
#'
#' # Get the results of enrichment testing.
#' enrich = results$results
#' 
#' @export
#' @include constants.R utils.R supported.R setup.R
#' @include read.R assign_peaks.R 
#' @include plot_dist_to_tss.R
#' @include test_proxReg.R
proxReg = function(
	peaks,
	out_name = "proxReg",
	out_path = getwd(),
	genome = supported_genomes(),
	reglocation = "tss",
	genesets = c(
		'GOBP',
		'GOCC',
		'GOMF'),
	randomization = NULL,
	qc_plots = TRUE,
	min_geneset_size = 15,
	max_geneset_size = 2000,
	n_cores = 1
) {
	# Kill if reglocation is not tss or enhancer
	if (!(reglocation %in% c("tss","enhancer"))) {
		stop("Unsupported regulatory location!")
	}
	
	# Set genome to hg due to enhancer-specific 
	if ("enhancer" %in% reglocation & genome != "hg19" ) {
		stop("Genome must be hg19 to use enhancer regulatory location")
	}
	
	genome = match.arg(genome)
	
	n_cores = reset_ncores_for_windows(n_cores)
	
	############################################################################
	# Collect options for opts output
	opts_list = as.list(sys.call())
	opts_list = opts_list[2:length(opts_list)]
	
	opts = data.frame(
		parameters = names(opts_list),
		values = as.character(opts_list),
		stringsAsFactors = FALSE
	)
	
	############################################################################
	# Setup locus definitions, genesets, and mappability
	
	#Stop if enhancer reglocation is selected and genome is not hg19
	if (reglocation == "enhancer" & genome != "hg19") {
		stop("Enhancer regulatory location only compatible with hg19 genome")
	}
	
	#Locus definition will always be NTSS for this
	ldef_list = setup_locusdef(ldef_code = "nearest_tss", genome, randomization = NULL)
	ldef = ldef_list[['ldef']]
	tss = ldef_list[['tss']]
	
	geneset_list = setup_genesets(gs_codes = genesets, ldef_obj = ldef, genome = genome, min_geneset_size = min_geneset_size, max_geneset_size = max_geneset_size)
	
	############################################################################
	############################################################################
	# Start enrichment process
	############################################################################
	############################################################################
	
	######################################################
	# Read in and format peaks (from data.frame or file)
	if (class(peaks) == "data.frame") {
		message('Reading peaks from data.frame...')
		peakobj = load_peaks(peaks)
	} else if (class(peaks) == "character") {
		peakobj = read_bed(peaks)
	}
	
	# Number of peaks in data.
	num_peaks = length(peakobj)
	
	######################################################
	# Assign peaks to genes.
	message("Assigning peaks to genes with assign_peaks(...) ..")
	assigned_peaks = assign_peaks(peakobj, ldef, tss)
	# Creating a column for adjusted DTSS
	if ("tss" %in% reglocation) {
		assigned_peaks$log_dtss = log(abs(assigned_peaks$dist_to_tss)+1)
		assigned_peaks$log_gene_ll = log(assigned_peaks$gene_locus_end-assigned_peaks$gene_locus_start)
		pred_log_dtss = as.numeric(mgcv::predict.gam(chipenrich.data::spline.log_dtss.90ENCODE, assigned_peaks, type="link"))
		assigned_peaks$scaled_dtss = assigned_peaks$log_dtss-pred_log_dtss
	}
	# Creating a column for enhancer distances
	if ("enhancer" %in% reglocation) {
		enhancers = chipenrich.data::enhancer.hg19
		gene.enh.desc = chipenrich.data::gene.enh.desc.hg19
		peakobj2 = GenomicRanges::makeGRangesFromDataFrame(assigned_peaks,
					seqnames.field = "chr", start.field = "peak_start",end.field = "peak_end")
		peak_mids = IRanges::mid(GenomicRanges::ranges(peakobj2))
		mids_gr = GenomicRanges::GRanges(
			seqnames = Seqinfo::seqnames(peakobj2),
			ranges = IRanges::IRanges(start = peak_mids, end = peak_mids),
			name = GenomicRanges::mcols(peakobj2)$name
		)
		enhancer_mids = IRanges::mid(GenomicRanges::ranges(enhancers))
		enhancer_mids_gr = GenomicRanges::GRanges(
			seqnames = Seqinfo::seqnames(enhancers),
			ranges = IRanges::IRanges(start = enhancer_mids, end = enhancer_mids),
			name = GenomicRanges::mcols(enhancers)$name
		)
		dist_to_enh = GenomicRanges::distanceToNearest(mids_gr, enhancer_mids_gr)
		assigned_peaks$dist_to_enh = dist_to_enh@elementMetadata$distance
		assigned_peaks$log_dtss = log(abs(assigned_peaks$dist_to_tss)+1)
		assigned_peaks$log_gene_ll = log(assigned_peaks$gene_locus_end-assigned_peaks$gene_locus_start)
		pred_log_dtss = as.numeric(mgcv::predict.gam(chipenrich.data::spline.log_dtss.90ENCODE, assigned_peaks, type="link"))
		assigned_peaks$avgdenh = sapply(assigned_peaks$gene_id, 
										function(x) {a = gene.enh.desc$avg_denh_emp[gene.enh.desc$gene_id == x]
                                            return(ifelse(length(a)==0,NA,a))
                                        })

        if (any(is.na(assigned_peaks$avgdenh))) { #Impute the missing genes as average of the others
        		message("Aha, is this the part that not working?")
        	dlm = stats::lm(avgdenh ~ log_gene_ll, data = assigned_peaks)
            assigned_peaks$avgdenh[is.na(assigned_peaks$avgdenh)] = stats::predict.lm(dlm, assigned_peaks[is.na(assigned_peaks$avgdenh),])
     		rm(dlm)
        }
        
		assigned_peaks$scaled_denh = log(abs(assigned_peaks$dist_to_enh)+1) - log(assigned_peaks$avgdenh+1)
	}
	
	#Randomizations
	if (!is.null(randomization)){
		if (randomization == "shuffle") { #Just shuffle gene ids
			message("Randomizing by shuffling gene ids...")
			assigned_peaks$gene_id = sample(assigned_peaks$gene_id)
		} else if (randomization == "unif") { #Uniformly pick from random available genes
			message("Randomizing by uniformly picking gene ids...")
			assigned_peaks$gene_id = as.integer(sample(names(table(assigned_peaks$gene_id)), nrow(assigned_peaks), replace = T))
		} else if (randomization == "bylength") { #bins of locus length, style of unif
			message("Randomizing by picking gene ids by locus length...")
			peakstemp = assigned_peaks[,c("peak_id","gene_id","gene_locus_start","gene_locus_end")]
			peakstemp$gene_locus_length = peakstemp$gene_locus_end-peakstemp$gene_locus_start
			peakstemp$dupe_genes = duplicated(peakstemp$gene_id)
			peakstemp = peakstemp[sample(1:nrow(peakstemp)),]
			peakstemp2 = peakstemp[!peakstemp$dupe_genes,]
			peakstemp2 = peakstemp2[order(peakstemp2$gene_locus_length),]
			rownames(peakstemp2) = 1:nrow(peakstemp2)
			peakstemp2$group = floor((as.numeric(rownames(peakstemp2))+99)/100)
			peakstemp$group = sapply(peakstemp$gene_id, function(x){peakstemp2$group[peakstemp2$gene_id==x]})
			peakstemp$gene_id = sapply(peakstemp$group, function(x){sample(peakstemp2$gene_id[peakstemp2$group==x],1)})
			assigned_peaks$gene_id_pre = assigned_peaks$gene_id
			assigned_peaks = merge(assigned_peaks[,-5], peakstemp[,c("peak_id","gene_id")], by = "peak_id")
		} else if (randomization == "byenh"){
			message("Randomizing by picking gene ids by enhancer dist...")
			peakstemp = assigned_peaks[,c("peak_id","gene_id","avgdenh")]
			peakstemp$dupe_genes = duplicated(peakstemp$gene_id)
			peakstemp = peakstemp[sample(1:nrow(peakstemp)),]
			peakstemp2 = peakstemp[!peakstemp$dupe_genes,]
			peakstemp2 = peakstemp2[sample(1:nrow(peakstemp2)),]
			peakstemp2 = peakstemp2[order(peakstemp2$avgdenh),]
			rownames(peakstemp2) = 1:nrow(peakstemp2)
			peakstemp2$group = floor((as.numeric(rownames(peakstemp2))+499)/500)
			peakstemp$group = sapply(peakstemp$gene_id, function(x){peakstemp2$group[peakstemp2$gene_id==x]})
			peakstemp$gene_id = sapply(peakstemp$group, function(x){sample(peakstemp2$gene_id[peakstemp2$group==x],1)})
			assigned_peaks$gene_id_pre = assigned_peaks$gene_id
			assigned_peaks = merge(assigned_peaks[,-5], peakstemp[,c("peak_id","gene_id")], by = "peak_id")
			
		} else {
			stop("Unsupported randomization!")
		}
	}
	
	######################################################
	# Enrichment
	results = list()
	for (gobj in geneset_list) {
		message(sprintf("Test: Proximity test"))
		message(sprintf("Genesets: %s",gobj@type))
		message("Running tests..")
		
		if ("tss" %in% reglocation) {
			message("Running proximity to TSS test...")
			rtemp = test_proxReg(gobj, assigned_peaks, regloc = "tss", n_cores)
		} else if ("enhancer" %in% reglocation) {
			message("Running proximity to enhancers test...")
			rtemp = test_proxReg(gobj, assigned_peaks, regloc = "enhancer", n_cores)
		}
		
		# Annotate with geneset descriptions.
		rtemp$"Description" = as.character(mget(rtemp$Geneset.ID, gobj@set.name, ifnotfound=NA))
		rtemp$"Geneset.Type" = gobj@type
		
		results[[gobj@type]] = rtemp
	}
	enrich = Reduce(rbind,results)
	
	######################################################
	# Post-process enrichment
	# Order columns, add enriched/depleted column as needed, remove bad tests,
	# sort by p-value, rename rownames to integers
	enrich = post_process_enrichments(enrich)
	
	######################################################
	# Write result objects to files
	if (!is.null(out_name)) {
		filename_analysis = file.path(out_path, sprintf("%s_results.tab", out_name))
		write.table(enrich, file = filename_analysis, row.names = FALSE, quote = FALSE, sep = "\t")
		message("Wrote results to: ", filename_analysis)
		
		filename_peaks = file.path(out_path, sprintf("%s_peaks.tab", out_name))
		write.table(assigned_peaks, file = filename_peaks, row.names = FALSE, quote = FALSE, sep = "\t")
		message("Wrote peak-to-gene assignments to: ", filename_peaks)
		
		filename_opts = file.path(out_path, sprintf("%s_opts.tab", out_name))
		write.table(opts, file = filename_opts, row.names = FALSE, quote = FALSE, sep = "\t")
		message("Wrote run options/arguments to: ", filename_opts)
		
		if (qc_plots) {
			filename_qcplots = file.path(out_path, sprintf("%s_qcplots.png", out_name))
			grDevices::png(filename_qcplots)
			#print(..plot_proxReg_spline(peaks = assigned_peaks, num_peaks = num_peaks))
			print(..plot_dist_to_tss(peakobj, tss))
			grDevices::dev.off()
			message("Wrote QC plots to: ",filename_qcplots)
		}
	}
	
	######################################################
	# Return objects as list
	return(list(
		peaks = assigned_peaks,
		results = enrich,
		opts = opts
	))
	
}

# Used in ..plot_chipenrich_spline(...)
avg_binned_peak = function(gpw) {
	bygroup = stats::aggregate(cbind(peak, length) ~ group, gpw, mean)
	bygroup$log_avg_length = log10(bygroup$length)
	names(bygroup) = c("group", "peak", "avg_length", "log_avg_length")

	return(bygroup)
}

# Used in ..plot_chipenrich_spline(...)
calc_weights_chipenrich = function(gpw) {
	# Create model.
	model = "peak ~ s(log10_length, bs = 'cr')"

	# Compute binomial spline fit.
	fit = gam(as.formula(model), data = gpw, family = "binomial")

	# Compute weights for each gene, based on the predicted prob(peak) for each gene.
	ppeak = fitted(fit)

	# These are not used, likely part of development
	# Perhaps of use for Chris in weighted polyenrich
	w0 = 1 / (ppeak / mean(gpw$peak, na.rm = TRUE))
	w0 = w0 / mean(w0, na.rm = TRUE)
	gpw$weight = w0
	gpw$resid.dev = resid(fit, type="deviance")

	# This is the only part that gets used
	gpw$prob_peak = ppeak

	return(gpw)
}

#' QC plot for ChIP-Enrich
#'
#' A plot showing an approximation of the empirical spline used to model the
#' relationship between a gene having a peak and the locus length. For visual
#' clarity, genes are binned into groups of 25 after sorting by locus length.
#' Expected fits assuming independence of locus length and presence of a peak,
#' and assuming proportionality of locus length and presence of a peak are given
#' to demonstrate deviation from either for the dataset.
#'
#' @param peaks Either a file path or a \code{data.frame} of peaks in BED-like
#' format. If a file path, the following formats are fully supported via their
#' file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
#' or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
#' without these extensions are supported under the conditions that the first 3
#' columns correspond to 'chr', 'start', and 'end' and that there is either no
#' header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
#' \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
#' acceptable column names.
#' @param locusdef One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
#' '1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
#' '10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
#' see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
#' a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
#' must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
#' example custom locus definition file, see the vignette.
#' @param genome One of the \code{supported_genomes()}.
#' @param mappability One of \code{NULL}, a file path to a custom mappability file,
#' or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
#' If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
#' Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
#' For an example custom mappability file, see the vignette. Default value is NULL.
#' @param legend If true, a legend will be drawn on the plot.
#' @param xlim Set the x-axis limit. NULL means select x-lim automatically.
#'
#' @return A trellis plot object.
#'
#' @examples
#'
#' # Spline plot for E2F4 example peak dataset.
#' data(peaks_E2F4, package = 'chipenrich.data')
#'
#' # Create the plot for a different locus definition
#' # to compare the effect.
#' plot_chipenrich_spline(peaks_E2F4, locusdef = 'nearest_gene', genome = 'hg19')
#'
#' @export
#' @include constants.R utils.R supported.R setup.R randomize.R
#' @include read.R assign_peaks.R peaks_per_gene.R
plot_chipenrich_spline = function(peaks, locusdef = "nearest_tss", genome = supported_genomes(), mappability = NULL, legend = TRUE, xlim = NULL) {
	genome = match.arg(genome)

	ldef_list = setup_locusdef(locusdef, genome)
	ldef = ldef_list[['ldef']]
	tss = ldef_list[['tss']]

	mappa = setup_mappa(mappa_code = mappability, genome = genome, ldef_code = locusdef, ldef_obj = ldef)

	if (class(peaks) == "data.frame") {
		peakobj = load_peaks(peaks)
	} else if (class(peaks) == "character") {
		peakobj = read_bed(peaks)
	}
	num_peaks = length(peakobj)

	assigned_peaks = assign_peaks(peakobj, ldef, tss)
	ppg = num_peaks_per_gene(assigned_peaks, ldef, mappa)

	# Make plot.
	plotobj = ..plot_chipenrich_spline(gpw = ppg, mappability = mappability, num_peaks = num_peaks, legend = legend, xlim = xlim)
	return(plotobj)
}

..plot_chipenrich_spline = function(gpw, mappability, num_peaks, legend = TRUE, xlim = NULL) {
	############################################################################
	# Prepare the gpw table
	gpw = gpw[order(gpw$log10_length), ]
	gpw$group = ceiling((1:nrow(gpw)) / 25)

	############################################################################
	# Quantities for the plot

	# Scatterplot stuff
	avg_bins = avg_binned_peak(gpw)

	# Spline stuff
	gpw = calc_weights_chipenrich(gpw) # gpw = genes, peaks, weights
	genome_length = sum(as.numeric(gpw$length))
	gpw$false_prob = 1 - (1 - (gpw$length / genome_length))^(num_peaks)

	############################################################################
	# Plotting parameters
	col_rand_gene = "grey35"
	col_rand_peak = "grey74"
	col_spline = "darkorange"

	panel_func = function(x,y,...) {
		lattice::panel.xyplot(x, y, ...)
		lattice::panel.abline(h = mean(y), col = col_rand_gene, lwd = 3)
	}

	custom_key = list(
		text = list(
		c(
			"Expected Fit - Peaks Independent of Locus Length",
			"Expected Fit - Peaks Proportional to Locus Length",
			"Binomial Smoothing Spline Fit",
			"Proportion of Genes in Bin with at Least 1 Peak"
			)
		),
		lines = list(
			pch = c(20, 15, 17, 20),
			type = c("l", "l", "l", "p"),
			col = c(col_rand_gene, col_rand_peak, col_spline, "black"),
			lwd = 3
		),
		cex = 1.2
	)

	if (!legend) {
		custom_key = NULL
	}

	if (is.null(mappability)) {
		xlab = expression(paste(Log[10], " Locus Length"))
	} else {
		xlab = expression(paste(Log[10], " Mappable Locus Length"))
	}

	xmin_nopad = base::ifelse(is.null(xlim[1]), floor(min(gpw$log10_length)), floor(xlim[1]))
	xmax_nopad = base::ifelse(is.null(xlim[2]), ceiling(max(gpw$log10_length)), ceiling(xlim[2]))

	scales = list(
		x = list(
			axs = 'i',
			at = seq(xmin_nopad, xmax_nopad, 1)
		),
		y = list(
			axs = 'i',
			at = seq(0, 1, 0.2)
		)
	)

	############################################################################
	# Plot
	plotobj = lattice::xyplot(
		false_prob + prob_peak ~ log10_length,
		gpw,
		xlab = list(label = xlab, cex = 1.4),
		ylab = list(label = "Prop. of Genes with at Least 1 Peak", cex = 1.4),
		ylim = c(-0.05, 1.05),
		xlim = c(xmin_nopad - 0.5, xmax_nopad + 0.5),
		panel = panel_func,
		type = "l",
		lwd = 3,
		key = custom_key,
		scales = list(cex = 1.4),
		par.settings = lattice::simpleTheme(pch = c(15,17), col = c(col_rand_peak, col_spline))
	) + latticeExtra::as.layer(lattice::xyplot(peak ~ log_avg_length, avg_bins, pch = 20, cex = 0.4, col = "black"))

	return(plotobj)
}

# Used in ..plot_gene_coverage(...)
avg_binned_coverage = function(gpw) {
	bygroup = stats::aggregate(cbind(ratio, length) ~ group, gpw, mean)
	bygroup$log_avg_length = log10(bygroup$length)
	names(bygroup) = c("group", "ratio", "length", "log_avg_length")

	return(bygroup)
}

#' QC plot for Broad-Enrich
#'
#' Create a plot showing the relationship between locus length and the
#' proportion of gene loci covered by peaks.
#'
#' @param peaks Either a file path or a \code{data.frame} of peaks in BED-like
#' format. If a file path, the following formats are fully supported via their
#' file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
#' or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
#' without these extensions are supported under the conditions that the first 3
#' columns correspond to 'chr', 'start', and 'end' and that there is either no
#' header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
#' \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
#' acceptable column names.
#' @param locusdef One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
#' '1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
#' '10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
#' see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
#' a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
#' must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
#' example custom locus definition file, see the vignette.
#' @param genome One of the \code{supported_genomes()}.
#' @param mappability One of \code{NULL}, a file path to a custom mappability file,
#' or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
#' If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
#' Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
#' For an example custom mappability file, see the vignette. Default value is NULL.
#' @param legend If true, a legend will be drawn on the plot.
#' @param xlim Set the x-axis limit. NULL means select x-lim automatically.
#'
#' @return A trellis plot object.
#'
#' @examples
#'
#' # Spline plot for E2F4 example peak dataset.
#' data(peaks_H3K4me3_GM12878, package = 'chipenrich.data')
#'
#' # Create the plot for a different locus definition
#' # to compare the effect.
#' plot_gene_coverage(peaks_H3K4me3_GM12878, locusdef = 'nearest_gene', genome = 'hg19')
#'
#' @export
#' @include constants.R utils.R supported.R setup.R randomize.R
#' @include read.R assign_peaks.R peaks_per_gene.R
plot_gene_coverage = function(peaks, locusdef = "nearest_tss", genome = supported_genomes(), mappability = NULL, legend = TRUE, xlim = NULL) {
	genome = match.arg(genome)

	ldef_list = setup_locusdef(locusdef, genome)
	ldef = ldef_list[['ldef']]

	mappa = setup_mappa(mappa_code = mappability, genome = genome, ldef_code = locusdef, ldef_obj = ldef)

	if (class(peaks) == "data.frame") {
		peakobj = load_peaks(peaks)
	} else if (class(peaks) == "character") {
		peakobj = read_bed(peaks)
	}
	num_peaks = length(peakobj)

	assigned_peaks = assign_peak_segments(peakobj, ldef)
	ppg = num_peaks_per_gene(assigned_peaks, ldef, mappa)
	ppg = calc_peak_gene_overlap(assigned_peaks, ppg)

	# Make plot.
	plotobj = ..plot_gene_coverage(gpw = ppg, mappability = mappability, num_peaks = num_peaks, xlim = xlim)
	return(plotobj)
}

..plot_gene_coverage = function(gpw, mappability, num_peaks, xlim = NULL) {
	############################################################################
	# Prepare the gpw table
	gpw = gpw[order(gpw$log10_length), ]
	gpw$group = ceiling((1:nrow(gpw)) / 25)

	############################################################################
	# Quantities for the plot

	# Scatterplot stuff
	avg_bins = avg_binned_coverage(gpw)

	############################################################################
	# Plotting parameters
	if (is.null(mappability)) {
		xlab = expression(paste(Log[10], " Locus Length"))
	} else {
		xlab = expression(paste(Log[10], " Mappable Locus Length"))
	}

	xmin_nopad = base::ifelse(is.null(xlim[1]), floor(min(gpw$log10_length)), floor(xlim[1]))
	xmax_nopad = base::ifelse(is.null(xlim[2]), ceiling(max(gpw$log10_length)), ceiling(xlim[2]))

	scales = list(
		x = list(
			axs = 'i',
			at = seq(xmin_nopad, xmax_nopad, 1)
		),
		y = list(
			axs = 'i',
			at = seq(0, 1, 0.2)
		)
	)

	############################################################################
	# Plot
	plotobj = lattice::xyplot(
		ratio ~ log_avg_length,
		avg_bins,
		main = 'Binned Locus Length versus Peak Coverage',
		xlab = list(label = xlab, cex = 1.4),
		ylab = list(label = "Proportion of Locus Covered by Peaks", cex = 1.4),
		ylim = c(-0.05, 1.05),
		xlim = c(xmin_nopad - 0.5, xmax_nopad + 0.5),
		pch = 20,
		cex = 0.4,
		col = 'black',
		scales = list(cex = 1.4)
	)

	return(plotobj)
}

#' Get the correct organism code based on genome
#'
#' Data from \code{chipenrich.data} uses three letter organism codes for the
#' \code{GeneSet} objects. This function ensures the correct objects are loaded.
#'
#' @param genome One of the \code{supported_genomes()}.
#'
#' @return A string for the three letter organism code. Convention is first letter
#' of the first word in the binomial name, and first two letters of the second word
#' in the binomial name. 'Homo sapiens' is then 'hsa', for example.
genome_to_organism = function(genome = supported_genomes()) {
    genome = match.arg(genome)

    if (grepl('^mm', genome)) {
        org = 'mmu'
    } else if (grepl('^hg', genome)) {
        org = 'hsa'
    } else if (grepl('^rn', genome)) {
        org = 'rno'
    } else if (grepl('^dm', genome)) {
        org = 'dme'
    } else if (grepl('^danRer', genome)) {
        org = 'dre'
    }

    return(org)
}

#' Get Entrez ID to gene symbol mappings for custom locus definitions
#'
#' If a custom locus definition is one of the \code{supported_genomes()}, then
#' the gene symbol column of the custom locus definition is populated using the
#' appropriate orgDb package.
#'
#' @param genome One of the \code{supported_genomes()}.
#'
#' @return A \code{data.frame} with \code{gene_id} and \code{symbol} columns.
genome_to_orgdb = function(genome = supported_genomes()) {
    genome = match.arg(genome)

    if(genome %in% c('hg19','hg38')) {
        egSYMBOL = org.Hs.eg.db::org.Hs.egSYMBOL
    } else if (genome %in% c('mm9','mm10')) {
        egSYMBOL = org.Mm.eg.db::org.Mm.egSYMBOL
    } else if (genome %in% c('rn4','rn5','rn6')) {
        egSYMBOL = org.Rn.eg.db::org.Rn.egSYMBOL
    } else if (genome %in% c('dm3','dm6')) {
        egSYMBOL = org.Dm.eg.db::org.Dm.egSYMBOL
    } else if (genome == 'danRer10') {
        egSYMBOL = org.Dr.eg.db::org.Dr.egSYMBOL
    }

    ### Build Entrez ID to gene symbol mapping
        mapped_genes = AnnotationDbi::mappedkeys(egSYMBOL)
        eg2symbol = as.data.frame(egSYMBOL[mapped_genes])
        eg2symbol$gene_id = as.integer(eg2symbol$gene_id)

    return(eg2symbol)
}

#' Get the test function name from the method name
#'
#' The \code{method} comes from what is used in \code{chipenrich()} or in
#' \code{polyenrich()}.
#'
#' @param method A character for the method used. One of the \code{supported_methods}
#' or one of the \code{HIDDEN_METHODS} in \code{constants.R}.
#'
#' @return A singleton named character vector with value of the test function and
#' name of the method.
get_test_method = function(method) {
    if (method %in% names(SUPPORTED_METHODS)) {
        return(SUPPORTED_METHODS[[method]])
    } else if (method %in% names(HIDDEN_METHODS)) {
        return(HIDDEN_METHODS[[method]])
    } else {
        stop(sprintf("Error: invalid enrichment test requested: %s",method))
    }
}

#' Post process the \code{data.frame} of enrichment results
#'
#' @param enrich A \code{data.frame} of the enrichment results from \code{broadenrich()},
#' \code{chipenrich()}, or \code{polyenrich()} created by \code{rbind}ing the
#' list of enrichment results for each of the \code{genesets}.
#'
#' @return A reformatted \code{data.frame} with columns in a specific order, filtered
#' of enrichment tests that failed, and ordered first by enrichment 'Status' (if
#' present) and then 'P.value'.
post_process_enrichments = function(enrich) {
    # Re-order the columns to something sensible.
    column_order = c(
        "Geneset.Type",
        "Geneset.ID",
        "Description",
        "P.value",
        "FDR",
        "Effect",
        "Odds.Ratio",
        "P.Success",
        "Status",
        "N.Geneset.Genes",
        "N.Geneset.Peak.Genes",
        "Geneset.Avg.Gene.Length",
        "Geneset.Avg.Gene.Coverage",
        "Geneset.Peak.Genes")
    column_order = intersect(column_order, names(enrich))
    enrich = enrich[, column_order]

    # Order results by p-value.
    enrich = enrich[order(enrich$P.value), ]

    # If there is a status column, re-sort so enriched terms are on top.
    if ("Status" %in% names(enrich)) {
        enrich = enrich[order(enrich$Status, decreasing=TRUE), ]
    }

    # Pull out tests that failed.
    bad_enrich = subset(enrich, is.na(enrich$P.value))
    enrich = subset(enrich, !is.na(enrich$P.value))
    rownames(enrich) = c(1:nrow(enrich))

    return(enrich)
}

#' Recode a vector of number of peaks to binary based on threshold
#'
#' @param num_peaks An \code{integer} vector representing numbers of peaks per gene.
#' @param threshold An \code{integer} specifying the minimum number of peaks
#' required to code as 1.
#'
#' @return An binary vector where an entry is 1 if the corresponding entry of
#' \code{num_peaks} is >= \code{threshold} and is otherwise 0.
recode_peaks = function(num_peaks, threshold = 1) {
  as.numeric(num_peaks >= threshold);
}

#' Reset n_cores for Windows
#'
#' We use \code{parallel::mclapply} for multicore geneset enrichment testing, but
#' this function supports more than one core if the OS is not Windows. If the OS
#' is windows, the number of cores (\code{mc.cores}) must be set to 1.
#'
#' @param n_cores An \code{integer} passed to \code{broadenrich()},
#' \code{chipenrich()}, or \code{polyenrich()} indicating the number of cores
#' to use for enrichment testing.
#'
#' @return Either the original \code{n_cores} if the OS is not Windows, or 1 if
#' the OS is Windows.
reset_ncores_for_windows = function(n_cores) {
    os = Sys.info()[1]
	if(os == 'Windows') {
		message('Setting n_cores = 1 because Windows detected as OS.')
		n_cores = 1
	}

    return(n_cores)
}

context('Test utility functions')

test_that('Recode peaks', {
    num_peaks = c(1,4,7,1,8,0,5)

    expect_true(all(recode_peaks(num_peaks) == c(1,1,1,1,1,0,1)))
    expect_true(all(recode_peaks(num_peaks, threshold = 2) == c(0,1,1,0,1,0,1)))
})

test_that('Genome to organism', {
    expect_error(genome_to_organism('test'), "'arg' should be one of")
    expect_equal(genome_to_organism('dm3'), 'dme')
    expect_equal(genome_to_organism('dm6'), 'dme')
    expect_equal(genome_to_organism('danRer10'), 'dre')
    expect_equal(genome_to_organism('hg19'), 'hsa')
    expect_equal(genome_to_organism('hg38'), 'hsa')
    expect_equal(genome_to_organism('mm9'), 'mmu')
    expect_equal(genome_to_organism('mm10'), 'mmu')
    expect_equal(genome_to_organism('rn4'), 'rno')
    expect_equal(genome_to_organism('rn5'), 'rno')
    expect_equal(genome_to_organism('rn6'), 'rno')
})

test_that('Genome to eg2symbol', {
    expect_error(genome_to_orgdb('test'), "'arg' should be one of")
    expect_equal(class(genome_to_orgdb('dm6')), 'data.frame')
    expect_equal(class(genome_to_orgdb('danRer10')), 'data.frame')
    expect_equal(class(genome_to_orgdb('hg19')), 'data.frame')
    expect_equal(class(genome_to_orgdb('mm10')), 'data.frame')
    expect_equal(class(genome_to_orgdb('rn6')), 'data.frame')
})

#' Function to filter genesets by locus definition and size
#'
#' This function filters gene sets based on the genes that are present in a particular
#' locus definition. After determining which genes are present in both the GeneSet,
#' \code{gs_obj}, and the LocusDefinition \code{ldef_obj}, gene sets are filtered
#' by size with \code{min_geneset_size} and \code{max_geneset_size}.
#'
#' @param gs_obj A valid GeneSet object
#' @param ldef_obj A valid LocusDefinition object
#' @param min_geneset_size An integer indicating the floor for genes in a geneset. Default 15.
#' @param max_geneset_size An integer indicating the ceiling for genes in a geneset. Default 2000.
#'
#' @return An altered \code{gs_obj} with changed \code{set.gene} and \code{all.genes}
#' slots reflecting \code{min_geneset_size} and \code{max_geneset_size} after intersecting
#' with the genes present in the particular locus definition.
filter_genesets = function(gs_obj, ldef_obj, min_geneset_size = 15, max_geneset_size = 2000) {
	if (class(gs_obj) != "GeneSet") {
		stop("Error: gs_obj not of class GeneSet")
	}

	if(class(ldef_obj) != 'LocusDefinition') {
		stop("Error: ldef_obj not of class LocusDefinition")
	}

	# Coerce to character, because gene sets store genes as character(), and
	# intersect() is much faster when the vectors are of the same type.
	ldef_gene_ids = as.character(unique(ldef_obj@granges$gene_id))
	tmp_set.gene = as.list(gs_obj@set.gene)

	tmp_set.gene = lapply(tmp_set.gene, function(gs){intersect(ldef_gene_ids, gs)})
	tmp_set.gene = Filter(function(gs) length(gs) >= min_geneset_size, tmp_set.gene)
	tmp_set.gene = Filter(function(gs) length(gs) <= max_geneset_size, tmp_set.gene)

	gs_obj@set.gene = as.environment(tmp_set.gene)
	gs_obj@all.genes = unique(as.character(unlist(tmp_set.gene, use.names = FALSE)))

	return(gs_obj)
}

#' Function to setup locus definitions
#'
#' @param ldef_code One of 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
#' '1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
#' '10kb', '10kb_outside', '10kb_outside_upstream'. Alternately, a file path for
#' a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
#' must have columns 'chr', 'start', 'end', and 'gene_id', or 'geneid'.
#' @param genome One of the \code{supported_genomes()}.
#' @param randomization One of \code{NULL}, 'complete', 'bylength', or 'bylocation'.
#' See the Randomizations section in \code{?chipenrich}. Default NULL.
#'
#' @return A list with components \code{ldef} and \code{tss}.
#' @include randomize.R
setup_locusdef = function(ldef_code, genome, randomization = NULL) {
	user_defined_ldef = file.exists(ldef_code)

	if (user_defined_ldef) {
		ldef = read_ldef(ldef_code, genome)
	} else {
		if (!any(supported_locusdefs()$genome == genome & ldef_code == supported_locusdefs()$locusdef)) {
			stop(sprintf("Error: invalid genome / definition combination requested: %s %s", genome, ldef_code))
		}

		ldef_rdata_code = sprintf("locusdef.%s.%s", genome, ldef_code)
		data(list = ldef_rdata_code, package = "chipenrich.data", envir = environment())
		ldef = get(ldef_rdata_code)
	}

	# Load TSS site info.
	tss_code = sprintf("tss.%s", genome)
	data(list=tss_code, package = "chipenrich.data", envir = environment())
	tss = get(tss_code)

	# Deal with randomizations
	if(is.null(randomization)) {

	} else if (randomization == 'complete') {
		ldef = randomize_ldef_complete(ldef)
	} else if (randomization == 'bylength') {
		ldef = randomize_ldef_bylength(ldef, resolution = 100)
	} else if (randomization == 'bylocation') {
		ldef = randomize_ldef_bylocation(ldef, resolution = 50)
	} else {
		stop("Invalid randomization. Must be one of c(NULL, 'complete', 'bylength', 'bylocation')")
	}

	return(list(
		ldef = ldef,
		tss = tss))
}

#' Function to setup genesets
#'
#' @param gs_codes A character vector of geneset databases to be tested for
#' enrichment. See \code{supported_genesets()}. Alternately, a file path to a
#' a tab-delimited text file with header and first column being the geneset ID
#' or name, and the second column being Entrez Gene IDs.
#' @param ldef_obj A \code{LocusDefinition} object to use for filtering gene sets
#' based on which genes are defined in the locus defintion.
#' @param genome One of the \code{supported_genomes()}.
#' @param min_geneset_size Sets the minimum number of genes a gene set may have
#' to be considered for enrichment testing.
#' @param max_geneset_size Sets the maximum number of genes a gene set may have
#' to be considered for enrichment testing.
#'
#' @return A list with components consisting of \code{GeneSet} objects for each
#' of the elements of \code{genesets}. NOTE: Custom genesets must be run separately
#' from built in gene sets.
setup_genesets = function(gs_codes, ldef_obj, genome, min_geneset_size, max_geneset_size) {
	if(class(ldef_obj) != 'LocusDefinition') {
		stop('ldef_obj not of class LocusDefinition.')
	}

	organism = genome_to_organism(genome)

	user_defined_geneset = file.exists(gs_codes)

	if(user_defined_geneset) {
		geneset_list = list()
		geneset_code = 'user-supplied'

		geneset_list[[geneset_code]] = read_geneset(gs_codes)
		geneset_list[[geneset_code]] = filter_genesets(geneset_list[[geneset_code]], ldef_obj, min_geneset_size, max_geneset_size)
	} else {
		if (!any(supported_genesets()$organism == organism & gs_codes %in% supported_genesets()$geneset)) {
			stop("Invalid organism / geneset combination requested. Please see supported_genesets().")
		}

		# If the user does not provide a path to the geneset.
		# That is, do the normal thing
		geneset_list = list()
		for (gs_code in gs_codes) {
			gs_rdata_code = sprintf("geneset.%s.%s", gs_code, organism)
			data(list = gs_rdata_code, package = "chipenrich.data", envir = environment())

			geneset_list[[gs_rdata_code]] = filter_genesets(get(gs_rdata_code), ldef_obj, min_geneset_size, max_geneset_size)
		}
	}

	return(geneset_list)
}

#' Function to setup mappability
#'
#' @param mappa_code One of \code{NULL}, a file path to a custom mappability file,
#' or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
#' If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
#' Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
#' Default value is NULL.
#' @param genome One of the \code{supported_genomes()}.
#' @param ldef_code One of 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
#' '1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
#' '10kb', '10kb_outside', '10kb_outside_upstream'. Alternately, a file path for
#' a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
#' must have columns 'chr', 'start', 'end', and 'gene_id', or 'geneid'.
#' @param ldef_obj A \code{LocusDefinition} object.
#'
#' @return A \code{data.frame} with columns \code{gene_id} and \code{mappa}.
setup_mappa = function(mappa_code, genome, ldef_code, ldef_obj) {
	if(is.null(mappa_code)) {
		return(NULL)
	}

	if(class(ldef_obj) != 'LocusDefinition') {
		stop('ldef_obj not of class LocusDefinition.')
	}

	# mappa_code		ldef_code		action
	# NULL				file			return(NULL)
	# NULL				built-in		return(NULL)
	# file				file			check overlap
	# file				built-in		check overlap
	# read_length		file			stop
	# read_length		built-in		check valid mappa_code

	user_defined_mappa = file.exists(as.character(mappa_code))
	user_defined_ldef = file.exists(ldef_code)

	if(user_defined_mappa) {
		message("Reading user-specified gene locus mappability file: ", mappa_code)
		mappa = read_mappa(mappa_code)

		# Check overlap of user defined mappability gene_id with locus definition gene_id
		total_unique_genes = union(mappa$gene_id, ldef_obj@dframe$gene_id)
		mappa_ldef_inters = intersect(mappa$gene_id, ldef_obj@dframe$gene_id)
		frac_overlap = length(mappa_ldef_inters) / length(total_unique_genes)

		if (frac_overlap < 0.95) {
			stop("Error: your mappability genes and locus definition genes overlap by less than 95% (they should match almost exactly)..")
		}
	} else {
		if(user_defined_ldef) {
			stop("Built-in mappability cannot be used with a user-defined locus definition. Please calculate mappability for your definition and pass its file path to the 'mappability' parameter.")
		} else {
			if (!any(supported_read_lengths()$genome == genome & supported_read_lengths()$locusdef == ldef_code & supported_read_lengths()$read_length == as.numeric(mappa_code))) {
				stop(sprintf("Error: bad genome / locusdef / read length combination requested: %s %s %s", genome, ldef_code, mappa_code))
			}

			mappa_rdata_code = sprintf("mappa.%s.%s.%smer", genome, ldef_code, mappa_code)
			data(list = mappa_rdata_code, package = "chipenrich.data", envir = environment())
			mappa = get(mappa_rdata_code)
			mappa = na.omit(mappa)
			colnames(mappa) = c('gene_id', 'mappa')
		}
	}

	return(mappa)
}

context('Test setup functions')

################################################################################
# Filter genesets

test_that('Filter genesets errors', {
	data(locusdef.hg19.nearest_tss, package='chipenrich.data')
	expect_error(filter_genesets('test', locusdef.hg19.nearest_tss, 20, 100), 'gs_obj not of class GeneSet')
})

test_that('Filter locus def errors', {
	data(geneset.GOBP.hsa, package='chipenrich.data')
	expect_error(filter_genesets(geneset.GOBP.hsa, 'test', 20, 100), 'ldef_obj not of class LocusDefinition')
})

test_that('Filter genesets works for nearest_tss', {
	data(locusdef.hg19.nearest_tss, package='chipenrich.data')
	data(geneset.GOBP.hsa, package='chipenrich.data')
	filtered = filter_genesets(geneset.GOBP.hsa, locusdef.hg19.nearest_tss, 15, 100)

	expect_equal(min(sapply(as.list(filtered@set.gene), length)),15)
	expect_equal(max(sapply(as.list(filtered@set.gene), length)), 100)
})

test_that('Filter genesets works for 5kb', {
	data(locusdef.hg19.5kb, package='chipenrich.data')
	data(geneset.GOBP.hsa, package='chipenrich.data')
	filtered = filter_genesets(geneset.GOBP.hsa, locusdef.hg19.5kb, 20, 100)

	expect_equal(min(sapply(as.list(filtered@set.gene), length)), 20)
	expect_equal(max(sapply(as.list(filtered@set.gene), length)), 100)
})

test_that('Errors in setup_locusdef()', {
	expect_error(setup_locusdef('blue', genome = 'hg19'), 'invalid genome / definition combination')
	expect_error(setup_locusdef('nearest_tss', genome = 'hg19', randomization = 'blue'), 'Invalid randomization')
})

test_that('Errors in setup_genesets()', {
	expect_error(setup_genesets(gs_codes = 'GOBP', ldef_obj = 'blue', genome = 'hg19', min_geneset_size = 15, max_geneset_size = 2000), 'ldef_obj not of class LocusDefinition')

	ldef = setup_locusdef('nearest_tss', genome = 'hg19')$ldef
	expect_error(setup_genesets(gs_codes = 'blue', ldef_obj = ldef, genome = 'hg19', min_geneset_size = 15, max_geneset_size = 2000), 'Invalid organism / geneset combination requested')
})

test_that('Errors in setup_mappa()', {
	mappa_file = system.file('extdata', 'test_mappa_good.txt', package = 'chipenrich')
	ldef_file = system.file('extdata', 'test_ldef_symbol.txt', package = 'chipenrich')

	ldef = setup_locusdef('nearest_tss', genome = 'hg19')$ldef
	custom_ldef = setup_locusdef(ldef_file, genome = 'hg19')$ldef

	expect_error(setup_mappa(mappa_code = 24, genome = 'hg19', ldef_code = 'blue', ldef_obj = 'blue'), 'ldef_obj not of class LocusDefinition')
	expect_error(setup_mappa(mappa_code = mappa_file, genome = 'hg19', ldef_code = 'nearest_tss', ldef_obj = ldef), 'your mappability genes and locus definition genes overlap')
	expect_error(setup_mappa(mappa_code = 24, genome = 'hg19', ldef_code = ldef_file, ldef_obj = custom_ldef), 'Built-in mappability cannot be used with a user-defined locus definition')
})

test_that('setup_mappa() works with two customs', {
	mappa_file = system.file('extdata', 'test_mappa_good.txt', package = 'chipenrich')
	ldef_file = system.file('extdata', 'test_ldef_symbol.txt', package = 'chipenrich')

	custom_ldef = setup_locusdef(ldef_file, genome = 'hg19')$ldef

	expect_true(all(setup_mappa(mappa_code = mappa_file, genome = 'hg19', ldef_code = ldef_file, ldef_obj = custom_ldef)$mappa == c(0.8,0.1,0.008)))
})

test_that('setup_mappa() works with built-ins', {
	ldef = setup_locusdef('nearest_tss', genome = 'hg19')$ldef

	expect_equal(nrow(setup_mappa(mappa_code = 24, genome = 'hg19', ldef_code = 'nearest_tss', ldef_obj = ldef)), 19051)
})

test_polyapprox = function(geneset, gpw, n_cores) {
	# Restrict our genes/weights/peaks to only those genes in the genesets.
	# Here, geneset is not all combined, but GOBP, GOCC, etc.
	# i.e. A specific one.
	gpw = subset(gpw, gpw$gene_id %in% geneset@all.genes)
	
    fitspl = mgcv::gam(num_peaks~s(log10_length,bs='cr'),data=gpw,family="nb")
    gpw$spline = as.numeric(predict(fitspl, gpw, type="terms"))
    
    #Need to use glm.nb to be able to use glm.scoretest later
    nullfit = MASS::glm.nb(num_peaks~spline,data=gpw)

	# Model formula not needed
	#model = "num_peaks ~ goterm + spline"
	
	# Run tests. NOTE: If os == 'Windows', n_cores is reset to 1 for this to work
	results_list = parallel::mclapply(as.list(ls(geneset@set.gene)), function(go_id) {
		single_polyapprox(go_id, geneset, gpw, 'polyenrich', nullmodel = nullfit)
	}, mc.cores = n_cores)
	
	# Collapse results into one table
	results = Reduce(rbind,results_list)
	
	# Correct for multiple testing
	results$FDR = p.adjust(results$P.value, method="BH")
	
	# Create enriched/depleted status column
	results$Status = ifelse(results$Effect > 0, 'enriched', 'depleted')
	
	results = results[order(results$P.value),]
	
	return(results)
}

glm.scoretest <- function(fit, x2, dispersion=NULL) #Directly lifted from statmod package,
#    because asking for more dependencies is asking for more chances for errors.
#    Score test for new covariate in glm
#    Gordon Smyth
#    27 March 2009. Last modified 20 Mar 2010.
{
    w <- fit$weights
    r <- fit$residuals
    if(any(w <= 0)) {
        r <- r[w>0]
        x2 <- x2[w>0]
        w <- w[w>0]
    }
    if (is.null(dispersion)) {
        fixed.dispersion <- (fit$family$family %in% c("poisson","binomial"))
        if(fixed.dispersion)
        dispersion <- 1
        else if(fit$df.residual > 0) {
            dispersion <- sum(w*r^2)/fit$df.residual
        } else {
            stop("No residual df available to estimate dispersion")
        }
    }
    ws <- sqrt(w)
    x2.1w <- qr.resid(fit$qr,ws*x2)
    zw <- ws*r
    colSums(as.matrix(x2.1w*zw))/sqrt(colSums(as.matrix(x2.1w * x2.1w)))/sqrt(dispersion)
}

single_polyapprox = function(go_id, geneset, gpw, method, nullmodel) {
	# Genes in the geneset
	go_genes = geneset@set.gene[[go_id]]
	
	# Background genes and the background presence of a peak
	b_genes = gpw$gene_id %in% go_genes
	sg_go = gpw$peak[b_genes]
	
	# Information about the geneset
	r_go_id = go_id
	r_go_genes_num = length(go_genes)
	r_go_genes_avg_length = mean(gpw$length[b_genes])
	
	# Information about peak genes
	go_genes_peak = gpw$gene_id[b_genes][sg_go==1]
	r_go_genes_peak = paste(go_genes_peak,collapse=", ")
	r_go_genes_peak_num = length(go_genes_peak)
	
    r_effect = NA
    r_pval = NA
    
    tryCatch(
    {r_effect = glm.scoretest(nullmodel, as.numeric(b_genes));
        r_pval = 2*stats::pnorm(abs(r_effect),lower.tail = F)
    },
    error = {function(e) {warning(
        sprintf("Error in geneset: %s. NAs given", go_id))
    }}
    )
	
	
	out = data.frame(
		"P.value"=r_pval,
		"Geneset ID"=r_go_id,
		"N Geneset Genes"=r_go_genes_num,
		"Geneset Peak Genes"=r_go_genes_peak,
		"N Geneset Peak Genes"=r_go_genes_peak_num,
		"Effect"=r_effect,
		"Odds.Ratio"=exp(r_effect),
		"Geneset Avg Gene Length"=r_go_genes_avg_length,
		stringsAsFactors = FALSE)
	
	return(out)
}

#' Display supported locus definitions
#'
#' The locus definitions are defined as below. For advice on selecting a locus
#' definition, see the 'Selecting A Locus Definition' section below.
#' \describe{
#'   \item{nearest_tss:}{The locus is the region spanning the midpoints between the TSSs of adjacent genes.}
#'   \item{nearest_gene:}{The locus is the region spanning the midpoints between the boundaries of each gene, where a gene is defined as the region between the furthest upstream TSS and furthest downstream TES for that gene. If two gene loci overlap each other, we take the midpoint of the overlap as the boundary between the two loci. When a gene locus is completely nested within another, we create a disjoint set of 3 intervals, where the outermost gene is separated into 2 intervals broken apart at the endpoints of the nested gene.}
#'   \item{1kb:}{The locus is the region within 1kb of any of the TSSs belonging to a gene. If TSSs from two adjacent genes are within 2 kb of each other, we use the midpoint between the two TSSs as the boundary for the locus for each gene.}
#'   \item{1kb_outside_upstream:}{The locus is the region more than 1kb upstream from a TSS to the midpoint between the adjacent TSS.}
#'   \item{1kb_outside:}{The locus is the region more than 1kb upstream or downstream from a TSS to the midpoint between the adjacent TSS.}
#'   \item{5kb:}{The locus is the region within 5kb of any of the TSSs belonging to a gene. If TSSs from two adjacent genes are within 10 kb of each other, we use the midpoint between the two TSSs as the boundary for the locus for each gene.}
#'   \item{5kb_outside_upstream:}{The locus is the region more than 5kb upstream from a TSS to the midpoint between the adjacent TSS.}
#'   \item{5kb_outside:}{The locus is the region more than 5kb upstream or downstream from a TSS to the midpoint between the adjacent TSS.}
#'   \item{10kb:}{The locus is the region within 10kb of any of the TSSs belonging to a gene. If TSSs from two adjacent genes are within 20 kb of each other, we use the midpoint between the two TSSs as the boundary for the locus for each gene.}
#'   \item{10kb_outside_upstream:}{The locus is the region more than 10kb upstream from a TSS to the midpoint between the adjacent TSS.}
#'   \item{10kb_outside:}{The locus is the region more than 10kb upstream or downstream from a TSS to the midpoint between the adjacent TSS.}
#'   \item{exon:}{Each gene has multiple loci corresponding to its exons. Overlaps between different genes are allowed.}
#'   \item{intron:}{Each gene has multiple loci corresponding to its introns. Overlaps between different genes are allowed.}
#' }
#'
#' @section Selecting A Locus Definition:
#' For a transcription factor ChIP-seq experiment, selecting a particular locus
#' definition for use in enrichment testing can have implications relating to how
#' the TF regulates genes. For example, selecting the '1kb' locus definition will
#' imply that the biological processes found enriched are a result of TF regulation
#' near the promoter. In contrast, selecting the '5kb_outside' locus definition
#' will imply that the biological processes found enriched are a result of TF
#' regulation distal from the promoter.
#'
#' Selecting a locus definition can also help reduce the noise in the enrichment
#' tests. For example, if a TF is known to primarily regulate genes by binding
#' around the promoter, then selecting the '1kb' locus definition can help to
#' reduce the noise from TSS-distal peaks in the enrichment testing.
#'
#' The \code{\link{plot_dist_to_tss}} QC plot displays
#' where genomic regions fall relative to TSSs genome-wide, and can help inform
#' the choice of locus definition. For example, if many peaks fall far from the
#' TSS, the 'nearest_tss' locus definition may be a good choice because it will
#' capture all input genomic regions, whereas the '1kb' locus definition may
#' not capture many of the input genomic regions and adversely affect the
#' enrichment testing.
#'
#' @return A \code{data.frame} with columns \code{genome, locusdef}.
#'
#' @examples
#'
#' supported_locusdefs()
#'
#' @export
supported_locusdefs = function() {
	piqr = data(package = "chipenrich.data")
	data_files = piqr$results[,3]

	ldefs = grep("locusdef", data_files, value = TRUE)
	combos = Reduce(rbind,sapply(ldefs, strsplit, '[.]'))[,c(2,3)]
	df = data.frame(
		'genome' = combos[,1],
		'locusdef' = combos[,2],
		stringsAsFactors = FALSE)
	return(df)
}

#' Display supported read lengths for mappability
#'
#' @return A \code{data.frame} with columns \code{genome, locusdef, read_length}.
#'
#' @examples
#'
#' supported_read_lengths()
#'
#' @export
supported_read_lengths = function() {
	piqr = data(package = "chipenrich.data")
	data_files = piqr$results[,3]

	mappas = grep("mappa", data_files, value = TRUE)
	combos = Reduce(rbind,sapply(mappas, strsplit, '[.]'))[,c(2,3,4)]
	df = data.frame(
		'genome' = combos[,1],
		'locusdef' = combos[,2],
		'read_length' = gsub('mer','',combos[,3]),
		stringsAsFactors = FALSE)
	return(df)
}

#' Display supported genesets for gene set enrichment.
#'
#' @return A \code{data.frame} with columns \code{geneset, organism}.
#'
#' @examples
#'
#' supported_genesets()
#'
#' @export
supported_genesets = function() {
	piqr = data(package = "chipenrich.data")
	data_files = piqr$results[,3]

	geneset_files = grep("geneset", data_files, value = TRUE)
	combos = Reduce(rbind,sapply(geneset_files, strsplit, '[.]'))[,c(2,3)]
	df = data.frame(
		'geneset' = combos[,1],
		'organism' = combos[,2],
		stringsAsFactors = FALSE)
	df = df[order(df$organism), ]

	return(df)
}

#' Display supported genomes.
#'
#' @return A vector indicating supported genomes.
#'
#' @examples
#'
#' supported_genomes()
#'
#' @export
supported_genomes = function() {
	piqr = data(package = "chipenrich.data")
	data_files = piqr$results[,3]

	as.character(stats::na.omit(unique(stringr::str_match(data_files,"(locusdef|mappa)\\.(\\w+)\\.")[,3])))
}

#' Display supported gene set enrichment methods.
#'
#' @return A vector indicating supported methods for gene set enrichment.
#'
#' @examples
#'
#' supported_methods()
#'
#' @export
supported_methods = function() {
	return(SUPPORTED_COMBOS)
}

context('Test supported functions')

test_that('Supported locus definitions works', {
	supported = supported_locusdefs()

	expect_equal(class(supported), 'data.frame')
})

test_that('Supported genesets works', {
	supported = supported_genesets()

	expect_equal(class(supported), 'data.frame')
})

test_that('Supported read lengths works', {
	supported = supported_read_lengths()

	expect_equal(class(supported), 'data.frame')
})

test_that('Supported genomes works', {
	supported = supported_genomes()

	expect_equal(class(supported), 'character')
	expect_equal(supported, c('danRer10', 'dm3', 'dm6', 'hg19', 'hg38', 'mm10', 'mm9', 'rn4', 'rn5', 'rn6'))
})

test_that('Supported methods works', {
	supported = supported_methods()

	expect_equal(class(supported), 'data.frame')
})

single_approx = function(go_id, geneset, gpw) {

	lrm.fast = function(x,y) {
		fit = rms::lrm.fit(x,y)
		vv = diag(fit$var)
		cof = fit$coef
		z = cof/sqrt(vv)
		pval = stats::pchisq(z^2, 1, lower.tail = FALSE)
		c(cof[2],pval[2])
	}

	# Genes in the geneset
	go_genes = geneset@set.gene[[go_id]]

	# Background genes, the background presence of a peak, and the background
	# weight of peaks
	b_genes = gpw$gene_id %in% go_genes
	sg_go = gpw$peak[b_genes]
	wg_go = gpw$weight[b_genes]

	# Information about the geneset
	r_go_id = go_id
	r_go_genes_num = length(go_genes)
	r_go_genes_avg_length = mean(gpw$length[b_genes])

    # Information about peak genes
	go_genes_peak = gpw$gene_id[b_genes][sg_go==1]
	r_go_genes_peak = paste(go_genes_peak, collapse = ", ")
	r_go_genes_peak_num = length(go_genes_peak)

	# Small correction for case where every gene in this geneset has a peak.
	if (all(as.logical(sg_go))) {
		cont_length = stats::quantile(gpw$length,0.0025)
		cont_gene = data.frame(
			gene_id = "continuity_correction",
			length = cont_length,
			log10_length = log10(cont_length),
			num_peaks = 0,
			peak = 0,
			weight = stats::quantile(gpw$weight,0.0025),
			prob_peak = stats::quantile(gpw$prob_peak,0.0025),
			resid.dev = stats::quantile(gpw$resid.dev,0.0025),
			stringsAsFactors = FALSE
		)
		if ("mappa" %in% names(gpw)) {
			cont_gene$mappa = 1
		}
		gpw = rbind(gpw,cont_gene)
		b_genes = c(b_genes,1)

		message(sprintf("Applying correction for geneset %s with %i genes...", go_id, length(go_genes)))
	}

	# The model is still essentially peak ~ goterm + s(log10_length,bs='cr')
	# except we're using the weights that are calculated once...
	# Also, y must be binary, so as.numeric(b_genes) (goterm) must be
	# predicted against the weights...
	testm = cbind(y = as.numeric(b_genes), x = gpw$weight*gpw$peak)
	ep = lrm.fast(testm[,"x"], testm[,"y"])

	# Results from quick regression
	r_effect = ep[1]
	r_pval = ep[2]

	out = data.frame(
		"P.value" = r_pval,
		"Geneset ID" = r_go_id,
		"N Geneset Genes" = r_go_genes_num,
		"Geneset Peak Genes" = r_go_genes_peak,
		"N Geneset Peak Genes" = r_go_genes_peak_num,
		"Effect" = r_effect,
		"Odds.Ratio" = exp(r_effect),
		"Geneset Avg Gene Length" = r_go_genes_avg_length,
		stringsAsFactors = FALSE)

	return(out)
}

test_approx = function(geneset, gpw, nwp = FALSE, n_cores) {

	if (!"weight" %in% names(gpw)) {
    	stop("Error: you must fit weights first using one of the calc_weights* functions.")
	}

	# Restrict our genes/weights/peaks to only those genes in the genesets.
	gpw = subset(gpw, gpw$gene_id %in% geneset@all.genes)

	# Re-normalize weights.
	# Not sure what nwp means, nor if we want to modify gpw like this.
	if (!nwp) {
		gpw$weight = gpw$weight / mean(gpw$weight)
	} else {
		b_haspeak = gpw$peak == 1
		gpw$weight[b_haspeak] = gpw$weight[b_haspeak] / mean(gpw$weight[b_haspeak])
	}

	# Run tests. NOTE: If os == 'Windows', n_cores is reset to 1 for this to work
	results_list = parallel::mclapply(as.list(ls(geneset@set.gene)), function(go_id) {
		single_approx(go_id, geneset, gpw)
	}, mc.cores = n_cores)

	# Collapse results into one table
	results = Reduce(rbind, results_list)

	# Correct for multiple testing
	results$FDR = stats::p.adjust(results$P.value, method="BH")

	# Create enriched/depleted status column
	results$Status = ifelse(results$Effect > 0, 'enriched', 'depleted')

	results = results[order(results$P.value),]

	return(results)
}

# Note: the counts input is to allow future methods to regress on other count data
test_polyenrich_weighted = function(geneset,gpw,n_cores,counts) {
	# Restrict our genes/weights/peaks to only those genes in the genesets.
	# Here, geneset is not all combined, but GOBP, GOCC, etc.
	# i.e. A specific one.
	gpw = subset(gpw,gpw$gene_id %in% geneset@all.genes);
	
	if (sum(gpw$peak) == 0) {
		stop("Error: no peaks in your data!");
	}
	
	if (!(counts %in% names(gpw))) {
		stop(sprintf("Error: %s is not a column in data!", counts))
	}
	
    fitspl = mgcv::gam(as.formula(sprintf("%s~s(log10_length,bs='cr')",counts)),data=gpw,family="nb")
	gpw$spline = as.numeric(predict(fitspl, gpw, type="terms"))
	
	
	# Construct model formula.
	model = sprintf("%s ~ goterm + spline",counts);
	
    # Run tests. NOTE: If os == 'Windows', n_cores is reset to 1 for this to work
    results_list = parallel::mclapply(as.list(ls(geneset@set.gene)), function(go_id) {
        single_polyenrich_weighted(go_id, geneset, gpw, fitspl, 'polyenrich', model)
    }, mc.cores = n_cores)
	
	# Collapse results into one table
	results = Reduce(rbind,results_list)
	
	# Correct for multiple testing
	results$FDR = p.adjust(results$P.value, method="BH");
	
	# Create enriched/depleted status column
	results$Status = ifelse(results$Effect > 0, 'enriched', 'depleted')
	
	results = results[order(results$P.value),];
	
	return(results);
}

single_polyenrich_weighted = function(go_id, geneset, gpw, fitspl, method, model) {
	final_model = as.formula(model);
	
	# Genes in the geneset
	go_genes = geneset@set.gene[[go_id]];
	
	# Filter genes in the geneset to only those in the gpw table.
	# The gpw table will be truncated depending on which geneset type we're in.
	go_genes = go_genes[go_genes %in% gpw$gene_id];
	
	# Background genes and the background presence of a peak
	b_genes = gpw$gene_id %in% go_genes;
	sg_go = gpw$peak[b_genes];
	
	# Information about the geneset
	r_go_id = go_id;
	r_go_genes_num = length(go_genes);
	r_go_genes_avg_length = mean(gpw$length[b_genes]);
	
	# Information about peak genes
	go_genes_peak = gpw$gene_id[b_genes][sg_go==1];
	r_go_genes_peak = paste(go_genes_peak,collapse=", ");
	r_go_genes_peak_num = length(go_genes_peak);
		
	# Logistic regression works no matter the method because final_model is chosen above
	# and the data required from gpw will automatically be correct based on the method used.
	fit = gam(final_model,data=cbind(gpw,goterm=as.numeric(b_genes)),family="nb");
	
	# Results from the logistic regression
	r_effect = coef(fit)[2];
	r_pval = summary(fit)$p.table[2,4];
	
	# The only difference between chipenrich and broadenrich here is
	# the Geneset Avg Gene Coverage column
	
	out = data.frame(
		"P.value"=r_pval,
		"Geneset ID"=r_go_id,
		"N Geneset Genes"=r_go_genes_num,
		"Geneset Peak Genes"=r_go_genes_peak,
		"N Geneset Peak Genes"=r_go_genes_peak_num,
		"Effect"=r_effect,
		"Odds.Ratio"=exp(r_effect),
		"Geneset Avg Gene Length"=r_go_genes_avg_length,
		stringsAsFactors=FALSE);
	
	return(out);
}

#' chipenrich: Gene Set Enrichment For ChIP-seq Peak Data and Other Genomic Regions
#'
#' The \code{chipenrich} package includes three classes of methods that adjust for potential confounders of gene set enrichment testing (locus length and mappability of the sequence reads). The first, \code{chipenrich}, is designed for use with transcription-factor (TF) based ChIP-seq experiments and other DNA sequencing experiments with narrow genomic regions. The second, \code{polyenrich}, is similarly designed for TF based ChIP-seq, but where the number of peaks present in gene loci may be important. The third, \code{broadenrich}, is designed for use with histone modification based ChIP-seq experiments and other DNA sequencing experiments with broad genomic regions.
#'
#' @docType package
#' @name chipenrich_package
#'
#' @importFrom AnnotationDbi mappedkeys
#' @importFrom BiocGenerics unlist
#' @import chipenrich.data
#' @import GenomicRanges
#' @importFrom Seqinfo seqinfo
#' @importFrom grDevices dev.off pdf
#' @import grid
#' @importFrom IRanges IRanges
#' @import lattice
#' @import latticeExtra
#' @import methods
#' @import mgcv
#' @import org.Dm.eg.db
#' @import org.Dr.eg.db
#' @import org.Hs.eg.db
#' @import org.Mm.eg.db
#' @import org.Rn.eg.db
#' @import parallel
#' @importFrom plyr rbind.fill
#' @import rms
#' @import rtracklayer
#' @importFrom S4Vectors queryHits subjectHits splitAsList
#' @importFrom stats as.formula binom.test coef complete.cases fisher.test fitted na.omit p.adjust pchisq predict qpois quantile resid
#' @importFrom stringr str_match str_sub
#' @importFrom utils data read.table tail write.table
NULL

SUPPORTED_METHODS = list(
    'broadenrich' = "test_broadenrich",
    'chipenrich' = "test_chipenrich",
    'fet' = "test_fisher_exact",
    'polyenrich' = "test_polyenrich"
)

HIDDEN_METHODS = list(
    'binomial' = "test_binomial",
    'broadenrich_splineless' = "test_broadenrich_splineless",
    'chipenrich_slow' = 'test_chipenrich_slow',
    'chipapprox_old' = 'test_approx',
    'polyenrich_slow' = 'test_polyenrich_slow',
    'polyenrich_weighted' = 'test_polyenrich_weighted',
    'chipapprox' = 'test_chipapprox',
    'polyapprox' = 'test_polyapprox'
)

METHOD_NAMES = list(
    'broadenrich' = "Broad-Enrich",
    'broadenrich_splineless' = "Broad-Enrich Splineless",
    'chipenrich' = "ChIP-Enrich",
    'chipenrich_slow' = "ChIP-Enrich Original",
	'chipapprox_old' = "ChIP-Enrich Approximate (Old)",
    'chipapprox' = "ChIP-Enrich Approximate",
    'fet' = "Fisher's Exact Test",
    'polyenrich' = "Poly-Enrich",
    'polyenrich_slow' = "Poly-Enrich Original",
    'polyenrich_weighted' = "Poly-Enrich Weighted",
    'polyapprox' = "Poly-Enrich Approximate"

)

SUPPORTED_COMBOS = data.frame(
    'function' = c("broadenrich","chipenrich","chipenrich","polyenrich","polyenrich","hybridenrich"),
    'method' = c(NA,"chipenrich","fet","polyenrich","polyenrich_weighted",NA)
)

#' Run ChIP-Enrich on narrow genomic regions
#'
#' ChIP-Enrich is designed for use with 1,000s or 10,000s of narrow
#' peaks which results in fewer gene loci containing a peak overall. For example,
#' ChIP-seq experiments for transcription factors. For more details, see the 'ChIP-Enrich
#' Method' section below. For help choosing a method, see the 'Choosing A Method'
#' section below, or see the vignette.
#'
#' @section ChIP-Enrich Method:
#' The ChIP-Enrich method uses the presence of a peak in its model for enrichment:
#' \code{peak ~ GO + s(log10_length)}. Here, \code{GO} is a binary vector indicating
#' whether a gene is in the gene set being tested, \code{peak} is a binary vector
#' indicating the presence of a peak in a gene, and \code{s(log10_length)} is a
#' binomial cubic smoothing spline which adjusts for the relationship between the
#' presence of a peak and locus length.
#'
#' @section Choosing A Method:
#' The following guidelines are intended to help select an enrichment function:
#' \describe{
#'	\item{broadenrich():}{ is designed for use with broad peaks that may intersect
#' multiple gene loci, and cumulatively cover greater than 5\% of the genome. For
#' example, ChIP-seq experiments for histone modifications.}
#'	\item{chipenrich():}{ is designed for use with 1,000s or 10,000s of narrow
#' peaks which results in fewer gene loci containing a peak overall. For example,
#' ChIP-seq experiments for transcription factors.}
#'	\item{polyenrich():}{ is also designed for narrow peaks, for experiments with
#' 100,000s of peaks, or in cases where the number of binding sites per gene affects
#' its regulation. If unsure whether to use chipenrich or polyenrich, then we recommend
#' hybridenrich.}
#'  \item{hybridenrich():}{ is a combination of chipenrich and polyenrich, to be
#' used when one is unsure which is the optimal method. }
#' }
#'
#' @section Randomizations:
#' Randomization of locus definitions allows for the assessment of Type I Error
#' under the null hypothesis. The randomization codes are:
#' \describe{
#'	\item{\code{NULL}:}{ No randomizations, the default.}
#' 	\item{'complete':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
#' \code{locusdef} together, without regard for the chromosome location, or locus length.
#' The null hypothesis is that there is no true gene set enrichment.}
#' 	\item{'bylength':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
#' \code{locusdef} together within bins of 100 genes sorted by locus length. The null
#' hypothesis is that there is no true gene set enrichment, but with preserved locus
#' length relationship.}
#' 	\item{'bylocation':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
#' \code{locusdef} together within bins of 50 genes sorted by genomic location. The null
#' hypothesis is that there is no true gene set enrichment, but with preserved
#' genomic location.}
#' }
#' The return value with a selected randomization is the same list as without.
#' To assess the Type I error, the \code{alpha} level for the particular data set
#' can be calculated by dividing the total number of gene sets with p-value < \code{alpha}
#' by the total number of tests. Users may want to perform multiple randomizations
#' for a set of peaks and take the median of the \code{alpha} values.
#'
#' @param peaks Either a file path or a \code{data.frame} of peaks in BED-like
#' format. If a file path, the following formats are fully supported via their
#' file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
#' or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
#' without these extensions are supported under the conditions that the first 3
#' columns correspond to 'chr', 'start', and 'end' and that there is either no
#' header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
#' \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
#' acceptable column names.
#' @param out_name Prefix string to use for naming output files. This should not
#' contain any characters that would be illegal for the system being used (Unix,
#' Windows, etc.) The default value is "chipenrich", and a file "chipenrich_results.tab"
#' is produced. If \code{qc_plots} is set, then a file "chipenrich_qcplots.png"
#' is produced containing a number of quality control plots. If \code{out_name}
#' is set to NULL, no files are written, and results then must be retrieved from
#' the list returned by \code{chipenrich}.
#' @param out_path Directory to which results files will be written out. Defaults
#' to the current working directory as returned by \code{\link{getwd}}.
#' @param genome One of the \code{supported_genomes()}.
#' @param genesets A character vector of geneset databases to be tested for
#' enrichment. See \code{supported_genesets()}. Alternately, a file path to a
#' a tab-delimited text file with header and first column being the geneset ID
#' or name, and the second column being Entrez Gene IDs. For an example custom
#' gene set file, see the vignette.
#' @param locusdef One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
#' '1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
#' '10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
#' see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
#' a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
#' must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
#' example custom locus definition file, see the vignette.
#' @param method A character string specifying the method to use for enrichment
#' testing. Must be one of ChIP-Enrich ('chipenrich') (default), or
#'Fisher's exact test ('fet').
#' @param mappability One of \code{NULL}, a file path to a custom mappability file,
#' or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
#' If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
#' Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
#' For an example custom mappability file, see the vignette. Default value is NULL.
#' @param fisher_alt If method is 'fet', this option indicates the alternative
#' for Fisher's exact test, and must be one of 'two-sided' (default), 'greater',
#' or 'less'.
#' @param qc_plots A logical variable that enables the automatic generation of
#' plots for quality control.
#' @param min_geneset_size Sets the minimum number of genes a gene set may have
#' to be considered for enrichment testing.
#' @param max_geneset_size Sets the maximum number of genes a gene set may have
#' to be considered for enrichment testing.
#' @param num_peak_threshold Sets the threshold for how many peaks a gene must
#' have to be considered as having a peak. Defaults to 1. Only relevant for
#' Fisher's exact test and ChIP-Enrich methods.
#' @param randomization One of \code{NULL}, 'complete', 'bylength', or 'bylocation'.
#' See the Randomizations section below.
#' @param n_cores The number of cores to use for enrichment testing. We recommend
#' using only up to the maximum number of \emph{physical} cores present, as
#' virtual cores do not significantly decrease runtime. Default number of cores
#' is set to 1. NOTE: Windows does not support multicore enrichment.
#'
#' @return A list, containing the following items:
#'
#' \item{opts }{A data frame containing the arguments/values passed to \code{chipenrich}.}
#'
#' \item{peaks }{
#' A data frame containing peak assignments to genes. Peaks which do not overlap
#' a gene locus are not included. Each peak that was assigned to a gene is listed,
#' along with the peak midpoint or peak interval coordinates (depending on which
#' was used), the gene to which the peak was assigned, the locus start and end
#' position of the gene, and the distance from the peak to the TSS.
#'
#' The columns are:
#'
#' \describe{
#'   \item{peak_id}{an ID given to unique combinations of chromosome, peak start, and peak end. }
#'   \item{chr}{the chromosome the peak originated from. }
#'   \item{peak_start}{start position of the peak. }
#'   \item{peak_end}{end position of the peak. }
#'   \item{peak_midpoint}{the midpoint of the peak. }
#'   \item{gene_id}{the Entrez ID of the gene to which the peak was assigned. }
#'   \item{gene_symbol}{the official gene symbol for the gene_id (above). }
#'   \item{gene_locus_start}{the start position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
#'   \item{gene_locus_end}{the end position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
#'   \item{nearest_tss}{the closest TSS to this peak (for any gene, not necessarily the gene this peak was assigned to.) }
#'   \item{nearest_tss_gene}{the gene having the closest TSS to the peak (should be the same as gene_id when using the nearest TSS locus definition.) }
#'   \item{nearest_tss_gene_strand}{the strand of the gene with the closest TSS. }
#' }}
#'
#' \item{peaks_per_gene }{
#' A data frame of the count of peaks per gene. The columns are:
#'
#' \describe{
#'   \item{gene_id}{the Entrez Gene ID. }
#'   \item{length}{the length of the gene's locus (depending on which locus definition you chose.)}
#'   \item{log10_length}{the log10(locus length) for the gene.}
#'   \item{num_peaks}{the number of peaks that were assigned to the gene, given the current locus definition. }
#'   \item{peak}{whether or not the gene is considered to have a peak, as defined by \code{num_peak_threshold}. }
#' }}
#'
#' \item{results }{
#' A data frame of the results from performing the gene set enrichment test on
#' each geneset that was requested (all genesets are merged into one final data
#' frame.) The columns are:
#'
#' \describe{
#'   \item{Geneset.ID}{the identifier for a given gene set from the selected database.  For example, GO:0000003. }
#'   \item{Geneset.Type}{ specifies from which database the Geneset.ID originates.  For example, "Gene Ontology Biological Process."}
#'   \item{Description}{ gives a definition of the geneset. For example, "reproduction."}
#'   \item{P.Value}{the probability of observing the degree of enrichment of the gene set given the null hypothesis that peaks are not associated with any gene sets.}
#'   \item{FDR}{the false discovery rate proposed by Bejamini \& Hochberg for adjusting the p-value to control for family-wise error rate.}
#'   \item{Odds.Ratio}{the estimated odds that peaks are associated with a given gene set compared to the odds that peaks are associated with other gene sets, after controlling for locus length and/or mappability.  An odds ratio greater than 1 indicates enrichment, and less than 1 indicates depletion.}
#'   \item{N.Geneset.Genes}{the number of genes in the gene set.}
#'   \item{N.Geneset.Peak.Genes}{the number of genes in the genes set that were assigned at least one peak.}
#'   \item{Geneset.Avg.Gene.Length}{the average length of the genes in the gene set.}
#'   \item{Geneset.Peak.Genes}{the list of genes from the gene set that had at least one peak assigned.}
#'
#' }}
#'
#' @family enrichment functions
#'
#' @examples
#'
#' # Run ChipEnrich using an example dataset, assigning peaks to the nearest TSS,
#' # and on a small custom geneset
#' data(peaks_E2F4, package = 'chipenrich.data')
#' peaks_E2F4 = subset(peaks_E2F4, peaks_E2F4$chrom == 'chr1')
#' gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
#' results = chipenrich(peaks_E2F4, method='chipenrich', locusdef='nearest_tss',
#' 			genome = 'hg19', genesets=gs_path, out_name=NULL)
#'
#' # Get the list of peaks that were assigned to genes.
#' assigned_peaks = results$peaks
#'
#' # Get the results of enrichment testing.
#' enrich = results$results
#'
#' @export
#' @include constants.R utils.R supported.R setup.R randomize.R
#' @include read.R assign_peaks.R peaks_per_gene.R
#' @include plot_dist_to_tss.R plot_chipenrich_spline.R
#' @include test_approx.R test_binomial.R test_fisher.R test_chipenrich_slow.R test_chipenrich.R test_chipapprox.R
chipenrich = function(
	peaks,
	out_name = "chipenrich",
	out_path = getwd(),
	genome = supported_genomes(),
	genesets = c(
		'GOBP',
		'GOCC',
		'GOMF'),
	locusdef = "nearest_tss",
	method = 'chipenrich',
	mappability = NULL,
	fisher_alt = "two.sided",
	qc_plots = TRUE,
	min_geneset_size = 15,
	max_geneset_size = 2000,
	num_peak_threshold = 1,
	randomization = NULL,
	n_cores = 1
) {
	genome = match.arg(genome)

	n_cores = reset_ncores_for_windows(n_cores)

	############################################################################
	# Collect options for opts output
	opts_list = as.list(sys.call())
	opts_list = opts_list[2:length(opts_list)]

	opts = data.frame(
		parameters = names(opts_list),
		values = as.character(opts_list),
		stringsAsFactors = FALSE
	)

	############################################################################
	# Setup locus definitions, genesets, and mappability

	ldef_list = setup_locusdef(locusdef, genome, randomization)
	ldef = ldef_list[['ldef']]
	tss = ldef_list[['tss']]

	geneset_list = setup_genesets(gs_codes = genesets, ldef_obj = ldef, genome = genome, min_geneset_size = min_geneset_size, max_geneset_size = max_geneset_size)

	mappa = setup_mappa(mappa_code = mappability, genome = genome, ldef_code = locusdef, ldef_obj = ldef)

	############################################################################
	# CHECK method and get() it if okay
	testf = get_test_method(method)
	test_func = get(testf)
	method_name = METHOD_NAMES[[method]]

	############################################################################
	############################################################################
    # Start enrichment process
	############################################################################
	############################################################################

	######################################################
	# Read in and format peaks (from data.frame or file)
	if (class(peaks) == "data.frame") {
		message('Reading peaks from data.frame...')
		peakobj = load_peaks(peaks)
	} else if (class(peaks) == "character") {
		peakobj = read_bed(peaks)
	}

	# Number of peaks in data.
	num_peaks = length(peakobj)

	######################################################
	# Assign peaks to genes.
	message("Assigning peaks to genes with assign_peaks(...) ..")
	assigned_peaks = assign_peaks(peakobj, ldef, tss)

	######################################################
	# Compute peaks per gene table
	ppg = num_peaks_per_gene(assigned_peaks, ldef, mappa)

	# Add relevant columns to ppg depending on the method
	#if(method == 'chipapprox_old') {
	#	message("Calculating weights for approximate method..")
	#	ppg = calc_approx_weights(ppg,mappa)
	#}

	######################################################
	# Enrichment
	results = list()
	for (gobj in geneset_list) {
		message(sprintf("Test: %s",method_name))
		message(sprintf("Genesets: %s",gobj@type))
		message("Running tests..")
		if (testf == "test_chipenrich_slow") {
			rtemp = test_func(gobj,ppg,n_cores)
		}
		if (testf == "test_fisher_exact") {
			rtemp = test_func(gobj,ppg,alternative=fisher_alt)
		}
		if (testf == "test_binomial") {
			rtemp = test_func(gobj,ppg)
		}
		if (testf == "test_approx") {
			rtemp = test_func(gobj,ppg,nwp=FALSE,n_cores)
		}
		if (testf == "test_chipenrich") {
			rtemp = test_func(gobj,ppg,n_cores)
		}
        if (testf == "test_chipapprox") {
            rtemp = test_func(gobj,ppg,n_cores)
        }

		# Annotate with geneset descriptions.
		rtemp$"Description" = as.character(mget(rtemp$Geneset.ID, gobj@set.name, ifnotfound=NA))
		rtemp$"Geneset.Type" = gobj@type

		results[[gobj@type]] = rtemp
	}
	enrich = Reduce(rbind,results)

	######################################################
	# Post-process enrichment
	# Order columns, add enriched/depleted column as needed, remove bad tests,
	# sort by p-value, rename rownames to integers
	enrich = post_process_enrichments(enrich)

	######################################################
	# Write result objects to files
	if (!is.null(out_name)) {
		filename_analysis = file.path(out_path, sprintf("%s_results.tab", out_name))
		write.table(enrich, file = filename_analysis, row.names = FALSE, quote = FALSE, sep = "\t")
		message("Wrote results to: ", filename_analysis)

		filename_peaks = file.path(out_path, sprintf("%s_peaks.tab", out_name))
		write.table(assigned_peaks, file = filename_peaks, row.names = FALSE, quote = FALSE, sep = "\t")
		message("Wrote peak-to-gene assignments to: ", filename_peaks)

		filename_opts = file.path(out_path, sprintf("%s_opts.tab", out_name))
		write.table(opts, file = filename_opts, row.names = FALSE, quote = FALSE, sep = "\t")
		message("Wrote run options/arguments to: ", filename_opts)

		filename_ppg = file.path(out_path, sprintf("%s_peaks-per-gene.tab", out_name))
		write.table(ppg, file = filename_ppg, row.names = FALSE, quote = FALSE, sep = "\t")
		message("Wrote count of peaks per gene to: ", filename_ppg)

		if (qc_plots) {
			filename_qcplots = file.path(out_path, sprintf("%s_qcplots-1.png", out_name))
			grDevices::png(filename_qcplots)
				print(..plot_chipenrich_spline(gpw = ppg, mappability = mappability, num_peaks = num_peaks))
			grDevices::dev.off()
            message("Wrote QC plots to: ",filename_qcplots)

            filename_qcplots = file.path(out_path, sprintf("%s_qcplots-2.png", out_name))
            grDevices::png(filename_qcplots)
                print(..plot_dist_to_tss(peakobj, tss))
            grDevices::dev.off()
			message("Wrote QC plots to: ",filename_qcplots)
		}
	}

	######################################################
	# Return objects as list
	return(list(
		peaks = assigned_peaks,
		results = enrich,
		opts = opts,
		peaks_per_gene = ppg
	))
}

# For each peak, find the nearest TSS, and distance to it.
peak_nearest_tss = function(peaks, tss) {
	########################################################
	# NOTE: THIS CODE IS RECYCLED FROM assign_peaks() with slight mods to
	# mid_dist_df to match previous output.

	# Determine midpoints of peaks and construct a GRanges object
	# on that basis. Include the peak name for later merging
	peak_mids = IRanges::mid(GenomicRanges::ranges(peaks))
	mids_gr = GenomicRanges::GRanges(
		seqnames = Seqinfo::seqnames(peaks),
		ranges = IRanges::IRanges(start = peak_mids, end = peak_mids),
		name = GenomicRanges::mcols(peaks)$name
	)

	mid_dist_to_ntss = GenomicRanges::distanceToNearest(mids_gr, tss)

	mid_indices = S4Vectors::queryHits(mid_dist_to_ntss)
	tss_indices = S4Vectors::subjectHits(mid_dist_to_ntss)
	mid_dist_df = data.frame(
		chr = Seqinfo::seqnames(mids_gr)[mid_indices],
		peak_midpoint = GenomicRanges::start(mids_gr)[mid_indices],
		nearest_tss = GenomicRanges::start(tss)[tss_indices],
		dist_to_tss = GenomicRanges::mcols(mid_dist_to_ntss)$distance,
		stringsAsFactors = FALSE
	)
	########################################################

	return(mid_dist_df)
}

#' Plot histogram of distance from peak to nearest TSS
#'
#' Create a histogram of the distance from each peak to the nearest transcription
#' start site (TSS) of any gene.
#'
#' @param peaks Either a file path or a \code{data.frame} of peaks in BED-like
#' format. If a file path, the following formats are fully supported via their
#' file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
#' or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
#' without these extensions are supported under the conditions that the first 3
#' columns correspond to 'chr', 'start', and 'end' and that there is either no
#' header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
#' \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
#' acceptable column names.
#' @param genome One of the \code{supported_genomes()}.
#'
#' @return A trellis plot object.
#'
#' @examples
#'
#' # Create histogram of distance from peaks to nearest TSS.
#' data(peaks_E2F4, package = 'chipenrich.data')
#' peaks_E2F4 = subset(peaks_E2F4, peaks_E2F4$chrom == 'chr1')
#' plot_dist_to_tss(peaks_E2F4, genome = 'hg19')
#'
#' @export
#' @include constants.R utils.R supported.R setup.R randomize.R
#' @include read.R assign_peaks.R peaks_per_gene.R
plot_dist_to_tss = function(peaks, genome = supported_genomes()) {
	# Get peaks from user's file.
	if (class(peaks) == "data.frame") {
		peakobj = load_peaks(peaks)
	} else if (class(peaks) == "character") {
		peakobj = read_bed(peaks)
	}

	# Load TSS site info.
	tss_code = sprintf("tss.%s", genome)
	data(list = tss_code, package = "chipenrich.data", envir = environment())
	tss = get(tss_code)

	plotobj = ..plot_dist_to_tss(peakobj, tss)
	return(plotobj)
}

..plot_dist_to_tss = function(peaks, tss) {
	# Calculate distance to each TSS.
	tss_peaks = peak_nearest_tss(peaks, tss)

	# Create distance to TSS plot.
	max_dist = max(tss_peaks$dist_to_tss)
	breaks = breaks=c(0, 100, 1000, 5000, 10000, 50000, 100000, max_dist)
	dist_table = table(cut(tss_peaks$dist_to_tss, breaks = breaks))
	dist_table = dist_table / sum(dist_table)
	names(dist_table) = c("< 0.1", "0.1 - 1", "1 - 5", "5 - 10", "10 - 50", "50 - 100", "> 100")

	pf = function(...) {
		args <- list(...)
		bar_labels = sprintf("%0.1f%%", args$y * 100)
		panel.text(seq(1, length(args$x)), args$y + 0.03, bar_labels, cex=1.5)
		panel.barchart(...)
	}

	plotobj = barchart(
		dist_table,
		panel = pf,
		horizontal = FALSE,
		scales = list(rot = 45, cex = 1.6),
		col = "gray",
		ylim = c(0,1),
		ylab = list(label = "Proportion of Peaks", cex=1.65),
		xlab = list(label = "Distance to TSS (kb)", cex=1.65),
		main = list(label = "Distribution of Distance from Peaks to Nearest TSS", cex=1.45)
	)

	return(plotobj)
}

#' Running Hybrid test, either from scratch or using two results files
#'
#' Hybrid test is designed for people unsure of which test between ChIP-Enrich
#' and Poly-Enrich to use, so it takes information of both and gives adjusted
#' P-values. For more about ChIP- and Poly-Enrich, consult their corresponding
#' documentation.
#'
#' @section Hybrid p-values:
#' Given n tests that test for the same hypothesis, same Type I error rate, and
#' converted to p-values: \code{p_1, ..., p_n}, the Hybrid p-value is computed as:
#' \code{n*min(p_1, ..., p_n)}. This hybrid test will have at most the same
#' Type I error as any individual test, and if any of the tests have 100\% power as
#' sample size goes to infinity, then so will the hybrid test.
#'
#' @section Function inputs:
#' Every input in hybridenrich is the same as in chipenrich and polyenrich. Inputs
#' unique to chipenrich are: num_peak_threshold; and inputs unique to polyenrich are:
#' weighting. Currently the test only supports running chipenrich and polyenrich, but
#' future plans will allow you to run any number of different support tests.
#'
#' @param peaks Either a file path or a \code{data.frame} of peaks in BED-like
#' format. If a file path, the following formats are fully supported via their
#' file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
#' or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
#' without these extensions are supported under the conditions that the first 3
#' columns correspond to 'chr', 'start', and 'end' and that there is either no
#' header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
#' \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
#' acceptable column names.
#' @param out_name Prefix string to use for naming output files. This should not
#' contain any characters that would be illegal for the system being used (Unix,
#' Windows, etc.) The default value is "chipenrich", and a file "chipenrich_results.tab"
#' is produced. If \code{qc_plots} is set, then a file "chipenrich_qcplots.pdf"
#' is produced containing a number of quality control plots. If \code{out_name}
#' is set to NULL, no files are written, and results then must be retrieved from
#' the list returned by \code{chipenrich}.
#' @param out_path Directory to which results files will be written out. Defaults
#' to the current working directory as returned by \code{\link{getwd}}.
#' @param genome One of the \code{supported_genomes()}.
#' @param genesets A character vector of geneset databases to be tested for
#' enrichment. See \code{supported_genesets()}. Alternately, a file path to a
#' a tab-delimited text file with header and first column being the geneset ID
#' or name, and the second column being Entrez Gene IDs. For an example custom
#' gene set file, see the vignette.
#' @param locusdef One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
#' '1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
#' '10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
#' see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
#' a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
#' must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
#' example custom locus definition file, see the vignette.
#' @param methods A character string array specifying the method to use for enrichment
#' testing. Currently actually unused as the methods are forced to be one chipenrich
#' and one polyenrich.
#' @param weighting A character string specifying the weighting method. Method name will
#' automatically be "polyenrich_weighted" if given weight options. Current options are:
#' 'signalValue', 'logsignalValue', and 'multiAssign'.
#' @param mappability One of \code{NULL}, a file path to a custom mappability file,
#' or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
#' If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
#' Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
#' For an example custom mappability file, see the vignette. Default value is NULL.
#' @param qc_plots A logical variable that enables the automatic generation of
#' plots for quality control.
#' @param min_geneset_size Sets the minimum number of genes a gene set may have
#' to be considered for enrichment testing.
#' @param max_geneset_size Sets the maximum number of genes a gene set may have
#' to be considered for enrichment testing.
#' @param num_peak_threshold Sets the threshold for how many peaks a gene must
#' have to be considered as having a peak. Defaults to 1. Only relevant for
#' Fisher's exact test and ChIP-Enrich methods.
#' @param randomization One of \code{NULL}, 'complete', 'bylength', or 'bylocation'.
#' See the Randomizations section below.
#' @param n_cores The number of cores to use for enrichment testing. We recommend
#' using only up to the maximum number of \emph{physical} cores present, as
#' virtual cores do not significantly decrease runtime. Default number of cores
#' is set to 1. NOTE: Windows does not support multicore enrichment.
#'
#' @section Joining two results files:
#' Combines two existing results files and returns one results file with hybrid
#' p-values and FDR included. Current allowed inputs are objects from any of
#' the supplied enrichment tests or a dataframe with at least the following columns:
#' \code{P.value, Geneset.ID}. Optional columns include: \code{Status}. Currently
#' we only allow for joining two results files, but future plans will allow you to join
#' any number of results files.
#'
#' @return A data.frame containing:
#' \item{results }{
#' A data frame of the results from performing the gene set enrichment test on
#' each geneset that was requested (all genesets are merged into one final data
#' frame.) The columns are:
#'
#' \describe{
#'   \item{Geneset.ID}{ is the identifier for a given gene set from the selected database.  For example, GO:0000003. }
#'   \item{P.Value.x}{ is the probability of observing the degree of enrichment of the gene set given the null hypothesis
#'                     that peaks are not associated with any gene sets, for the first test.}
#'   \item{P.Value.y}{ is the same as above except for the second test.}
#'   \item{P.Value.Hybrid}{ The calculated Hybrid p-value from the two tests}
#'   \item{FDR.Hybrid}{ is the false discovery rate proposed by Bejamini \& Hochberg for adjusting the p-value to control for family-wise error rate.}
#'  Other variables given will also be included, see the corresponding methods' documentation for their details.
#' }}
#'
#' @export
#' @include chipenrich.R polyenrich.R

hybridenrich <- function(	peaks,
    out_name = "hybridenrich",
    out_path = getwd(),
    genome = supported_genomes(),
    genesets = c('GOBP','GOCC','GOMF'),
    locusdef = "nearest_tss",
    methods = c('chipenrich','polyenrich'),
    weighting = NULL,
    mappability = NULL,
    qc_plots = TRUE,
    min_geneset_size = 15,
    max_geneset_size = 2000,
    num_peak_threshold = 1,
    randomization = NULL,
    n_cores = 1
) {
    genome = match.arg(genome)
    
    n_cores = reset_ncores_for_windows(n_cores)
    
    ############################################################################
    # Collect options for opts output
    opts_list = as.list(sys.call())
    opts_list = opts_list[2:length(opts_list)]
    
    opts = data.frame(
        parameters = names(opts_list),
        values = as.character(opts_list),
        stringsAsFactors = FALSE
    )
    
    ############################################################################
    # Setup locus definitions, genesets, and mappability
    
    ldef_list = setup_locusdef(locusdef, genome, randomization)
    ldef = ldef_list[['ldef']]
    tss = ldef_list[['tss']]
    
    geneset_list = setup_genesets(gs_codes = genesets, ldef_obj = ldef, genome = genome, min_geneset_size = min_geneset_size, max_geneset_size = max_geneset_size)
    
    mappa = setup_mappa(mappa_code = mappability, genome = genome, ldef_code = locusdef, ldef_obj = ldef)

    
    ############################################################################
    ############################################################################
    # Start enrichment process
    ############################################################################
    ############################################################################
    
    ######################################################
    # Read in and format peaks (from data.frame or file)
    if (class(peaks) == "data.frame") {
        message('Reading peaks from data.frame...')
        peakobj = load_peaks(peaks)
    } else if (class(peaks) == "character") {
        peakobj = read_bed(peaks)
    }
    
    # Number of peaks in data.
    num_peaks = length(peakobj)
    
    ######################################################
    # Assign peaks to genes.
    message("Assigning peaks to genes with assign_peaks(...) ..")
    assigned_peaks = assign_peaks(peakobj, ldef, tss)
    
    ######################################################
    # Compute peaks per gene table
    ppg = num_peaks_per_gene(assigned_peaks, ldef, mappa)
    
    if (!is.null(weighting)) {
        if (!all(weighting %in% c("logsignalValue","signalValue","multiAssign"))) {
            # Unsupported weights
            stop(sprintf("Unsupported weights: %s",
            paste(weighting[which(!(weighting %in% c("logsignalValue","signalValue","multiAssign")))],collapse=", ")))
        }
        assigned_peaks = calc_peak_weights(assigned_peaks, weighting)
        ppg = calc_genes_peak_weight(assigned_peaks, ppg)

    }
    # There is a for loop here to run through more than two tests, but later code only supports CE and PE for now. You need
    # to change the hybrid.join lines before deciding to add hybrid functionality to more than 2 methods!
    methodindex = 0
    
    # Initializing to be a list later.
    enrich = NULL
    for (method in methods) {
        methodindex = methodindex+1
        if (!is.null(weighting) & method=="polyenrich") {
            warning("Weights given but polyenrich_weighted method not chosen. Automatically switching to weighted.")
            method = "polyenrich_weighted"
        } # This isn't exactly the best workaround, as it forces weighted even if you don't want to. We need to change this
        ### if we want to hybrid between weighted and non-weighted. We also cannot hybrid two different types of weighting.
   
        ############################################################################
        # CHECK method and get() it if okay
        testf = get_test_method(method)
        test_func = get(testf)
        method_name = METHOD_NAMES[[method]]

        ######################################################
        # Enrichment
        results = list()
        for (gobj in geneset_list) {
            message(sprintf("Test: %s",method_name))
            message(sprintf("Genesets: %s",gobj@type))
            message("Running tests..")
            if (testf == "test_chipenrich_slow") {
                rtemp = test_func(gobj,ppg,n_cores)
            }
            if (testf == "test_fisher_exact") {
                rtemp = test_func(gobj,ppg,alternative="two.sided")
            }
            if (testf == "test_binomial") {
                rtemp = test_func(gobj,ppg)
            }
            if (testf == "test_approx") {
                rtemp = test_func(gobj,ppg,nwp=FALSE,n_cores)
            }
            if (testf == "test_chipenrich") {
                rtemp = test_func(gobj,ppg,n_cores)
            }
            if (testf == "test_chipapprox") {
                rtemp = test_func(gobj,ppg,n_cores)
            }
            if (testf == "test_polyenrich_slow") {
                rtemp = test_func(gobj,ppg,n_cores)
            }
            if (testf == "test_polyenrich") {
                rtemp = test_func(gobj,ppg,n_cores)
            }
            if (testf == "test_polyenrich_weighted") {
                # Note that there is an extra input for the name of the count column
                # In the future when we make a function purely for enrichment, this will be where you
                # give the name of the count column.
                rtemp = test_func(gobj,ppg,n_cores, "sum_peak_weight")
            }
            if (testf == "test_polyapprox") {
                rtemp = test_func(gobj,ppg,n_cores)
            }

            
            # Annotate with geneset descriptions.
            rtemp$"Description" = as.character(mget(rtemp$Geneset.ID, gobj@set.name, ifnotfound=NA))
            rtemp$"Geneset.Type" = gobj@type
            
            results[[gobj@type]] = rtemp
        }
        enrich[[methodindex]] = Reduce(rbind,results)
        
        ######################################################
        # Post-process enrichment
        # Order columns, add enriched/depleted column as needed, remove bad tests,
        # sort by p-value, rename rownames to integers
        enrich[[methodindex]] = post_process_enrichments(enrich[[methodindex]])
    
    }
    
    #Temporary only 2 tests hybridenrich solution:
    message("Combining tests..")
    hybrid = hybrid.join(enrich[[1]],enrich[[2]])
    
    
    ######################################################
    # Write result objects to files
    if (!is.null(out_name)) {
        filename_analysis = file.path(out_path, sprintf("%s_results.tab", out_name))
        write.table(hybrid, file = filename_analysis, row.names = FALSE, quote = FALSE, sep = "\t")
        message("Wrote results to: ", filename_analysis)
        
        filename_peaks = file.path(out_path, sprintf("%s_peaks.tab", out_name))
        write.table(assigned_peaks, file = filename_peaks, row.names = FALSE, quote = FALSE, sep = "\t")
        message("Wrote peak-to-gene assignments to: ", filename_peaks)
        
        filename_opts = file.path(out_path, sprintf("%s_opts.tab", out_name))
        write.table(opts, file = filename_opts, row.names = FALSE, quote = FALSE, sep = "\t")
        message("Wrote run options/arguments to: ", filename_opts)
        
        filename_ppg = file.path(out_path, sprintf("%s_peaks-per-gene.tab", out_name))
        write.table(ppg, file = filename_ppg, row.names = FALSE, quote = FALSE, sep = "\t")
        message("Wrote count of peaks per gene to: ", filename_ppg)
        
        if (qc_plots) {
            filename_qcplots = file.path(out_path, sprintf("%s_qcplots.png", out_name))
            #grDevices::png(filename_qcplots)
            #print(..plot_chipenrich_spline(gpw = ppg, mappability = mappability, num_peaks = num_peaks))
            #print(..plot_dist_to_tss(peakobj, tss))
            #grDevices::dev.off()
            
            
            filename_qcplots_chip = file.path(out_path, sprintf("%s_qcplots_chip.png", out_name))
            filename_qcplots_poly = file.path(out_path, sprintf("%s_qcplots_poly.png", out_name))
            filename_disttotss = file.path(out_path,sprintf("%s_locuslength.jpeg",out_name));
            
            
            grDevices::png(filename_qcplots_chip)
            print(..plot_chipenrich_spline(gpw = ppg, mappability = mappability, num_peaks = num_peaks))				
            grDevices::dev.off()
            
            grDevices::png(filename_qcplots_poly)
            print(..plot_polyenrich_spline(gpw = ppg, mappability = mappability, num_peaks = num_peaks))
            grDevices::dev.off()
            
            grDevices::png(filename_disttotss)
            print(..plot_dist_to_tss(peakobj, tss))
            grDevices::dev.off()
            
            
            message("Wrote QC plots to: ",filename_qcplots)
        }
    }
    
    ######################################################
    # Return objects as list
    return(list(
    peaks = assigned_peaks,
    results = hybrid,
    opts = opts,
    peaks_per_gene = ppg
    ))
}


hybridenrich.old <- function(	peaks,
						out_name = "hybridenrich",
						out_path = getwd(),
						genome = supported_genomes(),
						genesets = c(
							'GOBP',
							'GOCC',
							'GOMF'),
						locusdef = "nearest_tss",
						methods = c('chipenrich','polyenrich'),
						weighting = NULL,
						mappability = NULL,
						qc_plots = TRUE,
						min_geneset_size = 15,
						max_geneset_size = 2000,
						num_peak_threshold = 1,
						randomization = NULL,
						n_cores = 1
) {
    if (!is.null(out_name)) {
        out_chip = sprintf("%s_chip",out_name)
        out_poly = sprintf("%s_poly",out_name)
    } else {
        out_chip = NULL
        out_poly = NULL
    }
    
    #Check if methods number isn't 2. Will support more than 2 later.
    #Also not really relevant as you're forced to use chip and poly anyway.
    if (length(methods) != 2) {
        stop("Hybrid test currently only supports exactly two methods!")
    }
    
	results1 = chipenrich(
        peaks = peaks,
        out_name = out_chip,
        out_path = out_path,
        genome = genome,
        genesets = genesets,
        locusdef = locusdef,
        method = "chipenrich",
        mappability = mappability,
        qc_plots = qc_plots,
        min_geneset_size = min_geneset_size,
        max_geneset_size = max_geneset_size,
        num_peak_threshold = num_peak_threshold,
        randomization = randomization,
        n_cores = n_cores)
        
        
    if (is.null(weighting)) {
        polymeth = "polyenrich"
    } else {
        polymeth = "polyenrich_weighted"
    }
    results2 = polyenrich(
        peaks = peaks,
        out_name = out_poly,
        out_path = out_path,
        genome = genome,
        genesets = genesets,
        locusdef = locusdef,
        method = polymeth,
        weighting = weighting,
        mappability = mappability,
        qc_plots = qc_plots,
        min_geneset_size = min_geneset_size,
        max_geneset_size = max_geneset_size,
        randomization = randomization,
        n_cores = n_cores)
        
    hybrid = hybrid.join(results1,results2)
    
    return(hybrid)
}



#User gives results objects (object or just results), and appends hybrid results
hybrid.join <- function(test1, test2) {
	#Check if they inputed the test object or just the results file, checked by seeing
    # if object has a $results part
    if ("results" %in% names(test1)) {
        #If entire object, extract the results section
        results1 = test1$results
    } else if ("P.value" %in% names(test1)) {
        results1 = test1
    } else {
        stop("First object is not a valid output or does not have P.value column")
    }
	
    if ("results" %in% names(test2)) {
        #If entire object, extract the results section
        results2 = test2$results
    } else if ("P.value" %in% names(test2)) {
        results2 = test2
    } else {
        stop("Second object is not a valid output or does not have P.value column")
    }
    
    #Check for Geneset.ID column
    if (!("Geneset.ID" %in% names(results1))) {
        stop("First object does not have Geneset.ID column")
    }
    if (!("Geneset.ID" %in% names(results2))) {
        stop("Second object does not have Geneset.ID column")
    }
    
    
    

    
    #Separate tree if the data does not have Status column
    if ("Status" %in% names(results1) & "Status" %in% names(results2)) {
        #Extract p-value and status of both tests
        Pvals1 = results1[,c("Geneset.ID","P.value","Status")]
        Pvals2 = results2[,c("Geneset.ID","P.value","Status")]
    
    } else {
        #Extract p-value only
        Pvals1 = results1[,c("Geneset.ID","P.value")]
        Pvals2 = results2[,c("Geneset.ID","P.value")]

    }

    #Merge by Geneset.ID
    PvalsH = merge(Pvals1, Pvals2, by="Geneset.ID")
    #If 0 remain, stop.
    if (nrow(PvalsH) == 0) {
        stop("No common genesets in the two datasets!")
    }
    message(sprintf("Total of %s common Geneset.IDs", nrow(PvalsH)))


    PvalsH$P.value.Hybrid = 2*pmin(PvalsH$P.value.x, PvalsH$P.value.y, 0.5)
	
	#Run B-H to adjust for FDR for hybrid p-values
    PvalsH$FDR.Hybrid = stats::p.adjust(PvalsH$P.value.Hybrid, method = "BH")
    
    
    #Include enrich/depleted status if available and combine
    if ("Status" %in% names(results1) & "Status" %in% names(results2)) {
        PvalsH$Status.Hybrid = ifelse(PvalsH$Status.x == PvalsH$Status.y, PvalsH$Status.x, "Inconsistent")
        #Combine both results files together and append hybrid p-value and FDR
        resultsH = merge(results1[,-which(colnames(results1) %in% c("P.value","Status"))], PvalsH, by = "Geneset.ID")
    } else {
        resultsH = merge(results1[,-which(colnames(results1) %in% c("P.value"))], PvalsH, by = "Geneset.ID")
    }
    
    
    #Reorder columns?????
	
	#Output final results
	return(resultsH)
	
}

context('Test enrich functions')

test_that('Errors from no peaks', {
	ldef_file = system.file('extdata', 'test_ldef_symbol.txt', package = 'chipenrich')
	peak_file = system.file('extdata', 'test.broadPeak', package = 'chipenrich')
	gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')

	expect_error(
		suppressWarnings(chipenrich(peaks = peak_file, genome = 'hg19', genesets = gs_path, locusdef = ldef_file, method = 'binomial', qc_plots = F, out_name = NULL, n_cores = 1)),
		'Intersection between peak midpoints and locus definition is empty')

	expect_error(
		suppressWarnings(chipenrich(peaks = peak_file, genome = 'hg19', genesets = gs_path, locusdef = ldef_file, qc_plots = F, out_name = NULL, n_cores = 1)),
		'Intersection between peak midpoints and locus definition is empty')

	expect_error(
		suppressWarnings(polyenrich(peaks = peak_file, genome = 'hg19', genesets = gs_path, locusdef = ldef_file, qc_plots = F, out_name = NULL, n_cores = 1)),
		'Intersection between peak midpoints and locus definition is empty')

	expect_error(
		suppressWarnings(broadenrich(peaks = peak_file, genome = 'hg19', genesets = gs_path, locusdef = ldef_file, qc_plots = F, out_name = NULL, n_cores=1)),
		'Intersection between peaks and locus definition is empty')
})

data(peaks_E2F4, package = 'chipenrich.data')
peaks_E2F4 = subset(peaks_E2F4, peaks_E2F4$chrom == 'chr1')

data(peaks_H3K4me3_GM12878, package = 'chipenrich.data')
peaks_H3K4me3_GM12878 = subset(peaks_H3K4me3_GM12878, peaks_H3K4me3_GM12878$chrom == 'chr1')

gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')

test_that('Test chipenrich method without mappability', {
	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

	expect_equal(class(results), 'list')
})

test_that('Test chipenrich method with complete randomization', {
	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, randomization = 'complete', n_cores = 1))

	expect_equal(class(results), 'list')
})

test_that('Test chipenrich method with bylength randomization', {
	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, randomization = 'bylength', n_cores = 1))

	expect_equal(class(results), 'list')
})

test_that('Test chipenrich method with bylocation randomization', {
	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, randomization = 'bylocation', n_cores = 1))

	expect_equal(class(results), 'list')
})

test_that('Test chipenrich method with mappability', {
	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
		locusdef = "nearest_tss", mappability = 24, qc_plots = F,
		out_name = NULL,n_cores=1))

	expect_equal(class(results), 'list')
})

test_that('Test chipenrich_slow method', {
	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method = 'chipenrich_slow',
		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

	expect_equal(class(results), 'list')
})

test_that('Test chipapprox method', {
    results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method = 'chipapprox',
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

    expect_equal(class(results), 'list')
})

#test_that('Test old chipapprox method', {
#	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method = 'chipapprox_old',
#		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

#	expect_equal(class(results), 'list')
#})

test_that('Test polyenrich method', {
	results = suppressWarnings(polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method = 'polyenrich',
		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

	expect_equal(class(results), 'list')
})

test_that('Test polyenrich_slow method', {
	results = suppressWarnings(polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method = 'polyenrich_slow',
		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

	expect_equal(class(results), 'list')
})

test_that('Test polyapprox method', {
    results = suppressWarnings(polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method = 'polyapprox',
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

    expect_equal(class(results), 'list')
})

peaks_plus_sv = peaks_E2F4
peaks_plus_sv$signalValue = runif(nrow(peaks_E2F4))*100+10

test_that('Error from polyenrich weighted method without a signalValue column', {
    expect_error(
    suppressWarnings(polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,  method = 'polyenrich_weighted', weighting = 'signalValue',
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1)),
    'No signalValue column!')
})

test_that('Error from choosing weights but not polyenrich_weighted', {
    expect_error(
    suppressWarnings(polyenrich(peaks = peaks_plus_sv, genome = 'hg19', genesets = gs_path,  method = 'polyenrich', weighting = 'signalValue',
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1)),
    'You need to use method = polyenrich_weighted if choosing weighting!')
})

test_that('Error from polyenrich weighted picking both signalValue and logsignalValue', {
    expect_error(
    suppressWarnings(polyenrich(peaks = peaks_plus_sv, genome = 'hg19', genesets = gs_path,  method = 'polyenrich_weighted', weighting = c('signalValue','logsignalValue'),
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1)),
    'You can only choose one of signalValue and logsignalValue!')
})

# this test need to be fixed
# test_that('Error from polyenrich weighted picking no weighting option', {
#     expect_error(
#     suppressWarnings(polyenrich(peaks = peaks_plus_sv, genome = 'hg19', genesets = gs_path,  method = 'polyenrich_weighted', weighting = NULL,
#     locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1)),
#     'No weight options selected!')
# })

test_that('Error from polyenrich weighted picking unsupported weights', {
    expect_error(
    suppressWarnings(polyenrich(peaks = peaks_plus_sv, genome = 'hg19', genesets = gs_path,  method = 'polyenrich_weighted', weighting = c('foo','bar'),
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1)),
    'Unsupported weights: foo, bar')
})

test_that('Test polyenrich weighted method with signalValue', {
    results = suppressWarnings(polyenrich(peaks = peaks_plus_sv, genome = 'hg19', genesets = gs_path, method = 'polyenrich_weighted', weighting = 'signalValue',
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

    expect_equal(class(results), 'list')
})

test_that('Test polyenrich weighted method with logsignalValue', {
    results = suppressWarnings(polyenrich(peaks = peaks_plus_sv, genome = 'hg19', genesets = gs_path, method = 'polyenrich_weighted', weighting = 'logsignalValue',
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

    expect_equal(class(results), 'list')
})


test_that('Test polyenrich weighted method with signalValue AND multiAssign', {
    results = suppressWarnings(polyenrich(peaks = peaks_plus_sv, genome = 'hg19', genesets = gs_path, method = 'polyenrich_weighted', weighting = 'signalValue',
    multiAssign = TRUE, locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))
    
    expect_equal(class(results), 'list')
})

test_that('Test broadenrich method', {
	results = suppressWarnings(broadenrich(peaks = peaks_H3K4me3_GM12878, genome = 'hg19', genesets = gs_path,
		locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores=1))

	expect_equal(class(results), 'list')
})

test_that('Test FET method', {
	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, locusdef = "5kb",
		method = "fet", fisher_alt = "two.sided", qc_plots = F, out_name = NULL))

	expect_equal(class(results), 'list')
})

test_that('Test binomial method', {
	results = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, locusdef = "5kb",
		method = "binomial", qc_plots = F, out_name = NULL))

	expect_equal(class(results), 'list')
})

test_that('Test hybrid from scratch', {
    resultsH = suppressWarnings(hybridenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

    expect_equal(class(resultsH), 'list')
})

test_that('Test hybrid from inputting two different types of results', {
    resultsC = suppressWarnings(chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))

    resultsP = suppressWarnings(polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
    locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1))


    resultsH1 = suppressWarnings(hybrid.join(resultsC, resultsP$results))
    resultsH2 = suppressWarnings(hybrid.join(resultsC$results, resultsP))
})

test_that('Test hybrid.join from inputting invalid results', {
    dummyresults = data.frame('P.value' = 1:5/10, 'Geneset.ID' = 5:9, 'foo' = 5:1)

    expect_error(
        suppressWarnings(hybrid.join(1:5, 2:6)),
        'First object is not a valid output or does not have P.value column'
    )

    expect_error(
        suppressWarnings(hybrid.join(dummyresults[,-2], dummyresults)),
        'First object does not have Geneset.ID column'
    )

    expect_error(
        suppressWarnings(hybrid.join(dummyresults[1:2,], dummyresults[3:5,])),
        'No common genesets in the two datasets!'
    )


})

test_that('Test proxReg tss method', {
    results = suppressWarnings(proxReg(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, reglocation = "tss", qc_plots = F, out_name = NULL))
})

test_that('Test proxReg enhancer method', {
    results = suppressWarnings(proxReg(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, reglocation = "enhancer", qc_plots = F, out_name = NULL))
})

# This function calculates a weight for each peak according to how many loci it intersects.
calc_peak_weights = function(assigned_peaks, weighting, multiAssign) {
    if (multiAssign == TRUE) {
        peak_counts = table(assigned_peaks$peak_id)
        peak_weights = data.frame(
            'peak_id' = names(peak_counts),
            'peak_weights' = as.numeric(1 / peak_counts),
            stringsAsFactors=FALSE)
	
        assigned_peaks = merge(assigned_peaks, peak_weights, by='peak_id')
        message("Assigning weights based on multiple peak assignments...")

    } else {
        assigned_peaks$peak_weights = 1
    }
    
    if (any(c("signalValue", "logsignalValue") %in% weighting)) {
        if (!all(is.numeric(assigned_peaks$signalValue))) {
            stop("signalValue column is not all numeric!")
        } else if (any(assigned_peaks$signalValue<=0)) {
            stop("signalValue column should be all positive!")
        }
    }
    
    
    if ("signalValue" %in% weighting) {
        #assigned_peaks$peak_weights = assigned_peaks$peak_weights*log(assigned_peaks$signalValue)
        assigned_peaks$peak_weights = assigned_peaks$peak_weights*assigned_peaks$signalValue
        message("Assigning weights based on signalValue...")

    }
    if ("logsignalValue" %in% weighting) {
        #assigned_peaks$peak_weights = assigned_peaks$peak_weights*log(assigned_peaks$signalValue)
        assigned_peaks$peak_weights = assigned_peaks$peak_weights*log(assigned_peaks$signalValue)
        message("Assigning weights based on logsignalValue...")
        
    }
	return(assigned_peaks)
}


# Adds gene_weight column to the peaks per gene result
# Currently the gene_weight is the sum of the peak weights overlapping the locus
calc_genes_peak_weight = function(assigned_peaks, ppg) {
    message("Summing the peak weights...")
	# Sum up the normalized peak weights for each peak in a gene
    assigned_peaks$peak_weights = assigned_peaks$peak_weights/mean(assigned_peaks$peak_weights, na.rm=TRUE)
	rpg = stats::aggregate(peak_weights ~ gene_id, assigned_peaks, sum)
	
	d_rpg = data.frame(gene_id = rpg$gene_id, sum_peak_weight = rpg$peak_weights, stringsAsFactors=FALSE)
	
	result = merge(ppg, d_rpg, by='gene_id', all.x=TRUE)
	result$sum_peak_weight[is.na(result$sum_peak_weight)] = 0
	
	# Order by number of peaks in a gene.
	result = result[order(result$num_peaks,decreasing=TRUE),]
	
	return(result)
}

# Used in plot_expected_peaks(...) and chipenrich(...)
#' Aggregate peak assignments over the \code{gene_id} column
#'
#' For each \code{gene_id}, determine the locus length and the number of peaks.
#'
#' Typically, this function will not be used alone, but inside \code{chipenrich()}.
#'
#' @param assigned_peaks A \code{data.frame} resulting from \code{assign_peaks()} or \code{assign_peak_segments()}.
#' @param locusdef A locus definition object from \code{chipenrich.data}.
#' @param mappa A mappability object from \code{chipenrich.data}.
#'
#' @return A \code{data.frame} with columns \code{gene_id, length, log10_length, num_peaks, peak}. The result is used directly in the gene set enrichment tests in \code{chipenrich()}.
#'
#' @examples
#'
#' data('locusdef.hg19.nearest_tss', package = 'chipenrich.data')
#' data('tss.hg19', package = 'chipenrich.data')
#'
#' file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' assigned_peaks = assign_peaks(
#' 	peaks = peaks,
#' 	locusdef = locusdef.hg19.nearest_tss,
#' 	tss = tss.hg19)
#'
#' ppg = num_peaks_per_gene(
#' 	assigned_peaks = assigned_peaks,
#' 	locusdef = locusdef.hg19.nearest_tss,
#' 	mappa = NULL)
#'
#' @export
num_peaks_per_gene = function(assigned_peaks, locusdef, mappa=NULL) {
	# Add in gene lengths to locusdef data
	# NOTE: the locusdef@dframe is the basis for the returned object because
	# the enrichment tests need to know which genes have no peaks as well as
	# which genes have peaks. The assigned_peaks object just tells us which
	# peaks overlapped a locus.
	d = locusdef@dframe
	d$length = d$end - d$start

	# Sum up lengths for each gene.
	d = stats::aggregate(length ~ gene_id, d, sum)
	d$log10_length = log10(d$length+1)

	# Compute the total number of peaks assigned to each gene.
	ppg = table(assigned_peaks$gene_id)
	d_ppg = data.frame(
		gene_id = names(ppg),
		num_peaks = as.numeric(ppg),
		stringsAsFactors = FALSE)
	result = merge(
		x = d,
		y = d_ppg,
		by = 'gene_id',
		all.x= TRUE)
	result[is.na(result$num_peaks), ]$num_peaks = 0

	# Mappable length if requested
	if (!is.null(mappa)) {
		message("Mappability adjustment is enabled..")
		result = merge(
			x = result,
			y = mappa,
			by = 'gene_id',
			sort= FALSE)
		result$length = as.numeric((result$mappa * result$length) + 1)
		result$log10_length = log10(result$length)
	}

	# Order by number of peaks in a gene.
	result = result[order(result$num_peaks, decreasing= TRUE), ]

	# Add in peak vector (0,1).
	result$peak = as.numeric(result$num_peaks >= 1)

	return(result)
}

# Used for method='broadenrich'
#' Add peak overlap and ratio to result of \code{num_peaks_per_gene()}
#'
#' In particular, for \code{method = 'broadenrich'} in \code{chipenrich()}, when using \code{assign_peak_segments()}. This function will add aggregated \code{peak_overlap} (in base pairs) and \code{ratio} (relative to \code{length}) columns to the result of \code{num_peaks_per_gene()} so the right data is present for the \code{method = 'broadenrich'} model.
#'
#' Typically, this function will not be used alone, but inside \code{chipenrich()} with \code{method = 'broadenrich'}.
#'
#' @param assigned_peaks A \code{data.frame} resulting from \code{assign_peak_segments()}.
#' @param ppg The aggregated peak assignments over \code{gene_id} from \code{num_peaks_per_gene()}.
#'
#' @return A \code{data.frame} with columns \code{gene_id, length, log10_length, num_peaks, peak, peak_overlap, ratio}. The result is used directly in the gene set enrichment tests in \code{chipenrich()} when \code{method = 'broadenrich'}.
#'
#' @examples
#'
#' data('locusdef.hg19.nearest_tss', package = 'chipenrich.data')
#' data('tss.hg19', package = 'chipenrich.data')
#'
#' file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
#' peaks = read_bed(file)
#'
#' assigned_peaks = assign_peak_segments(
#' 	peaks = peaks,
#' 	locusdef = locusdef.hg19.nearest_tss)
#'
#' ppg = num_peaks_per_gene(
#' 	assigned_peaks = assigned_peaks,
#' 	locusdef = locusdef.hg19.nearest_tss,
#' 	mappa = NULL)
#'
#' ppg = calc_peak_gene_overlap(
#' 	assigned_peaks = assigned_peaks,
#' 	ppg = ppg)
#'
#' @export
calc_peak_gene_overlap = function(assigned_peaks, ppg) {
	# Sum up the lengths for each peak in a gene
	rpg = stats::aggregate(peak_overlap ~ gene_id, assigned_peaks, sum)

	d_rpg = data.frame(
		gene_id = rpg$gene_id,
		peak_overlap = rpg$peak_overlap,
		stringsAsFactors = FALSE)
	result = merge(
		x = ppg,
		y = d_rpg,
		by = 'gene_id',
		all.x= TRUE)
	result$peak_overlap[is.na(result$peak_overlap)] = 0

	result$ratio = result$peak_overlap / result$length

	result$ratio[result$ratio > 1] = 1

	# Order by number of peaks in a gene.
	result = result[order(result$num_peaks,decreasing= TRUE),]

	return(result)
}

# Used for method='chipapprox'
# Identical to post-mappa part of calc_weights_chipenrich(...)
# Adds weight, prob_peak, resid.dev columns to peaks per gene result
calc_approx_weights = function(ppg, mappa) {
	d = ppg

	# Create model.
	model = "peak ~ s(log10_length, bs = 'cr')"

	# Compute binomial spline fit.
	fit = gam(as.formula(model), data = d, family = 'binomial')

	# Compute weights for each gene, based on the predicted prob(peak) for each gene.
	ppeak = fitted(fit)
	w0 = 1 / (ppeak/mean(d$peak,na.rm= TRUE))
	w0 = w0 / mean(w0,na.rm= TRUE)

	d$weight = w0
	d$prob_peak = ppeak
	d$resid.dev = resid(fit,type="deviance")

	cols = c("gene_id","length","log10_length","mappa","num_peaks","peak","weight","prob_peak","resid.dev")
	if (is.null(mappa)) {
		cols = setdiff(cols,c("mappa"))
	}
	d = subset(d,select=cols)

	return(d)
}

test_binomial = function(geneset, ppg) {
	warning("The binomial test is provided for comparison purposes only. It will almost always give biased results favoring gene sets with short average locus length.")

	# Restrict to genes only within the genesets.
	ppg = subset(ppg, ppg$gene_id %in% geneset@all.genes)

	# Total # of peaks.
	total_peaks = sum(ppg$num_peaks)

	# Gene IDs must be character to be stored into an environment as names.
	ppg$gene_id = as.character(ppg$gene_id)

	# Index gene --> peak count.
	gene2peakc = new.env(parent = emptyenv())
	apply(ppg, 1, function(x) assign(x["gene_id"], x["num_peaks"], envir = gene2peakc))

	# Index gene --> log10 length.
	gene2length = new.env(parent = emptyenv())
	apply(ppg, 1, function(x) assign(x["gene_id"], x["length"], envir = gene2length))

	# Length of the genome.
	genome_length = sum(as.numeric(ppg$length))

	estimates = c()
	pvals = c()
	goids = c()
	gonums = c()
	n_go_peak_genes = c()
	r_go_peak_genes = c()
	for (i in 1:length(geneset@set.gene)) {
		go_term = ls(geneset@set.gene)[i]
		go_genes = as.character(geneset@set.gene[[go_term]])

		# Eliminate GO genes that aren't in the ppg.
		#go_genes = Filter(function(x) x %in% ppg$gene_id,go_genes)
		go_genes = go_genes[go_genes %in% ppg$gene_id]

		go_gene_lengths = as.numeric(sapply(go_genes,function(x) get(x, gene2length)))
		go_term_length = sum(go_gene_lengths)

		go_gene_counts = as.numeric(sapply(go_genes,function(x) get(x, gene2peakc)))
		go_term_count = sum(go_gene_counts)

		binom_p = go_term_length / genome_length
		btest = stats::binom.test(go_term_count, total_peaks, binom_p, alternative = "greater")

		goids[i] = go_term
		pvals[i] = btest$p.value
		estimates[i] = btest$estimate
		gonums[i] = length(go_genes)

		go_peak_genes = ppg[(ppg$gene_id %in% go_genes) & (ppg$num_peaks > 0),]$gene_id
		r_go_peak_genes[i] = paste(go_peak_genes, collapse = ", ")
		n_go_peak_genes[i] = length(go_peak_genes)
	}

	fdr = stats::p.adjust(pvals,method="BH")

	results = data.frame(
		"Geneset ID" = goids,
		"N Geneset Genes" = gonums,
		"N Geneset Peak Genes" = n_go_peak_genes,
		"P.Success" = estimates,
		"P-value" = pvals,
		"FDR" = fdr,
		"Geneset Peak Genes" = r_go_peak_genes,
		stringsAsFactors = FALSE)

	results = results[order(results$P.value), ]
	return(results)
}

## -----------------------------------------------------------------------------
library(chipenrich)

## ----warning = FALSE, message = FALSE-----------------------------------------
data(peaks_E2F4, package = 'chipenrich.data')
data(peaks_H3K4me3_GM12878, package = 'chipenrich.data')

head(peaks_E2F4)
head(peaks_H3K4me3_GM12878)

## -----------------------------------------------------------------------------
supported_genomes()

## -----------------------------------------------------------------------------
# Take head because it's long
head(supported_locusdefs())

## ----eval=FALSE---------------------------------------------------------------
# chr	start	end	geneid
# chr1	839460	839610	148398
# chr1	840040	840190	148398
# chr1	840040	840190	57801
# chr1	840800	840950	148398
# chr1	841160	841310	148398

## -----------------------------------------------------------------------------
# Take head because it's long
head(supported_genesets())

## ----eval=FALSE---------------------------------------------------------------
# gs_id	gene_id
# GO:0006631	30
# GO:0006631	31
# GO:0006631	32
# GO:0006631	33
# GO:0006631	34
# GO:0006631	35
# GO:0006631	36
# GO:0006631	37
# GO:0006631	51
# GO:0006631	131
# GO:0006631	183
# GO:0006631	207
# GO:0006631	208
# GO:0006631	215
# GO:0006631	225

## ----eval = F-----------------------------------------------------------------
# library(EGSEAdata)
# egsea.data("mouse")
# temp = Mm.H #Or some other gene sets
# geneset = do.call(rbind,lapply(1:length(temp), function(index) {data.frame(gs_id = rep(names(temp)[index], length(temp[[index]])),gene_id = unlist(strsplit(temp[[index]],split = " ")),stringsAsFactors = F)}))
# write.table(geneset, "./custom_geneset.txt", quote=F, sep="\t",row.names = F, col.names = T)

## -----------------------------------------------------------------------------
# Take head because it's long
head(supported_read_lengths())

## ----eval=FALSE---------------------------------------------------------------
# mappa	gene_id
# 0.8	8487
# 0.1	84
# 0.6	91
# 1	1000

## ----warning = FALSE, message = FALSE-----------------------------------------
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = broadenrich(peaks = peaks_H3K4me3_GM12878, genome = 'hg19', genesets = gs_path, locusdef = "nearest_tss", qc_plots = FALSE, out_name = NULL, n_cores=1)
results.be = results$results
print(results.be[1:5,1:5])

## ----warning = FALSE, message = FALSE-----------------------------------------
# Without mappability
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,locusdef = "nearest_tss", qc_plots = FALSE, out_name = NULL, n_cores = 1)
results.ce = results$results
print(results.ce[1:5,1:5])

## ----warning = FALSE, message = FALSE-----------------------------------------
# With mappability
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, locusdef = "nearest_tss", mappability=24, qc_plots = FALSE,out_name = NULL,n_cores=1)
results.cem = results$results
print(results.cem[1:5,1:5])

## ----warning = FALSE, message = FALSE-----------------------------------------
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method = 'polyenrich', locusdef = "nearest_tss", qc_plots = FALSE, out_name = NULL, n_cores = 1)
results.pe = results$results
print(results.pe[1:5,1:5])

## ----warning = FALSE, message = FALSE-----------------------------------------
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method="polyenrich_weighted", locusdef = "enhancer_plus5kb", qc_plots = FALSE, out_name = NULL, n_cores = 1)
results.pe = results$results
print(results.pe[1:5,1:5])

## ----warning = FALSE, message = FALSE-----------------------------------------
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = hybridenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1)
results.hybrid = results$results
print(results.hybrid[1:5,1:5])

## ----warning = FALSE, message = FALSE-----------------------------------------
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results.prox = proxReg(peaks_E2F4, reglocation = 'tss',	genome = 'hg19', genesets=gs_path, out_name=NULL)
results.prox = results.prox$results
print(results.prox[1:5,1:7])

## ----fig.align='center', fig.cap='E2F4 peak distances to TSS', fig.height=6, fig.width=6, fig.show='hold', warning = FALSE, message = FALSE----
# Output in chipenrich and polyenrich
plot_dist_to_tss(peaks = peaks_E2F4, genome = 'hg19')

## ----fig.align='center', fig.cap='E2F4 chipenrich spline without mappability', fig.height=6, fig.width=6, fig.show='hold', warning = FALSE, message = FALSE----
# Output in chipenrich
plot_chipenrich_spline(peaks = peaks_E2F4, locusdef = 'nearest_tss', genome = 'hg19')

## ----fig.align='center', fig.cap='E2F4 polyenrich spline without mappability', fig.height=6, fig.width=6, fig.show='hold', warning = FALSE, message = FALSE----
# Output in polyenrich
plot_polyenrich_spline(peaks = peaks_E2F4, locusdef = 'nearest_tss', genome = 'hg19')

## ----fig.align='center', fig.cap='H3K4me3 gene coverage', fig.height=6, fig.width=6, fig.show='hold', warning = FALSE, message = FALSE----
# Output in broadenrich
plot_gene_coverage(peaks = peaks_H3K4me3_GM12878, locusdef = 'nearest_tss',  genome = 'hg19')

## -----------------------------------------------------------------------------
head(results$peaks)

## -----------------------------------------------------------------------------
head(results$peaks_per_gene)

## -----------------------------------------------------------------------------
head(results$results)

## ----warning = FALSE, message = FALSE-----------------------------------------
# Assessing if alpha = 0.05
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
	locusdef = "nearest_tss", qc_plots = FALSE, randomization = 'complete',
    out_name = NULL, n_cores = 1)
alpha = sum(results$results$P.value < 0.05) / nrow(results$results)
# NOTE: This is for
print(alpha)

---
title: "`chipenrich`: Gene Set Enrichment For ChIP-seq Peak Data"
author: "Ryan P. Welch, Chee Lee, Raymond G. Cavalcante, Chris Lee, Kai Wang, Laura J. Scott, Maureen A. Sartor"
date: "`r Sys.Date()`"
output:
  BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{chipenrich_vignette}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

Gene set enrichment (GSE) testing enables the interpretation of lists of genes (e.g. from RNA-seq), or lists of genomic regions (e.g. from ChIP-seq), in terms of pathways and other biologically meaningful sets of genes. The `chipenrich` package is designed for GSE testing of large sets of genomic regions with different properties. The primary innovation of `chipenrich` is its accounting for biases that are known to affect the Type I error of such testing, and other properties of the data. It offers many options for gene set databases, and several other methods/options: tests for wide versus narrow genomic regions, a thousand versus > a million genomic regions, regions in promoters vs enhancers vs gene bodies, or accounting for the strength/score of each genomic region.

The `chipenrich` package includes different enrichment methods for different use cases:

* `broadenrich()` is designed for use with broad peaks that may intersect multiple gene loci, and cumulatively cover greater than 5% of the genome. For example, most ChIP-seq experiments for histone modifications.
* `chipenrich()` is designed for use with 1,000s or 10,000s of narrow genomic regions which results in a relatively small percent of genes being assigned a genomic region. For example, many ChIP-seq experiments for transcription factors.
* `polyenrich()` is also designed for narrow peaks, but for experiments with > ~50,000 genomic regions, or in cases where the number of binding sites per gene is thought to be important. If unsure whether to use chipenrich or polyenrich, then we recommend hybridenrich.
* `hybridenrich()` is used when one is unsure of which method, between ChIP-Enrich or Poly-Enrich, is the optimal method. It is slightly more conservative than either test for an individual gene set, but is usually more powerful overall.

We recommend using polyenrich(method='polyenrich-weighted', multiAssign=TRUE, .) with any enhancer locus definition. See example below for more details.

# Concepts and Usage

```{r}
library(chipenrich)
```

Due to several required dependencies, installation may take some time if the dependencies are not already installed.

## Peaks

A ChIP-seq peak is a genomic region that represents a transcription factor binding event or the presence of a histone complex with a particular histone modification. Typically peaks are called with a peak caller (such as  [MACS2](https://github.com/taoliu/MACS) or [PePr](https://github.com/shawnzhangyx/PePr/)) and represent relative enrichment of reads in a sample where the antibody is present versus input. Typically, peaks are output by a peak caller in [`BED`-like](https://genome.ucsc.edu/FAQ/FAQformat.html) format.

User input for `chipenrich()`, `broadenrich()`, or `polyenrich()` may be peaks called from a ChIP-seq or ATAC-seq experiment or other large sets of genomic regions (e.g. from a family of repetitive elements or hydroxymethylation experiments) but we shall continue to refer to the input genomic regions as 'peaks' for simplicity. Peaks can be input as either a file path or a `data.frame`.

If a file path, the following formats are fully supported via their file extensions: `.bed`, `.broadPeak`, `.narrowPeak`, `.gff3`, `.gff2`, `.gff`, and `.bedGraph` or `.bdg`. BED3 through BED6 files are supported under the `.bed` extension ([BED specification](https://genome.ucsc.edu/FAQ/FAQformat.html)). Files without these extensions are supported under the conditions that the first 3 columns correspond to `chr`, `start`, and `end` and that there is either no header column, or it is commented out. Files may be compressed with `gzip`, and so might end in `.narrowPeak.gz`, for example. For files with extension support, the `rtracklayer::import()` function is used to read peaks, so adherence to the mentioned file formats is necessary.

If peaks are already in the R environment as a `data.frame`, the `GenomicRanges::makeGRangesFromDataFrame()` function is used to convert to a `GRanges` object. For the acceptable column names needed for correct interpretation, see `?GenomicRanges::makeGRangesFromDataFrame`.

For the purpose of the vignette, we'll load some ChIP-seq peaks from the `chipenrich.data` companion package:

```{r, warning = FALSE, message = FALSE}
data(peaks_E2F4, package = 'chipenrich.data')
data(peaks_H3K4me3_GM12878, package = 'chipenrich.data')

head(peaks_E2F4)
head(peaks_H3K4me3_GM12878)
```

## Genomes

Genomes for fly, human, mouse, rat, and zebrafish are supported. Particular supported genome builds are given by:

```{r}
supported_genomes()
```

## Locus Definitions

A gene locus definition is a way of defining the gene regulatory regions, and enables us to associate peaks with genes. The terms 'gene regulatory region' and 'gene locus' are used interchangeably in the vignette. An example is the '5kb' gene locus definition, which assigns the 5000 bp immediately up- and down- stream of transcription start sites (TSS) to the respective gene. A locus definition can also express from where one expects a transcription factor or histone modification to regulate genes. For example, a locus definition defined as 1kb upstream and downstream of a TSS (the 1kb definition) would capture TFs binding in proximal-promoter regions.

### Built-in locus definitions

A number of locus definitions representing different regulatory paradigms are included in the package:

* `nearest_tss`: The locus is the region spanning the midpoints between the TSSs of adjacent genes. Thus, each genomic region is assigned to the gene with the nearest TSS.
* `nearest_gene`: The locus is the region spanning the midpoints between the boundaries of each gene, where a gene is defined as the region between the furthest upstream TSS and furthest downstream TES for that gene. If gene loci overlap, the midpoint of the overlap is used as a border. If a gene locus is nested in another, the larger locus is split in two.
* `exon`: Each gene has multiple loci corresponding to its exons. Overlaps between different genes are allowed.
* `intron`: Each gene has multiple loci corresponding to its introns. Overlaps between different genes are allowed.
* `1kb`: The locus is the region within 1kb of any of the TSSs belonging to a gene. If TSSs from two adjacent genes are within 2 kb of each other, we use the midpoint between the two TSSs as the boundary for the locus for each gene.
* `1kb_outside_upstream`: The locus is the region more than 1kb upstream from a TSS to the midpoint between the adjacent TSS.
* `1kb_outside`: The locus is the region more than 1kb upstream or downstream from a TSS to the midpoint between the adjacent TSS.
* `5kb`: The locus is the region within 5kb of any of the TSSs belonging to a gene. If TSSs from two adjacent genes are within 10 kb of each other, we use the midpoint between the two TSSs as the boundary for the locus for each gene.
* `5kb_outside_upstream`: The locus is the region more than 5kb upstream from a TSS to the midpoint between the adjacent TSS.
* `5kb_outside`: The locus is the region more than 5kb upstream or downstream from a TSS to the midpoint between the adjacent TSS.
* `10kb`: The locus is the region within 10kb of any of the TSSs belonging to a gene. If TSSs from two adjacent genes are within 20 kb of each other, we use the midpoint between the two TSSs as the boundary for the locus for each gene.
* `10kb_outside_upstream`: The locus is the region more than 10kb upstream from a TSS to the midpoint between the adjacent TSS.
* `10kb_outside`: The locus is the region more than 10kb upstream or downstream from a TSS to the midpoint between the adjacent TSS.

The complete listing of genome build and locus definition pairs can be listed with `supported_locusdefs()`:

```{r}
# Take head because it's long
head(supported_locusdefs())
```

### Custom locus definitions

Users can create custom locus definitions for any of the `supported_genomes()`, and pass the file path as the value of the `locusdef` parameter in `broadenrich()`, `chipenrich()`, or `polyenrich()`. Custom locus definitions should be defined in a tab-delimited text file with column names `chr`, `start`, `end`, and `gene_id`. For example:

```{r, eval=FALSE}
chr	start	end	geneid
chr1	839460	839610	148398
chr1	840040	840190	148398
chr1	840040	840190	57801
chr1	840800	840950	148398
chr1	841160	841310	148398
```

### Selecting a locus definition

For a transcription factor ChIP-seq experiment, selecting a particular locus definition for use in enrichment testing implies how the TF is assumed to regulate genes. For example, selecting the `1kb` locus definition will imply that the biological processes found enriched are a result of TF regulation near the promoter. In contrast, selecting the `5kb_outside` locus definition will imply that the biological processes found enriched are a result of TF regulation distal from the promoter.

Selecting a locus definition can also help reduce the noise in the enrichment tests. For example, if a TF is known to primarily regulate genes by binding around the promoter, then selecting the `1kb` locus definition can help to reduce the noise from TSS-distal peaks in the enrichment testing.

The [`plot_dist_to_tss()` QC plot](#peak-distance-to-tss-distribution) displays where peak midpoints fall relative to TSSs genome-wide, and can help inform the choice of locus definition. For example, if many peaks fall far from the TSS, the `nearest_tss` locus definition may be a good choice because it will capture *all* peaks, whereas the `1kb` locus definition may not capture many of the peaks and adversely affect the enrichment testing.

## Gene Sets

Gene sets are sets of genes that represent a particular biological function.

### Built-in gene sets

Gene sets for fly, human, mouse, rat, and zebrafish are built in to `chipenrich`. Some organisms have gene sets that others do not, so check with:

```{r}
# Take head because it's long
head(supported_genesets())
```

Descriptions of our built-in gene sets:

* `GOBP`: Gene Ontology-Biological Processes, Bioconductor ver. 3.4.2. (http://www.geneontology.org/)

* `GOCC`: Gene Ontology-Cell Component, Bioconductor ver. 3.4.2. (geneontology.org)

* `GOMF`: Gene Ontology-Molecular Function, Bioconductor ver 3.4.2. (geneontology.org)

* `biocarta_pathway`: BioCarta Pathway, Ver 6.0. (cgap.nci.nih.gov/Pathways/BioCarta_Pathways)

* `ctd`: Comparative Toxicogenomics Database, Last updated June 06, 2017. Groups of genes that interact with specific chemicals to help understand enviromental exposures that affect human health. (ctdbase.org)

* `cytoband`: Cytobands (NCBI). Groups of genes that reside in the same area of a chromosome.

* `drug_bank`: Sets of gene that are targeted by a specific drug. Ver 5.0.7. (www.drugbank.ca)

* `hallmark`: Hallmark gene sets (MSigDB). Ver 6.0. Specific biological states or processes that display coherent expression. (software.broadinstitute.org/gsea/msigdb/collections.jsp)

* `immunologic`: Immunologic signatures (MSigDB). Ver 6.0. Gene sets that represent cell states within the immune system. (software.broadinstitute.org/gsea/msigdb/collections.jsp)

* `kegg_pathway`: Kyoto Encyclopedia of Genes and Genomes. Ver 3.2.3. (genome.jp/kegg)

* `mesh`: Gene Annotation with MeSH, the National Library of Medicine's controlled vocabulary for biology and medicine. Useful for testing hypotheses related to diseases, processes, other genes, confounders such as populations and experimental techniques, etc. based on knowledge from the literature that may not yet be formally described in any other gene sets. Last updated ~2013. (gene2mesh.ncibi.org)

* `metabolite`: Metabolite concepts, defined from Edinburgh Human Metabolic Network database (Ma, et al., 2007) Contains gene sets coding for metabolic enzymes that catalyze reactions involving the respective metabolite.

* `microrna`: microRNA targets (MSigDB). Ver 6.0. Gene sets containing genes with putative target sites of human mature miRNA. (software.broadinstitute.org/gsea/msigdb/collections.jsp)

* `oncogenic`: Oncogenic signatures (MSigDB). Ver 6.0. Gene sets that represent signatures of pathways often disregulated in cancer. (software.broadinstitute.org/gsea/msigdb/collections.jsp)

* `panther_pathway`: PANTHER Pathway. Ver 3.5. Contains primarily signaling pathways with subfamilies. (pantherdb.org/pathway)

* `pfam`: Pfam Ver 31.0 (March 2017). A large collection of protein families. (pfam.xfam.org)

* `protein_interaction_biogrid`: Protein Interaction from Biological General Repository for Interaction Datasets. Ver 3.4.151. (thebiogrid.org)

* `reactome`: Reactome Pathway Database. Ver 61. (reactome.org)

* `transcription_factors`: Transcription Factors (MSigDB). Ver 6.0. Gene sets that share upstream cis-regulatory motifs which can function as potential transcription factor binding sites. (software.broadinstitute.org/gsea/msigdb/collections.jsp)

### Custom gene sets

Users can perform GSE on custom gene sets for any supported organism by passing the file path as the value of `genesets` parameter in `broadenrich()`, `chipenrich()`, `polyenrich()`, or `hybridenrich()`. Custom gene set definitions should be defined in a tab-delimited text file with a header. The first column should be the geneset ID or name, and the second column should be the Entrez IDs belonging to the geneset. For example:

```{r, eval=FALSE}
gs_id	gene_id
GO:0006631	30
GO:0006631	31
GO:0006631	32
GO:0006631	33
GO:0006631	34
GO:0006631	35
GO:0006631	36
GO:0006631	37
GO:0006631	51
GO:0006631	131
GO:0006631	183
GO:0006631	207
GO:0006631	208
GO:0006631	215
GO:0006631	225
```

If a gene set from another database is in the form of a list with an array of genes (e.g. EGSEA PMID: 29333246), you can convert it into the neccessary form by the following and input the path directly to the `genesets` parameter. Note: Converting from R to text back to R is quite redundant and we are planning to be able to bypass this step.

```{r eval = F}
library(EGSEAdata)
egsea.data("mouse")
temp = Mm.H #Or some other gene sets
geneset = do.call(rbind,lapply(1:length(temp), function(index) {data.frame(gs_id = rep(names(temp)[index], length(temp[[index]])),gene_id = unlist(strsplit(temp[[index]],split = " ")),stringsAsFactors = F)}))
write.table(geneset, "./custom_geneset.txt", quote=F, sep="\t",row.names = F, col.names = T)
```

## Mappability

We define base pair mappability as the average read mappability of all possible reads of size K that encompass a specific base pair location, $b$. Mappability files from UCSC Genome Browser mappability track were used to calculate base pair mappability. The mappability track provides values for theoretical read mappability, or the number of places in the genome that could be mapped by a read that begins with the base pair location $b$. For example, a value of 1 indicates a Kmer read beginning at $b$ is mappable to one area in the genome. A value of 0.5 indicates a Kmer read beginning at $b$ is mappable to two areas in the genome. For our purposes, we are only interested in uniquely mappable reads; therefore, all reads with mappability less than 1 were set to 0 to indicate non-unique mappability. Then, base pair mappability is calculated as:

$$
\begin{equation}
M_{i} = (\frac{1}{2K-1}) \sum_{j=i-K+1}^{i+(K-1)} M_{j}
\end{equation}
$$

where $M_{i}$ is the mappability of base pair $i$, and $M_{j}$ is mappability (from UCSC's mappability track) of read $j$ where j is the start position of the K length read.

### Built-in mappability

Base pair mappability for reads of lengths 24, 36, 40, 50, 75, and 100 base pairs for `hg19` and for reads of lengths 36, 40, 50, 75, and 100 base pairs `mm9` a included. See the complete list with:

```{r}
# Take head because it's long
head(supported_read_lengths())
```

### Custom mappability

Users can use custom mappability with any built-in locus definition (if, for example, the read length needed is not present), or with a custom locus definition. Custom mappability should be defined in a tab-delimited text file with columns named `gene_id` and `mappa`. Gene IDs should be Entrez Gene IDs, and mappability should be in [0,1]. A check is performed to verify that the gene IDs in the locus definition and mappability overlap by at least 95\%. An example custom mappability file looks like:

```{r, eval=FALSE}
mappa	gene_id
0.8	8487
0.1	84
0.6	91
1	1000
```

## Testing for enrichment

As stated in the introduction, the `chipenrich` package includes three classes of methods for doing GSE testing. For each method, we describe the intended use case, the model used for enrichment, and an example using the method.

### `broadenrich()`

Broad-Enrich is designed for use with broad peaks that may intersect multiple gene loci, and/or cumulatively cover greater than 5\% of the genome. For example, ChIP-seq experiments for histone modifications or large sets of copy number alterations.

The Broad-Enrich method uses the cumulative peak coverage of genes in its logistic regression model for enrichment: `GO ~ ratio + s(log10_length)`. Here, `GO` is a binary vector indicating whether a gene is in the gene set being tested, `ratio` is a numeric vector indicating the ratio of the gene covered by peaks, and `s(log10_length)` is a binomial cubic smoothing spline which adjusts for the relationship between gene coverage and locus length.

```{r, warning = FALSE, message = FALSE}
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = broadenrich(peaks = peaks_H3K4me3_GM12878, genome = 'hg19', genesets = gs_path, locusdef = "nearest_tss", qc_plots = FALSE, out_name = NULL, n_cores=1)
results.be = results$results
print(results.be[1:5,1:5])
```

### `chipenrich()`

ChIP-Enrich is designed for use with 1,000s or 10,000s of narrow genomic regions which results in a relatively small percent of genes being assigned a genomic region. For example, many ChIP-seq experiments for transcription factors.

The ChIP-Enrich method uses the presence of a peak in its logistic regression model for enrichment: `peak ~ GO + s(log10_length)`. Here, `GO` is a binary vector indicating whether a gene is in the gene set being tested, `peak` is a binary vector indicating the presence of a peak in a gene, and `s(log10_length)` is a binomial cubic smoothing spline which adjusts for the relationship between the presence of a peak and locus length.

```{r, warning = FALSE, message = FALSE}
# Without mappability
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,locusdef = "nearest_tss", qc_plots = FALSE, out_name = NULL, n_cores = 1)
results.ce = results$results
print(results.ce[1:5,1:5])
```

```{r, warning = FALSE, message = FALSE}
# With mappability
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, locusdef = "nearest_tss", mappability=24, qc_plots = FALSE,out_name = NULL,n_cores=1)
results.cem = results$results
print(results.cem[1:5,1:5])
```

### `polyenrich()`

Poly-Enrich is also designed for narrow peaks, for experiments with 100,000s of peaks, or in cases where the number of binding sites per gene affects its regulation. If unsure whether to use chipenrich or polyenrich, then we recommend hybridenrich.

The Poly-Enrich method uses the number of peaks in genes in its negative binomial regression model for enrichment: `num_peaks ~ GO + s(log10_length)`. Here, `GO` is a binary vector indicating whether a gene is in the gene set being tested, `num_peaks` is a numeric vector indicating the number of peaks in each gene, and `s(log10_length)` is a negative binomial cubic smoothing spline which adjusts for the relationship between the number of peaks in a gene and locus length.

```{r, warning = FALSE, message = FALSE}
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method = 'polyenrich', locusdef = "nearest_tss", qc_plots = FALSE, out_name = NULL, n_cores = 1)
results.pe = results$results
print(results.pe[1:5,1:5])
```

Poly-Enrich can also be used for linking human distal enhancer regions to their target genes, which are not necessarily the adjacent genes. We optimized human distal enhancer to target gene locus definitions (locusdef="enhancer" or locusdef="enhancer_plus5kb"). locusdef="enhancer" uses only distal regions >5kb from a TSS, while locusdef="enhancer_plus5kb" combines distal enhancers (>5kb from a TSS) with promoters (<=5kb from a TSS) to capture all genomic regions. Poly-Enrich is strongly recommended when using either the 'enhancer' or 'enhancer_plus5kb' gene locus definition, because only polyenrich is able to properly split the weight of genomic regions that are assigned to multiple genes (multiAssign=TRUE). The performance of Poly-Enrich using the enhancer locusdefs can be found in our recent study (details in reference 5). Like chipenrich, polyenrich is designed for narrow peaks, but for experiments with > ~50,000 genomic regions, or in cases where the number of binding sites per gene is thought to be important.

Poly-Enrich also allows weighting of individual genomic regions based on a score, which can be useful for differential methylation enrichment analysis and ChIP-seq. Currently the options are: `signalValue` and `logsignalValue`. `signalValue` weighs each genomic region or peak based on the Signal Value given in the narrowPeak format or a user-supplied column (column name should be `signalValue`), while `logsignalValue` takes the log of these values. To use weighted Poly-Enrich, use method = `polyenrich_weighted`. When using the `enhancer` or `enhancer_plus5kb` locus definition, we recommend selecting the `polyenrich_weighted` method, which in this case will automatically set multiAssign as `TRUE`.

```{r, warning = FALSE, message = FALSE}
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = polyenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, method="polyenrich_weighted", locusdef = "enhancer_plus5kb", qc_plots = FALSE, out_name = NULL, n_cores = 1)
results.pe = results$results
print(results.pe[1:5,1:5])
```

### `hybridenrich()`

The hybrid method is used when one is unsure of which method, between ChIP-Enrich or Poly-Enrich, is the optimal method, and the most statistically powerful results are desired for each gene set.  The runtime for this method, however, is ~2X that of the others.

The hybrid p-value is given as `2*min(chipenrich_pvalue, polyenrich_pvalue)`. This test will retain the same Type 1 level and will be a consistent test if one of chipenrich or polyenrich is consistent. This can be extended to any number of tests, but currently we only allow a hybrid test for chipenrich and polyenrich. For more information about chipenrich or polyenrich, see their respective sections.

```{r, warning = FALSE, message = FALSE}
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = hybridenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path, locusdef = "nearest_tss", qc_plots = F, out_name = NULL, n_cores = 1)
results.hybrid = results$results
print(results.hybrid[1:5,1:5])
```

### `proxReg()`

The proximity to regulatory regions (proxReg) test is a complementary test to any gene set enrichment test on a set of genomic regions, not just exclusive to the methods in this package. It tests if the genomic regions tend to be closer to (or farther from) gene transcription start sites or enhancer regions in each gene set tested. Currently, testing proximity to enhancer regions is only compatible with the hg19 genome. The purpose of ProxReg is to provide additional information for interpreting gene set enrichment test results, as a gene set enrichment test alone does not give information about whether the genomic regions occur near promoter or enhancer regions.

ProxReg first calculates the distance between the midpoints of peaks and the nearest transcription start site or the nearest enhancer region midpoint for each genomic region. Each genomic region is then assigned to the gene with the nearest transcription start site. The distances are then classified according to whether the gene is in the gene set or not, and a signed Wilcoxon rank-sum test is used to calculate if the regions are closer or farther in the gene set than average.

```{r, warning = FALSE, message = FALSE}
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results.prox = proxReg(peaks_E2F4, reglocation = 'tss',	genome = 'hg19', genesets=gs_path, out_name=NULL)
results.prox = results.prox$results
print(results.prox[1:5,1:7])
```

## QC Plots

Each enrich function outputs QC plots if `qc_plots = TRUE`. There are also stand-alone functions to make the QC plots without the need for GSE testing. The QC plots can be used to help determine which locus definition to use, or which enrichment method is more appropriate.

### Peak distance to TSS distribution

This plot gives a distribution of the distance of the peak midpoints to the TSSs. It can help in [selecting a locus definition](#selecting-a-locus-definition). For example, if genes are primarily within 5kb of TSSs, then the `5kb` locus definition may be a good choice. In contrast, if most genes fall far from TSSs, the `nearest_tss` locus definition may be a good choice.

```{r, fig.align='center', fig.cap='E2F4 peak distances to TSS', fig.height=6, fig.width=6, fig.show='hold', warning = FALSE, message = FALSE}
# Output in chipenrich and polyenrich
plot_dist_to_tss(peaks = peaks_E2F4, genome = 'hg19')
```

### Presence of peak versus locus length

This plot visualizes the relationship between the presence of at least one peak in a gene locus and the locus length (on the log10 scale). For clarity of visualization, each point represents 25 gene loci binned after sorting by locus length. The expected fit under the assumptions of Fisher's Exact Test (horizontal line), and a binomial-based test (gray curve) are displayed to indicate how the dataset being enriched conforms to the assumption of each. The empirical spline used in the `chipenrich` method is in orange.

```{r, fig.align='center', fig.cap='E2F4 chipenrich spline without mappability', fig.height=6, fig.width=6, fig.show='hold', warning = FALSE, message = FALSE}
# Output in chipenrich
plot_chipenrich_spline(peaks = peaks_E2F4, locusdef = 'nearest_tss', genome = 'hg19')
```

### Number of peaks versus locus length

This plot visualizes the relationship between the number of peaks assigned to a gene and the locus length (on the log10 scale). For clarity of visualization, each point represents 25 gene loci binned after sorting by locus length. The empirical spline used in the `polyenrich` method is in orange.

If many gene loci have multiple peaks assigned to them, `polyenrich` is likely an appropriate method. If there are a low number of peaks per gene, then `chipenrich()` may be the appropriate method.

```{r, fig.align='center', fig.cap='E2F4 polyenrich spline without mappability', fig.height=6, fig.width=6, fig.show='hold', warning = FALSE, message = FALSE}
# Output in polyenrich
plot_polyenrich_spline(peaks = peaks_E2F4, locusdef = 'nearest_tss', genome = 'hg19')
```

### Gene coverage versus locus length

This plot visualizes the relationship between proportion of the gene locus covered by peaks and the locus length (on the log10 scale). For clarity of visualization, each point represents 25 gene loci binned after sorting by locus length.

```{r, fig.align='center', fig.cap='H3K4me3 gene coverage', fig.height=6, fig.width=6, fig.show='hold', warning = FALSE, message = FALSE}
# Output in broadenrich
plot_gene_coverage(peaks = peaks_H3K4me3_GM12878, locusdef = 'nearest_tss',  genome = 'hg19')
```

## Output

The output of `broadenrich()`, `chipenrich()`, and `polyenrich()` is a list with components corresponding to each section below. If `out_name` is not `NULL`, then a file for each component will be written to the `out_path` with prefixes of `out_name`.

### Assigned peaks

Peak assignments are stored in `$peaks`. This is a peak-level summary. Each line corresponds to a peak intersecting a particular gene locus defined in the selected locus definition. In the case of `broadenrich()` peaks may be assigned to multiple gene loci. Doing `table()` on the `peak_id` column will indicate how many genes are assigned to each peak.

```{r}
head(results$peaks)
```

### Peaks-per-gene

Peak information aggregated over gene loci is stored in `$peaks_per_gene`. This is a gene-level summary. Each line corresponds to aggregated peak information over the `gene_id` such as the number of peaks assigned to the gene locus or the ratio of the gene locus covered in the case of `broadenrich()`.

```{r}
head(results$peaks_per_gene)
```

### Gene set enrichment results

GSE results are stored in `$results`. For convenience, gene set descriptions are provided in addition to the gene set ID (which is the same as the ID from the originating database). The `Status` column takes values of `enriched` if the `Effect` is \> 0 and `depleted` if \< 0, with `enriched` results being of primary importance. Finally, the `Geneset.Peak.Genes` column gives a list of gene IDs that had signal contributing to the test for enrichment. This list can be used to cross reference information from `$peaks` or `$peaks_per_gene` if desired.

```{r}
head(results$results)
```

## Assessing Type I Error with Randomizations

Randomization of locus definitions allows for the assessment of Type I Error under the null hypothesis of no true gene set enrichment. A well-calibrated Type I Error means that the false positive rate is controlled, and the p-values reported for actual data can be trusted. In both Welch & Lee, et al. and Cavalcante, et al., we demonstrated that both `chipenrich()` and `broadenrich()` have well-calibrated Type I Error over dozens of publicly available ENCODE ChIP-seq datasets. Unpublished data suggests the same is true for `polyenrich()`.

Within `chipenrich()`, `broadenrich()`, and `polyenrich()`, the `randomization` parameters can be used to assess the Type I Error for the data being analyzed.

The randomization codes, and their effects are:

* `NULL`: No randomizations, the default.
* `complete`: Shuffle the `gene_id` and `symbol` columns of the `locusdef` together, without regard for the chromosome location, or locus length. The null hypothesis is that there is no true gene set enrichment.
* `bylength`: Shuffle the `gene_id` and `symbol` columns of the `locusdef` together, within bins of 100 genes sorted by locus length. The null hypothesis is that there is no true gene set enrichment, but with preserved locus length relationship.
* `bylocation`: Shuffle the `gene_id` and `symbol` columns of the `locusdef` together, within bins of 50 genes sorted by genomic location. The null hypothesis is that there is no true gene set enrichment, but with preserved genomic location.

The return value of `chipenrich()`, `broadenrich()`, or `polyenrich()` with a selected randomization is the same list object described above. To assess the Type I error, the `alpha` level for the particular data set can be calculated by dividing the total number of gene sets with p-value < `alpha` by the total number of tests tested. Users may want to perform multiple randomizations for a set of peaks and take the median of the `alpha` values.

```{r, warning = FALSE, message = FALSE}
# Assessing if alpha = 0.05
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = chipenrich(peaks = peaks_E2F4, genome = 'hg19', genesets = gs_path,
	locusdef = "nearest_tss", qc_plots = FALSE, randomization = 'complete',
    out_name = NULL, n_cores = 1)
alpha = sum(results$results$P.value < 0.05) / nrow(results$results)
# NOTE: This is for
print(alpha)
```

# References

1. R.P. Welch^, C. Lee^, R.A. Smith, P. Imbriano, S. Patil, T. Weymouth, L.J. Scott, M.A. Sartor. "ChIP-Enrich: gene set enrichment testing for ChIP-seq data." Nucl. Acids Res. (2014) 42(13):e105. [doi:10.1093/nar/gku463](https://academic.oup.com/nar/article-lookup/doi/10.1093/nar/gku463)

2. R.G. Cavalcante, C. Lee, R.P. Welch, S. Patil, T. Weymouth, L.J. Scott, and M.A. Sartor. "Broad-Enrich: functional interpretation of large sets of broad genomic regions." Bioinformatics (2014) 30(17):i393-i400 [doi:10.1093/bioinformatics/btu444](https://academic.oup.com/bioinformatics/article-lookup/doi/10.1093/bioinformatics/btu444)

3. C.T. Lee , R.G. Cavalcante, C. Lee, T. Qin, S. Patil, S. Wang, Z. Tsai, A.P. Boyle, M.A. Sartor. "Poly-Enrich: Count-based Methods for Gene Set Enrichment Testing with Genomic Regions." NAR genomics and bioinformatics 2.1 (2020): lqaa006. [doi.org/10.1093/nargab/lqaa006
](https://academic.oup.com/nargab/article/2/1/lqaa006/5728474)

4. C.T. Lee, K. Wang, T. Qin, M.A. Sartor. "Testing proximity of genomic regions to transcription start sites and enhancers complements gene set enrichment testing." Frontiers in genetics 11 (2020): 199. [doi.org/10.3389/fgene.2020.00199](https://www.frontiersin.org/articles/10.3389/fgene.2020.00199/full)

5. T. Qin, C.T. Lee, R.G. Cavalcante, P. Orchard, H. Yao, H. Zhang, S. Wang, S. Patil, A.P. Boyle, M.A. Sartor, "Comprehensive enhancer-target gene assignments improve gene set level interpretation of genome-wide regulatory data." (2020) bioRxiv. [doi.org/10.1101/2020.10.22.351049](https://www.biorxiv.org/content/10.1101/2020.10.22.351049v1)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supported.R
\name{supported_genomes}
\alias{supported_genomes}
\title{Display supported genomes.}
\usage{
supported_genomes()
}
\value{
A vector indicating supported genomes.
}
\description{
Display supported genomes.
}
\examples{

supported_genomes()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign_peaks.R
\name{assign_peak_segments}
\alias{assign_peak_segments}
\title{Assign whole peaks to all overlapping defined gene loci.}
\usage{
assign_peak_segments(peaks, locusdef)
}
\arguments{
\item{peaks}{A \code{GRanges} object representing regions to be used for enrichment.}

\item{locusdef}{A locus definition object from \code{chipenrich.data}.}
}
\value{
A \code{data.frame} with columns for \code{peak_id, chr, peak_start, peak_end, gene_locus_start, gene_locus_end, gene_id, overlap_start, overlap_end, peak_overlap}. The result is used in \code{num_peaks_per_gene()}.
}
\description{
Determine all overlaps between the set of input regions \code{peaks} and the given locus definition \code{locusdef}. In addition, report where each overlap begins and ends, as well as the length of the overlap.
}
\details{
Typically, this function will not be used alone, but inside \code{chipenrich()} with \code{method = 'broadenrich'}.
}
\examples{

data('locusdef.hg19.nearest_tss', package = 'chipenrich.data')
data('tss.hg19', package = 'chipenrich.data')

file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
peaks = read_bed(file)

assigned_peaks = assign_peak_segments(
	peaks = peaks,
	locusdef = locusdef.hg19.nearest_tss)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_dist_to_tss.R
\name{plot_dist_to_tss}
\alias{plot_dist_to_tss}
\title{Plot histogram of distance from peak to nearest TSS}
\usage{
plot_dist_to_tss(peaks, genome = supported_genomes())
}
\arguments{
\item{peaks}{Either a file path or a \code{data.frame} of peaks in BED-like
format. If a file path, the following formats are fully supported via their
file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
without these extensions are supported under the conditions that the first 3
columns correspond to 'chr', 'start', and 'end' and that there is either no
header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
\code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
acceptable column names.}

\item{genome}{One of the \code{supported_genomes()}.}
}
\value{
A trellis plot object.
}
\description{
Create a histogram of the distance from each peak to the nearest transcription
start site (TSS) of any gene.
}
\examples{

# Create histogram of distance from peaks to nearest TSS.
data(peaks_E2F4, package = 'chipenrich.data')
peaks_E2F4 = subset(peaks_E2F4, peaks_E2F4$chrom == 'chr1')
plot_dist_to_tss(peaks_E2F4, genome = 'hg19')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{post_process_enrichments}
\alias{post_process_enrichments}
\title{Post process the \code{data.frame} of enrichment results}
\usage{
post_process_enrichments(enrich)
}
\arguments{
\item{enrich}{A \code{data.frame} of the enrichment results from \code{broadenrich()},
\code{chipenrich()}, or \code{polyenrich()} created by \code{rbind}ing the
list of enrichment results for each of the \code{genesets}.}
}
\value{
A reformatted \code{data.frame} with columns in a specific order, filtered
of enrichment tests that failed, and ordered first by enrichment 'Status' (if
present) and then 'P.value'.
}
\description{
Post process the \code{data.frame} of enrichment results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{postprocess_peak_grs}
\alias{postprocess_peak_grs}
\title{A helper function to post-process peak GRanges}
\usage{
postprocess_peak_grs(gr)
}
\arguments{
\item{gr}{A \code{GRanges} of input peaks.}
}
\value{
A \code{GRanges} that is sorted if the \code{seqinfo} is set, and has named peaks.
}
\description{
Check for overlapping input regions, sort peaks, and force peak names
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supported.R
\name{supported_genesets}
\alias{supported_genesets}
\title{Display supported genesets for gene set enrichment.}
\usage{
supported_genesets()
}
\value{
A \code{data.frame} with columns \code{geneset, organism}.
}
\description{
Display supported genesets for gene set enrichment.
}
\examples{

supported_genesets()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{setup_locusdef}
\alias{setup_locusdef}
\title{Function to setup locus definitions}
\usage{
setup_locusdef(ldef_code, genome, randomization = NULL)
}
\arguments{
\item{ldef_code}{One of 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
'1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
'10kb', '10kb_outside', '10kb_outside_upstream'. Alternately, a file path for
a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
must have columns 'chr', 'start', 'end', and 'gene_id', or 'geneid'.}

\item{genome}{One of the \code{supported_genomes()}.}

\item{randomization}{One of \code{NULL}, 'complete', 'bylength', or 'bylocation'.
See the Randomizations section in \code{?chipenrich}. Default NULL.}
}
\value{
A list with components \code{ldef} and \code{tss}.
}
\description{
Function to setup locus definitions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{read_ldef}
\alias{read_ldef}
\title{Function to read custom locus definition from file}
\usage{
read_ldef(file_path, genome = NA)
}
\arguments{
\item{file_path}{A file path for the custom locus definition.}

\item{genome}{A genome from \code{supported_genomes()}, default \code{NA}.}
}
\value{
A \code{LocusDefinition} class object with slots \code{dframe},
\code{granges}, \code{genome.build}, and \code{organism}.
}
\description{
This function reads a tab-delimited text (with a header) file that should have
columns 'chr', 'start', 'end', and a column named 'gene_id' (or 'geneid') with the
Entrez Gene ID. If a \code{supported_genomes()} is given, then a column
of gene symbols named 'symbol', will be added. If an unsupported genome is
used there are two options: 1) Have a column named 'symbols' with the gene
symbols in the custom locus definition, and leave \code{genome = NA}, or 2)
leave \code{genome = NA}, do not provide gene symbols, and NAs will be used.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{read_mappa}
\alias{read_mappa}
\title{Function to read custom mappability files}
\usage{
read_mappa(file_path)
}
\arguments{
\item{file_path}{A file path for the custom mappability.}
}
\value{
A \code{data.frame} containing \code{gene_id} and \code{mappa} columns.
}
\description{
This function reads a two-columned tab-delimited text file (with header). Expected
column names are 'mappa' and 'gene_id'. Each line is for a unique 'gene_id'
and contains the mappability (between 0 and 1) for that gene.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/peaks_per_gene.R
\name{num_peaks_per_gene}
\alias{num_peaks_per_gene}
\title{Aggregate peak assignments over the \code{gene_id} column}
\usage{
num_peaks_per_gene(assigned_peaks, locusdef, mappa = NULL)
}
\arguments{
\item{assigned_peaks}{A \code{data.frame} resulting from \code{assign_peaks()} or \code{assign_peak_segments()}.}

\item{locusdef}{A locus definition object from \code{chipenrich.data}.}

\item{mappa}{A mappability object from \code{chipenrich.data}.}
}
\value{
A \code{data.frame} with columns \code{gene_id, length, log10_length, num_peaks, peak}. The result is used directly in the gene set enrichment tests in \code{chipenrich()}.
}
\description{
For each \code{gene_id}, determine the locus length and the number of peaks.
}
\details{
Typically, this function will not be used alone, but inside \code{chipenrich()}.
}
\examples{

data('locusdef.hg19.nearest_tss', package = 'chipenrich.data')
data('tss.hg19', package = 'chipenrich.data')

file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
peaks = read_bed(file)

assigned_peaks = assign_peaks(
	peaks = peaks,
	locusdef = locusdef.hg19.nearest_tss,
	tss = tss.hg19)

ppg = num_peaks_per_gene(
	assigned_peaks = assigned_peaks,
	locusdef = locusdef.hg19.nearest_tss,
	mappa = NULL)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{recode_peaks}
\alias{recode_peaks}
\title{Recode a vector of number of peaks to binary based on threshold}
\usage{
recode_peaks(num_peaks, threshold = 1)
}
\arguments{
\item{num_peaks}{An \code{integer} vector representing numbers of peaks per gene.}

\item{threshold}{An \code{integer} specifying the minimum number of peaks
required to code as 1.}
}
\value{
An binary vector where an entry is 1 if the corresponding entry of
\code{num_peaks} is >= \code{threshold} and is otherwise 0.
}
\description{
Recode a vector of number of peaks to binary based on threshold
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{read_bed}
\alias{read_bed}
\title{Read files containing peaks or genomic regions}
\usage{
read_bed(file_path)
}
\arguments{
\item{file_path}{A path to a file with input peaks/regions. See extended
description above for details about file support.}
}
\value{
A \code{GRanges} with \code{mcols} matching any extra columns.
}
\description{
The following formats are fully supported via their file extensions: .bed,
.broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph or .bdg. BED3 through
BED6 files are supported under the .bed extension. Files without these extensions
are supported under the conditions that the first 3 columns correspond to
chr, start, and end and that there is either no header column, or it is
commented out. Files may be compressed with gzip, and so might end in .narrowPeak.gz,
for example. For files with extension support, the \code{rtracklayer::import()}
function is used to read peaks, so adherence to the mentioned file formats is
necessary.
}
\details{
NOTE: Header rows must be commented with \code{#} to be ignored. Otherwise,
an error may result.

NOTE: A warning is given if any input regions overlap. In the case of enrichment
testing with \code{method = 'broadenrich'}, regions should be disjoint.

Typically, this function will not be used alone, but inside \code{chipenrich()}.
}
\examples{

# Example of generic .txt file with peaks
file = system.file('extdata', 'test_header.txt', package = 'chipenrich')
peaks = read_bed(file)

# Example of BED3
file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
peaks = read_bed(file)

# Example of narrowPeak
file = system.file('extdata', 'test.narrowPeak', package = 'chipenrich')
peaks = read_bed(file)

# Example of gzipped broadPeak
file = system.file('extdata', 'test.broadPeak.gz', package = 'chipenrich')
peaks = read_bed(file)

# Example of gzipped gff3 Fly peaks
file = system.file('extdata', 'test.gff3.gz', package = 'chipenrich')
peaks = read_bed(file)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_test_method}
\alias{get_test_method}
\title{Get the test function name from the method name}
\usage{
get_test_method(method)
}
\arguments{
\item{method}{A character for the method used. One of the \code{supported_methods}
or one of the \code{HIDDEN_METHODS} in \code{constants.R}.}
}
\value{
A singleton named character vector with value of the test function and
name of the method.
}
\description{
The \code{method} comes from what is used in \code{chipenrich()} or in
\code{polyenrich()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/peaks2genes.R
\name{peaks2genes}
\alias{peaks2genes}
\title{Run the test process up to, but not including the enrichment tests.}
\usage{
peaks2genes(
  peaks,
  out_name = "readyToEnrich",
  out_path = getwd(),
  genome = supported_genomes(),
  locusdef = "nearest_tss",
  weighting = NULL,
  mappability = NULL,
  qc_plots = TRUE,
  num_peak_threshold = 1,
  randomization = NULL
)
}
\arguments{
\item{peaks}{Either a file path or a \code{data.frame} of peaks in BED-like
format. If a file path, the following formats are fully supported via their
file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
without these extensions are supported under the conditions that the first 3
columns correspond to 'chr', 'start', and 'end' and that there is either no
header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
\code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
acceptable column names.}

\item{out_name}{Prefix string to use for naming output files. This should not
contain any characters that would be illegal for the system being used (Unix,
Windows, etc.) The default value is "polyenrich", and a file "polyenrich_results.tab"
is produced. If \code{qc_plots} is set, then a file "polyenrich_qcplots.pdf"
is produced containing a number of quality control plots. If \code{out_name}
is set to NULL, no files are written, and results then must be retrieved from
the list returned by \code{polyenrich}.}

\item{out_path}{Directory to which results files will be written out. Defaults
to the current working directory as returned by \code{\link{getwd}}.}

\item{genome}{One of the \code{supported_genomes()}.}

\item{locusdef}{One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
'1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
'10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
example custom locus definition file, see the vignette.}

\item{weighting}{(Poly-Enrich only) character string specifying the weighting method if method is
chosen to be 'polyenrich_weighted'. Current options are: 'signalValue',
'logsignalValue', and 'multiAssign'.}

\item{mappability}{One of \code{NULL}, a file path to a custom mappability file,
or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
For an example custom mappability file, see the vignette. Default value is NULL.}

\item{qc_plots}{A logical variable that enables the automatic generation of
plots for quality control.}

\item{num_peak_threshold}{(ChIP-Enrich only) Sets the threshold for how many peaks a gene must
have to be considered as having a peak. Defaults to 1. Only relevant for
Fisher's exact test and ChIP-Enrich methods.}

\item{randomization}{One of \code{NULL}, 'complete', 'bylength', or 'bylocation'.
See the Randomizations section below.}
}
\value{
A list, containing the following items:

\item{opts }{A data frame containing the arguments/values passed to \code{polyenrich}.}

\item{peaks }{
A data frame containing peak assignments to genes. Peaks which do not overlap
a gene locus are not included. Each peak that was assigned to a gene is listed,
along with the peak midpoint or peak interval coordinates (depending on which
was used), the gene to which the peak was assigned, the locus start and end
position of the gene, and the distance from the peak to the TSS.

The columns are:

\describe{
  \item{peak_id}{ is an ID given to unique combinations of chromosome, peak start, and peak end. }
  \item{chr}{ is the chromosome the peak originated from. }
  \item{peak_start}{ is start position of the peak. }
  \item{peak_end}{ is end position of the peak. }
  \item{peak_midpoint}{ is the midpoint of the peak. }
  \item{gene_id}{ is the Entrez ID of the gene to which the peak was assigned. }
  \item{gene_symbol}{ is the official gene symbol for the gene_id (above). }
  \item{gene_locus_start}{ is the start position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
  \item{gene_locus_end}{ is the end position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
  \item{nearest_tss}{ is the closest TSS to this peak (for any gene, not necessarily the gene this peak was assigned to.) }
  \item{nearest_tss_gene}{ is the gene having the closest TSS to the peak (should be the same as gene_id when using the nearest TSS locus definition.) }
  \item{nearest_tss_gene_strand}{ is the strand of the gene with the closest TSS. }
}}

\item{peaks_per_gene }{
A data frame of the count of peaks per gene. The columns are:

\describe{
  \item{gene_id}{ is the Entrez Gene ID. }
  \item{length}{ is the length of the gene's locus (depending on which locus definition you chose.)}
  \item{log10_length}{ is the log10(locus length) for the gene.}
  \item{num_peaks}{ is the number of peaks that were assigned to the gene, given the current locus definition. }
  \item{peak}{ is whether or not the gene has a peak. }
}}
}
\description{
This function is used to create the *_peaks and *_peaks-per-gene files
This way one does not need to remake these files whenever one just wants
to test enrichment methods.
}
\section{Randomizations}{

Randomization of locus definitions allows for the assessment of Type I Error
under the null hypothesis. The randomization codes are:
\describe{
\item{\code{NULL}:}{ No randomizations, the default.}
	\item{'complete':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
\code{locusdef} together, without regard for the chromosome location, or locus length.
The null hypothesis is that there is no true gene set enrichment.}
	\item{'bylength':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
\code{locusdef} together within bins of 100 genes sorted by locus length. The null
hypothesis is that there is no true gene set enrichment, but with preserved locus
length relationship.}
	\item{'bylocation':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
\code{locusdef} together within bins of 50 genes sorted by genomic location. The null
hypothesis is that there is no true gene set enrichment, but with preserved
genomic location.}
}
The return value with a selected randomization is the same list as without.
To assess the Type I error, the \code{alpha} level for the particular data set
can be calculated by dividing the total number of gene sets with p-value < \code{alpha}
by the total number of tests. Users may want to perform multiple randomizations
for a set of peaks and take the median of the \code{alpha} values.
}

\section{Poly-Enrich Weighting Options}{

Poly-Enrich also allows weighting of individual peaks. Currently the options are:
\describe{
 \item{'signalValue:'}{ weighs each peak based on the Signal Value given in the
narrowPeak format or a user-supplied column, normalized to have mean 1.}
 \item{'logsignalValue:'}{ weighs each peak based on the log Signal Value given in the
narrowPeak format or a user-supplied column, normalized to have mean 1.}
 \item{'multiAssign:'}{ weighs each peak by the inverse of the number of genes
it is assigned to.}
}
}

\examples{

# Run peaks2genes using an example dataset, assigning peaks to the nearest TSS
data(peaks_E2F4, package = 'chipenrich.data')
peaks_E2F4 = subset(peaks_E2F4, peaks_E2F4$chrom == 'chr1')
gs_path = system.file('extdata', package='chipenrich')
results = peaks2genes(peaks_E2F4, locusdef='nearest_tss',
			genome = 'hg19', out_name=NULL)

# Get the list of peaks that were assigned to genes.
assigned_peaks = results$peaks

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{setup_genesets}
\alias{setup_genesets}
\title{Function to setup genesets}
\usage{
setup_genesets(gs_codes, ldef_obj, genome, min_geneset_size, max_geneset_size)
}
\arguments{
\item{gs_codes}{A character vector of geneset databases to be tested for
enrichment. See \code{supported_genesets()}. Alternately, a file path to a
a tab-delimited text file with header and first column being the geneset ID
or name, and the second column being Entrez Gene IDs.}

\item{ldef_obj}{A \code{LocusDefinition} object to use for filtering gene sets
based on which genes are defined in the locus defintion.}

\item{genome}{One of the \code{supported_genomes()}.}

\item{min_geneset_size}{Sets the minimum number of genes a gene set may have
to be considered for enrichment testing.}

\item{max_geneset_size}{Sets the maximum number of genes a gene set may have
to be considered for enrichment testing.}
}
\value{
A list with components consisting of \code{GeneSet} objects for each
of the elements of \code{genesets}. NOTE: Custom genesets must be run separately
from built in gene sets.
}
\description{
Function to setup genesets
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{genome_to_orgdb}
\alias{genome_to_orgdb}
\title{Get Entrez ID to gene symbol mappings for custom locus definitions}
\usage{
genome_to_orgdb(genome = supported_genomes())
}
\arguments{
\item{genome}{One of the \code{supported_genomes()}.}
}
\value{
A \code{data.frame} with \code{gene_id} and \code{symbol} columns.
}
\description{
If a custom locus definition is one of the \code{supported_genomes()}, then
the gene symbol column of the custom locus definition is populated using the
appropriate orgDb package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supported.R
\name{supported_methods}
\alias{supported_methods}
\title{Display supported gene set enrichment methods.}
\usage{
supported_methods()
}
\value{
A vector indicating supported methods for gene set enrichment.
}
\description{
Display supported gene set enrichment methods.
}
\examples{

supported_methods()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{reset_ncores_for_windows}
\alias{reset_ncores_for_windows}
\title{Reset n_cores for Windows}
\usage{
reset_ncores_for_windows(n_cores)
}
\arguments{
\item{n_cores}{An \code{integer} passed to \code{broadenrich()},
\code{chipenrich()}, or \code{polyenrich()} indicating the number of cores
to use for enrichment testing.}
}
\value{
Either the original \code{n_cores} if the OS is not Windows, or 1 if
the OS is Windows.
}
\description{
We use \code{parallel::mclapply} for multicore geneset enrichment testing, but
this function supports more than one core if the OS is not Windows. If the OS
is windows, the number of cores (\code{mc.cores}) must be set to 1.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polyenrich.R
\name{polyenrich}
\alias{polyenrich}
\title{Run Poly-Enrich on narrow genomic regions}
\usage{
polyenrich(
  peaks,
  out_name = "polyenrich",
  out_path = getwd(),
  genome = supported_genomes(),
  genesets = c("GOBP", "GOCC", "GOMF"),
  locusdef = "nearest_tss",
  method = "polyenrich",
  multiAssign = NULL,
  weighting = NULL,
  mappability = NULL,
  qc_plots = TRUE,
  min_geneset_size = 15,
  max_geneset_size = 2000,
  randomization = NULL,
  n_cores = 1
)
}
\arguments{
\item{peaks}{Either a file path or a \code{data.frame} of peaks in BED-like
format. If a file path, the following formats are fully supported via their
file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
without these extensions are supported under the conditions that the first 3
columns correspond to 'chr', 'start', and 'end' and that there is either no
header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
\code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
acceptable column names.}

\item{out_name}{Prefix string to use for naming output files. This should not
contain any characters that would be illegal for the system being used (Unix,
Windows, etc.) The default value is "polyenrich", and a file "polyenrich_results.tab"
is produced. If \code{qc_plots} is set, then a file "polyenrich_qcplots.pdf"
is produced containing a number of quality control plots. If \code{out_name}
is set to NULL, no files are written, and results then must be retrieved from
the list returned by \code{polyenrich}.}

\item{out_path}{Directory to which results files will be written out. Defaults
to the current working directory as returned by \code{\link{getwd}}.}

\item{genome}{One of the \code{supported_genomes()}.}

\item{genesets}{A character vector of geneset databases to be tested for
enrichment. See \code{supported_genesets()}. Alternately, a file path to a
a tab-delimited text file with header and first column being the geneset ID
or name, and the second column being Entrez Gene IDs. For an example custom
gene set file, see the vignette.}

\item{locusdef}{One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
'1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
'10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
example custom locus definition file, see the vignette.}

\item{method}{A character string specifying the method to use for enrichment
testing. Current options are \code{polyenrich} and \code{polyenrich_weighted}.}

\item{multiAssign}{A boolean value for performing a multiassignment of peaks.
Default is \code{NULL}. When selecting \code{polyenrich_weighted} method and
\code{locusdef} equals to \code{enhancer} or \code{enhancer_plus5kb}, this
\code{multiAssign} will be automatically set as \code{TRUE}.}

\item{weighting}{A character string specifying the weighting method if method is
chosen to be 'polyenrich_weighted'. Current options are: 'signalValue' and
'logsignalValue'.}

\item{mappability}{One of \code{NULL}, a file path to a custom mappability file,
or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
For an example custom mappability file, see the vignette. Default value is NULL.}

\item{qc_plots}{A logical variable that enables the automatic generation of
plots for quality control.}

\item{min_geneset_size}{Sets the minimum number of genes a gene set may have
to be considered for enrichment testing.}

\item{max_geneset_size}{Sets the maximum number of genes a gene set may have
to be considered for enrichment testing.}

\item{randomization}{One of \code{NULL}, 'complete', 'bylength', or 'bylocation'.
See the Randomizations section below.}

\item{n_cores}{The number of cores to use for enrichment testing. We recommend
using only up to the maximum number of \emph{physical} cores present, as
virtual cores do not significantly decrease runtime. Default number of cores
is set to 1. NOTE: Windows does not support multicore enrichment.}
}
\value{
A list, containing the following items:

\item{opts }{A data frame containing the arguments/values passed to \code{polyenrich}.}

\item{peaks }{
A data frame containing peak assignments to genes. Peaks which do not overlap
a gene locus are not included. Each peak that was assigned to a gene is listed,
along with the peak midpoint or peak interval coordinates (depending on which
was used), the gene to which the peak was assigned, the locus start and end
position of the gene, and the distance from the peak to the TSS.

The columns are:

\describe{
  \item{peak_id}{ is an ID given to unique combinations of chromosome, peak start, and peak end. }
  \item{chr}{ is the chromosome the peak originated from. }
  \item{peak_start}{ is start position of the peak. }
  \item{peak_end}{ is end position of the peak. }
  \item{peak_midpoint}{ is the midpoint of the peak. }
  \item{gene_id}{ is the Entrez ID of the gene to which the peak was assigned. }
  \item{gene_symbol}{ is the official gene symbol for the gene_id (above). }
  \item{gene_locus_start}{ is the start position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
  \item{gene_locus_end}{ is the end position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
  \item{nearest_tss}{ is the closest TSS to this peak (for any gene, not necessarily the gene this peak was assigned to.) }
  \item{nearest_tss_gene}{ is the gene having the closest TSS to the peak (should be the same as gene_id when using the nearest TSS locus definition.) }
  \item{nearest_tss_gene_strand}{ is the strand of the gene with the closest TSS. }
}}

\item{peaks_per_gene }{
A data frame of the count of peaks per gene. The columns are:

\describe{
  \item{gene_id}{ is the Entrez Gene ID. }
  \item{length}{ is the length of the gene's locus (depending on which locus definition you chose.)}
  \item{log10_length}{ is the log10(locus length) for the gene.}
  \item{num_peaks}{ is the number of peaks that were assigned to the gene, given the current locus definition. }
  \item{peak}{ is whether or not the gene has a peak. }
}}

\item{results }{
A data frame of the results from performing the gene set enrichment test on
each geneset that was requested (all genesets are merged into one final data
frame.) The columns are:

\describe{
  \item{Geneset.ID}{ is the identifier for a given gene set from the selected database.  For example, GO:0000003. }
  \item{Geneset.Type}{ specifies from which database the Geneset.ID originates.  For example, "Gene Ontology Biological Process."}
  \item{Description}{ gives a definition of the geneset. For example, "reproduction."}
  \item{P.Value}{ is the probability of observing the degree of enrichment of the gene set given the null hypothesis that peaks are not associated with any gene sets.}
  \item{FDR}{ is the false discovery rate proposed by Bejamini \& Hochberg for adjusting the p-value to control for family-wise error rate.}
  \item{Odds.Ratio}{ is the estimated odds that peaks are associated with a given gene set compared to the odds that peaks are associated with other gene sets, after controlling for locus length and/or mappability.  An odds ratio greater than 1 indicates enrichment, and less than 1 indicates depletion.}
  \item{N.Geneset.Genes}{ is the number of genes in the gene set.}
  \item{N.Geneset.Peak.Genes}{ is the number of genes in the genes set that were assigned at least one peak.}
  \item{Geneset.Avg.Gene.Length}{ is the average length of the genes in the gene set.}
  \item{Geneset.Peak.Genes}{ is the list of genes from the gene set that had at least one peak assigned.}

}}
}
\description{
Poly-Enrich is also designed for narrow peaks, for experiments with 100,000s of
peaks, or in cases where the number of binding sites per gene affects its regulation.
If unsure whether to use chipenrich or polyenrich, then we recommend hybridenrich.
For more details, see the 'Poly-Enrich Method' section below. For help choosing a method,
see the 'Choosing A Method' section below, or see the vignette.
}
\section{Poly-Enrich Method}{

The Poly-Enrich method uses the number of peaks in genes in its model for
enrichment: \code{num_peaks ~ GO + s(log10_length)}. Here, \code{GO} is a
binary vector indicating whether a gene is in the gene set being tested,
\code{num_peaks} is a numeric vector indicating the number of peaks in each
gene, and \code{s(log10_length)} is a negative binomial cubic smoothing spline
which adjusts for the relationship between the number of peaks in a gene and
locus length.
}

\section{Poly-Enrich Weighting Options}{

Poly-Enrich also allows weighting of individual peaks. Currently the options are:
\describe{
 \item{'signalValue:'}{ weighs each peak based on the Signal Value given in the
narrowPeak format or a user-supplied column, normalized to have mean 1.}
 \item{'logsignalValue:'}{ weighs each peak based on the log Signal Value given in the
narrowPeak format or a user-supplied column, normalized to have mean 1.}
}
}

\section{Choosing A Method}{

The following guidelines are intended to help select an enrichment function:
\describe{
\item{broadenrich():}{ is designed for use with broad peaks that may intersect
multiple gene loci, and cumulatively cover greater than 5\% of the genome. For
example, ChIP-seq experiments for histone modifications.}
\item{chipenrich():}{ is designed for use with 1,000s or 10,000s of narrow
peaks which results in fewer gene loci containing a peak overall. For example,
ChIP-seq experiments for transcription factors.}
\item{polyenrich():}{ is also designed for narrow peaks, for experiments with
100,000s of peaks, or in cases where the number of binding sites per gene affects
its regulation. If unsure whether to use chipenrich or polyenrich, then we recommend
hybridenrich.}
 \item{hybridenrich():}{ is a combination of chipenrich and polyenrich, to be
used when one is unsure which is the optimal method. }
}
}

\section{Randomizations}{

Randomization of locus definitions allows for the assessment of Type I Error
under the null hypothesis. The randomization codes are:
\describe{
\item{\code{NULL}:}{ No randomizations, the default.}
	\item{'complete':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
\code{locusdef} together, without regard for the chromosome location, or locus length.
The null hypothesis is that there is no true gene set enrichment.}
	\item{'bylength':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
\code{locusdef} together within bins of 100 genes sorted by locus length. The null
hypothesis is that there is no true gene set enrichment, but with preserved locus
length relationship.}
	\item{'bylocation':}{ Shuffle the \code{gene_id} and \code{symbol} columns of the
\code{locusdef} together within bins of 50 genes sorted by genomic location. The null
hypothesis is that there is no true gene set enrichment, but with preserved
genomic location.}
}
The return value with a selected randomization is the same list as without.
To assess the Type I error, the \code{alpha} level for the particular data set
can be calculated by dividing the total number of gene sets with p-value < \code{alpha}
by the total number of tests. Users may want to perform multiple randomizations
for a set of peaks and take the median of the \code{alpha} values.
}

\examples{

# Run Poly-Enrich using an example dataset, assigning peaks to the nearest TSS,
# and on a small custom geneset
data(peaks_E2F4, package = 'chipenrich.data')
peaks_E2F4 = subset(peaks_E2F4, peaks_E2F4$chrom == 'chr1')
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = polyenrich(peaks_E2F4, method='polyenrich', locusdef='nearest_tss',
			genome = 'hg19', genesets=gs_path, out_name=NULL)

# Get the list of peaks that were assigned to genes.
assigned_peaks = results$peaks

# Get the results of enrichment testing.
enrich = results$results

}
\seealso{
Other enrichment functions: 
\code{\link{broadenrich}()},
\code{\link{chipenrich}()}
}
\concept{enrichment functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hybrid.R
\name{hybridenrich}
\alias{hybridenrich}
\title{Running Hybrid test, either from scratch or using two results files}
\usage{
hybridenrich(
  peaks,
  out_name = "hybridenrich",
  out_path = getwd(),
  genome = supported_genomes(),
  genesets = c("GOBP", "GOCC", "GOMF"),
  locusdef = "nearest_tss",
  methods = c("chipenrich", "polyenrich"),
  weighting = NULL,
  mappability = NULL,
  qc_plots = TRUE,
  min_geneset_size = 15,
  max_geneset_size = 2000,
  num_peak_threshold = 1,
  randomization = NULL,
  n_cores = 1
)
}
\arguments{
\item{peaks}{Either a file path or a \code{data.frame} of peaks in BED-like
format. If a file path, the following formats are fully supported via their
file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
without these extensions are supported under the conditions that the first 3
columns correspond to 'chr', 'start', and 'end' and that there is either no
header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
\code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
acceptable column names.}

\item{out_name}{Prefix string to use for naming output files. This should not
contain any characters that would be illegal for the system being used (Unix,
Windows, etc.) The default value is "chipenrich", and a file "chipenrich_results.tab"
is produced. If \code{qc_plots} is set, then a file "chipenrich_qcplots.pdf"
is produced containing a number of quality control plots. If \code{out_name}
is set to NULL, no files are written, and results then must be retrieved from
the list returned by \code{chipenrich}.}

\item{out_path}{Directory to which results files will be written out. Defaults
to the current working directory as returned by \code{\link{getwd}}.}

\item{genome}{One of the \code{supported_genomes()}.}

\item{genesets}{A character vector of geneset databases to be tested for
enrichment. See \code{supported_genesets()}. Alternately, a file path to a
a tab-delimited text file with header and first column being the geneset ID
or name, and the second column being Entrez Gene IDs. For an example custom
gene set file, see the vignette.}

\item{locusdef}{One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
'1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
'10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
example custom locus definition file, see the vignette.}

\item{methods}{A character string array specifying the method to use for enrichment
testing. Currently actually unused as the methods are forced to be one chipenrich
and one polyenrich.}

\item{weighting}{A character string specifying the weighting method. Method name will
automatically be "polyenrich_weighted" if given weight options. Current options are:
'signalValue', 'logsignalValue', and 'multiAssign'.}

\item{mappability}{One of \code{NULL}, a file path to a custom mappability file,
or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
For an example custom mappability file, see the vignette. Default value is NULL.}

\item{qc_plots}{A logical variable that enables the automatic generation of
plots for quality control.}

\item{min_geneset_size}{Sets the minimum number of genes a gene set may have
to be considered for enrichment testing.}

\item{max_geneset_size}{Sets the maximum number of genes a gene set may have
to be considered for enrichment testing.}

\item{num_peak_threshold}{Sets the threshold for how many peaks a gene must
have to be considered as having a peak. Defaults to 1. Only relevant for
Fisher's exact test and ChIP-Enrich methods.}

\item{randomization}{One of \code{NULL}, 'complete', 'bylength', or 'bylocation'.
See the Randomizations section below.}

\item{n_cores}{The number of cores to use for enrichment testing. We recommend
using only up to the maximum number of \emph{physical} cores present, as
virtual cores do not significantly decrease runtime. Default number of cores
is set to 1. NOTE: Windows does not support multicore enrichment.}
}
\value{
A data.frame containing:
\item{results }{
A data frame of the results from performing the gene set enrichment test on
each geneset that was requested (all genesets are merged into one final data
frame.) The columns are:

\describe{
  \item{Geneset.ID}{ is the identifier for a given gene set from the selected database.  For example, GO:0000003. }
  \item{P.Value.x}{ is the probability of observing the degree of enrichment of the gene set given the null hypothesis
                    that peaks are not associated with any gene sets, for the first test.}
  \item{P.Value.y}{ is the same as above except for the second test.}
  \item{P.Value.Hybrid}{ The calculated Hybrid p-value from the two tests}
  \item{FDR.Hybrid}{ is the false discovery rate proposed by Bejamini \& Hochberg for adjusting the p-value to control for family-wise error rate.}
 Other variables given will also be included, see the corresponding methods' documentation for their details.
}}
}
\description{
Hybrid test is designed for people unsure of which test between ChIP-Enrich
and Poly-Enrich to use, so it takes information of both and gives adjusted
P-values. For more about ChIP- and Poly-Enrich, consult their corresponding
documentation.
}
\section{Hybrid p-values}{

Given n tests that test for the same hypothesis, same Type I error rate, and
converted to p-values: \code{p_1, ..., p_n}, the Hybrid p-value is computed as:
\code{n*min(p_1, ..., p_n)}. This hybrid test will have at most the same
Type I error as any individual test, and if any of the tests have 100\% power as
sample size goes to infinity, then so will the hybrid test.
}

\section{Function inputs}{

Every input in hybridenrich is the same as in chipenrich and polyenrich. Inputs
unique to chipenrich are: num_peak_threshold; and inputs unique to polyenrich are:
weighting. Currently the test only supports running chipenrich and polyenrich, but
future plans will allow you to run any number of different support tests.
}

\section{Joining two results files}{

Combines two existing results files and returns one results file with hybrid
p-values and FDR included. Current allowed inputs are objects from any of
the supplied enrichment tests or a dataframe with at least the following columns:
\code{P.value, Geneset.ID}. Optional columns include: \code{Status}. Currently
we only allow for joining two results files, but future plans will allow you to join
any number of results files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/peaks_per_gene.R
\name{calc_peak_gene_overlap}
\alias{calc_peak_gene_overlap}
\title{Add peak overlap and ratio to result of \code{num_peaks_per_gene()}}
\usage{
calc_peak_gene_overlap(assigned_peaks, ppg)
}
\arguments{
\item{assigned_peaks}{A \code{data.frame} resulting from \code{assign_peak_segments()}.}

\item{ppg}{The aggregated peak assignments over \code{gene_id} from \code{num_peaks_per_gene()}.}
}
\value{
A \code{data.frame} with columns \code{gene_id, length, log10_length, num_peaks, peak, peak_overlap, ratio}. The result is used directly in the gene set enrichment tests in \code{chipenrich()} when \code{method = 'broadenrich'}.
}
\description{
In particular, for \code{method = 'broadenrich'} in \code{chipenrich()}, when using \code{assign_peak_segments()}. This function will add aggregated \code{peak_overlap} (in base pairs) and \code{ratio} (relative to \code{length}) columns to the result of \code{num_peaks_per_gene()} so the right data is present for the \code{method = 'broadenrich'} model.
}
\details{
Typically, this function will not be used alone, but inside \code{chipenrich()} with \code{method = 'broadenrich'}.
}
\examples{

data('locusdef.hg19.nearest_tss', package = 'chipenrich.data')
data('tss.hg19', package = 'chipenrich.data')

file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
peaks = read_bed(file)

assigned_peaks = assign_peak_segments(
	peaks = peaks,
	locusdef = locusdef.hg19.nearest_tss)

ppg = num_peaks_per_gene(
	assigned_peaks = assigned_peaks,
	locusdef = locusdef.hg19.nearest_tss,
	mappa = NULL)

ppg = calc_peak_gene_overlap(
	assigned_peaks = assigned_peaks,
	ppg = ppg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_chipenrich_spline.R
\name{plot_chipenrich_spline}
\alias{plot_chipenrich_spline}
\title{QC plot for ChIP-Enrich}
\usage{
plot_chipenrich_spline(
  peaks,
  locusdef = "nearest_tss",
  genome = supported_genomes(),
  mappability = NULL,
  legend = TRUE,
  xlim = NULL
)
}
\arguments{
\item{peaks}{Either a file path or a \code{data.frame} of peaks in BED-like
format. If a file path, the following formats are fully supported via their
file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
without these extensions are supported under the conditions that the first 3
columns correspond to 'chr', 'start', and 'end' and that there is either no
header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
\code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
acceptable column names.}

\item{locusdef}{One of: 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
'1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
'10kb', '10kb_outside', '10kb_outside_upstream'. For a description of each,
see the vignette or \code{\link{supported_locusdefs}}. Alternately, a file path for
a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
must have columns 'chr', 'start', 'end', and 'gene_id' or 'geneid'. For an
example custom locus definition file, see the vignette.}

\item{genome}{One of the \code{supported_genomes()}.}

\item{mappability}{One of \code{NULL}, a file path to a custom mappability file,
or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
For an example custom mappability file, see the vignette. Default value is NULL.}

\item{legend}{If true, a legend will be drawn on the plot.}

\item{xlim}{Set the x-axis limit. NULL means select x-lim automatically.}
}
\value{
A trellis plot object.
}
\description{
A plot showing an approximation of the empirical spline used to model the
relationship between a gene having a peak and the locus length. For visual
clarity, genes are binned into groups of 25 after sorting by locus length.
Expected fits assuming independence of locus length and presence of a peak,
and assuming proportionality of locus length and presence of a peak are given
to demonstrate deviation from either for the dataset.
}
\examples{

# Spline plot for E2F4 example peak dataset.
data(peaks_E2F4, package = 'chipenrich.data')

# Create the plot for a different locus definition
# to compare the effect.
plot_chipenrich_spline(peaks_E2F4, locusdef = 'nearest_gene', genome = 'hg19')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{genome_to_organism}
\alias{genome_to_organism}
\title{Get the correct organism code based on genome}
\usage{
genome_to_organism(genome = supported_genomes())
}
\arguments{
\item{genome}{One of the \code{supported_genomes()}.}
}
\value{
A string for the three letter organism code. Convention is first letter
of the first word in the binomial name, and first two letters of the second word
in the binomial name. 'Homo sapiens' is then 'hsa', for example.
}
\description{
Data from \code{chipenrich.data} uses three letter organism codes for the
\code{GeneSet} objects. This function ensures the correct objects are loaded.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{setup_mappa}
\alias{setup_mappa}
\title{Function to setup mappability}
\usage{
setup_mappa(mappa_code, genome, ldef_code, ldef_obj)
}
\arguments{
\item{mappa_code}{One of \code{NULL}, a file path to a custom mappability file,
or an \code{integer} for a valid read length given by \code{supported_read_lengths}.
If a file, it should contain a header with two column named 'gene_id' and 'mappa'.
Gene IDs should be Entrez IDs, and mappability values should range from 0 and 1.
Default value is NULL.}

\item{genome}{One of the \code{supported_genomes()}.}

\item{ldef_code}{One of 'nearest_tss', 'nearest_gene', 'exon', 'intron', '1kb',
'1kb_outside', '1kb_outside_upstream', '5kb', '5kb_outside', '5kb_outside_upstream',
'10kb', '10kb_outside', '10kb_outside_upstream'. Alternately, a file path for
a custom locus definition. NOTE: Must be for a \code{supported_genome()}, and
must have columns 'chr', 'start', 'end', and 'gene_id', or 'geneid'.}

\item{ldef_obj}{A \code{LocusDefinition} object.}
}
\value{
A \code{data.frame} with columns \code{gene_id} and \code{mappa}.
}
\description{
Function to setup mappability
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign_peaks.R
\name{assign_peaks}
\alias{assign_peaks}
\title{Assign peak midpoints to defined gene loci.}
\usage{
assign_peaks(peaks, locusdef, tss, weighting = NULL)
}
\arguments{
\item{peaks}{A \code{GRanges} object representing regions to be used for enrichment.}

\item{locusdef}{A locus definition object from \code{chipenrich.data}.}

\item{tss}{A \code{GRanges} object representing the TSSs for the genome build. Includes \code{mcols} for Entrez Gene ID \code{gene_id} and gene symbol \code{symbol}.}

\item{weighting}{A string defining what weighting option they want. Current options are 'multiAssign', 'signalValue', and 'logSignal Value'. Default is NULL.}
}
\value{
A \code{data.frame} with columns for \code{peak_id, chr, peak_start, peak_end, gene_locus_start, gene_locus_end, gene_id, nearest_tss, nearest_tss_gene, dist_to_tss, nearest_tss_gene_strand}. The result is used in \code{num_peaks_per_gene()}.
}
\description{
Determine the midpoints of a set of input regions \code{peaks} and the overlap of the midpoints with a given locus definition \code{locusdef}. Also report the TSS that is nearest each region (peak) overlapping a defined locus and its distance.
}
\details{
Typically, this function will not be used alone, but inside \code{chipenrich()}.
}
\examples{

data('locusdef.hg19.nearest_tss', package = 'chipenrich.data')
data('tss.hg19', package = 'chipenrich.data')

file = system.file('extdata', 'test_assign.bed', package = 'chipenrich')
peaks = read_bed(file)

assigned_peaks = assign_peaks(
	peaks = peaks,
	locusdef = locusdef.hg19.nearest_tss,
	tss = tss.hg19)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supported.R
\name{supported_read_lengths}
\alias{supported_read_lengths}
\title{Display supported read lengths for mappability}
\usage{
supported_read_lengths()
}
\value{
A \code{data.frame} with columns \code{genome, locusdef, read_length}.
}
\description{
Display supported read lengths for mappability
}
\examples{

supported_read_lengths()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{read_geneset}
\alias{read_geneset}
\title{Function to read custom gene sets from file}
\usage{
read_geneset(file_path)
}
\arguments{
\item{file_path}{A file path for the custom gene set.}
}
\value{
A \code{GeneSet} class object.
}
\description{
This function reads a two-columned tab-delimited text file (with header). Column
names are ignored, but the first column should be geneset names or IDs and the
second column should be Entrez Gene IDs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxReg.R
\name{proxReg}
\alias{proxReg}
\title{Run Proximity Regulation test on a set of narrow genomic regions}
\usage{
proxReg(
  peaks,
  out_name = "proxReg",
  out_path = getwd(),
  genome = supported_genomes(),
  reglocation = "tss",
  genesets = c("GOBP", "GOCC", "GOMF"),
  randomization = NULL,
  qc_plots = TRUE,
  min_geneset_size = 15,
  max_geneset_size = 2000,
  n_cores = 1
)
}
\arguments{
\item{peaks}{Either a file path or a \code{data.frame} of peaks in BED-like
format. If a file path, the following formats are fully supported via their
file extensions: .bed, .broadPeak, .narrowPeak, .gff3, .gff2, .gff, and .bedGraph
or .bdg. BED3 through BED6 files are supported under the .bed extension. Files
without these extensions are supported under the conditions that the first 3
columns correspond to 'chr', 'start', and 'end' and that there is either no
header column, or it is commented out. If a \code{data.frame} A BEDX+Y style
\code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame} for
acceptable column names.}

\item{out_name}{Prefix string to use for naming output files. This should not
contain any characters that would be illegal for the system being used (Unix,
Windows, etc.) The default value is "proxReg", and a file "proxReg_results.tab"
is produced. If \code{qc_plots} is set, then a file "proxReg_qcplots.pdf"
is produced containing a number of quality control plots. If \code{out_name}
is set to NULL, no files are written, and results then must be retrieved from
the list returned by \code{proxReg}.}

\item{out_path}{Directory to which results files will be written out. Defaults
to the current working directory as returned by \code{\link{getwd}}.}

\item{genome}{One of the \code{supported_genomes()}. If reglocation = enhancer,
genome MUST be 'hg19'.}

\item{reglocation}{One of: 'tss', 'enhancer'. Details in the "Regulatory locations" section}

\item{genesets}{A character vector of geneset databases to be tested for
enrichment. See \code{supported_genesets()}. Alternately, a file path to
a tab-delimited text file with header and first column being the geneset ID
or name, and the second column being Entrez Gene IDs. For an example custom
gene set file, see the vignette.}

\item{randomization}{One of: 'shuffle', 'unif', 'bylength', 'byenh'. These were used to
test for Type I error under the null hypothesis. A general user will never have to
use these.}

\item{qc_plots}{A logical variable that enables the automatic generation of
plots for quality control.}

\item{min_geneset_size}{Sets the minimum number of genes a gene set may have
to be considered for testing.}

\item{max_geneset_size}{Sets the maximum number of genes a gene set may have
to be considered for testing.}

\item{n_cores}{The number of cores to use for testing. We recommend
using only up to the maximum number of \emph{physical} cores present, as
virtual cores do not significantly decrease runtime. Default number of cores
is set to 1. NOTE: Windows does not support multicore testing.}
}
\value{
A list, containing the following items:

\item{opts }{A data frame containing the arguments/values passed to \code{polyenrich}.}

\item{peaks }{
A data frame containing peak assignments to genes. Peaks which do not overlap
a gene locus are not included. Each peak that was assigned to a gene is listed,
along with the peak midpoint or peak interval coordinates (depending on which
was used), the gene to which the peak was assigned, the locus start and end
position of the gene, and the distance from the peak to the TSS.

The columns are:

\describe{
  \item{peak_id}{an ID given to unique combinations of chromosome, peak start, and peak end. }
  \item{chr}{the chromosome the peak originated from. }
  \item{peak_start}{start position of the peak. }
  \item{peak_end}{end position of the peak. }
  \item{gene_id}{the Entrez ID of the gene to which the peak was assigned. }
  \item{gene_symbol}{the official gene symbol for the gene_id (above). }
  \item{gene_locus_start}{the start position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
  \item{gene_locus_end}{the end position of the locus for the gene to which the peak was assigned (specified by the locus definition used.) }
  \item{nearest_tss}{the closest TSS to this peak (for any gene, not necessarily the gene this peak was assigned to.) }
  \item{dist_to_tss}{the distance in bp to the closest TSS to this peak. }
  \item{nearest_tss_gene}{the gene having the closest TSS to the peak (should be the same as gene_id when using the nearest TSS locus definition.) }
  \item{nearest_tss_gene_strand}{the strand of the gene with the closest TSS. }
  \item{log_dtss}{ log of dist_to_tss}
  \item{log_gene_ll}{the log of length of the gene locus in bp }
  \item{scaled_dtss}{the adjusted distance to TSS, used in the calculations. Shown if reglocation = "tss" }
  \item{dist_to_enh}{the distance to the nearest enhancer. Shown if reglocation = "enhancer" }
  \item{avg_denh}{the empirical average for distance to the nearest enhancer for the gene the peak is assigned to. Shown if reglocation = enhancer }
  \item{scaled_denh}{the adjusted distance to the nearest enhancer. Shown if reglocation = enhancer }
}}

\item{results }{
A data frame of the results from performing the proxReg test on
each geneset that was requested (all genesets are merged into one final data
frame.) The columns are:

\describe{
  \item{Geneset.ID}{the identifier for a given gene set from the selected database.  For example, GO:0000003. }
  \item{Geneset.Type}{ specifies from which database the Geneset.ID originates.  For example, "Gene Ontology Biological Process."}
  \item{Description}{ gives a definition of the geneset. For example, "reproduction."}
  \item{P.Value}{the probability of observing the proxmity of genomic regions in the gene set given the null hypothesis that peaks are not closer or farther in the gene set.}
  \item{FDR}{the false discovery rate proposed by Bejamini \& Hochberg for adjusting the p-value to control for family-wise error rate.}
  \item{Effect}{the signed Wilcoxon statistic, with positive values meaning the gene set has closer genomic regions than expected by chance.}
  \item{Status}{specifies if the peaks in the gene set tend to be closer or farther than those not in the gene set.}
  \item{Odds.Ratio}{the estimated odds that peaks are associated with a given gene set compared to the odds that peaks are associated with other gene sets, after controlling for locus length and/or mappability.  An odds ratio greater than 1 indicates enrichment, and less than 1 indicates depletion.}
  \item{N.Geneset.Genes}{the number of genes in the gene set.}
  \item{N.Geneset.Peak.Genes}{the number of genes in the genes set that were assigned at least one peak.}
  \item{Geneset.Peak.Genes}{the list of genes from the gene set that had at least one peak assigned.}
}}
}
\description{
This method is designed for a set of narrow genomic regions (e.g. TF peaks) and is used to test
whether the genomic regions assigned to genes in a gene set are closer to 
regulatory locations (i.e. promoters or enhancers) than by chance.
}
\section{Regulatory locations}{

Current supported regulatory locations are gene transcription
start sites (tss) or enhancer locations (hg19 only)
}

\section{Method}{

ProxReg first calculates the distance between each peak midpoint and 
regulatory location in base pairs. For gene transcription start sites, 
since parts of the chromosome are more sparse than others, there is an
association with gene locus length that needs to be adjusted for.
When using tss as the regulatory location, the peak distances are 
adjusted for this confounding variable based on an average of 90 ENCODE
ChIP-seq experiments (details in citation pending). Similarly, for enhancers, 
distances depend on the density of enhancers within a gene locus, so distance
to enhancer is adjusted using an empirical average of 90 ChIP-seq ENCODE
experiments.

For each gene set of interest, the genomic regions are divided into two groups indicating
the gene with the nearest tss is in the gene set or not. A Wilcoxon Rank-Sum test is 
then done to test for a difference in the adjusted distances (either to tss or enhancer).
}

\examples{
# Run proxReg using an example dataset, assigning peaks to the nearest TSS,
# and on a small custom geneset
data(peaks_E2F4, package = 'chipenrich.data')
peaks_E2F4 = subset(peaks_E2F4, peaks_E2F4$chrom == 'chr1')
gs_path = system.file('extdata','vignette_genesets.txt', package='chipenrich')
results = proxReg(peaks_E2F4, reglocation = 'tss',
			genome = 'hg19', genesets=gs_path, out_name=NULL)

# Get the list of peaks that were assigned to genes and their distances to 
# regulatory regions.
assigned_peaks = results$peaks

# Get the results of enrichment testing.
enrich = results$results

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chipenrich_package_doc.R
\docType{package}
\name{chipenrich_package}
\alias{chipenrich_package}
\title{chipenrich: Gene Set Enrichment For ChIP-seq Peak Data and Other Genomic Regions}
\description{
The \code{chipenrich} package includes three classes of methods that adjust for potential confounders of gene set enrichment testing (locus length and mappability of the sequence reads). The first, \code{chipenrich}, is designed for use with transcription-factor (TF) based ChIP-seq experiments and other DNA sequencing experiments with narrow genomic regions. The second, \code{polyenrich}, is similarly designed for TF based ChIP-seq, but where the number of peaks present in gene loci may be important. The third, \code{broadenrich}, is designed for use with histone modification based ChIP-seq experiments and other DNA sequencing experiments with broad genomic regions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{filter_genesets}
\alias{filter_genesets}
\title{Function to filter genesets by locus definition and size}
\usage{
filter_genesets(
  gs_obj,
  ldef_obj,
  min_geneset_size = 15,
  max_geneset_size = 2000
)
}
\arguments{
\item{gs_obj}{A valid GeneSet object}

\item{ldef_obj}{A valid LocusDefinition object}

\item{min_geneset_size}{An integer indicating the floor for genes in a geneset. Default 15.}

\item{max_geneset_size}{An integer indicating the ceiling for genes in a geneset. Default 2000.}
}
\value{
An altered \code{gs_obj} with changed \code{set.gene} and \code{all.genes}
slots reflecting \code{min_geneset_size} and \code{max_geneset_size} after intersecting
with the genes present in the particular locus definition.
}
\description{
This function filters gene sets based on the genes that are present in a particular
locus definition. After determining which genes are present in both the GeneSet,
\code{gs_obj}, and the LocusDefinition \code{ldef_obj}, gene sets are filtered
by size with \code{min_geneset_size} and \code{max_geneset_size}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{load_peaks}
\alias{load_peaks}
\title{Convert a BEDX+Y data.frame and into GRanges}
\usage{
load_peaks(dframe, keep.extra.columns = TRUE)
}
\arguments{
\item{dframe}{A BEDX+Y style \code{data.frame}. See \code{GenomicRanges::makeGRangesFromDataFrame}
for acceptable column names for appropriate conversion to \code{GRanges}.}

\item{keep.extra.columns}{Keep extra columns parameter from \code{GenomicRanges::makeGRangesFromDataFrame()}.}
}
\value{
A \code{GRanges} that may or may not \code{keep.extra.columns}, and
that may or may not be stranded, depending on whether there is strand column
in the \code{dframe}.
}
\description{
Given a \code{data.frame} in BEDX+Y format, use the built-in function
\code{GenomicRanges::makeGRangesFromDataFrame()} to convert to \code{GRanges}.
}
\details{
Typically, this function will not be used alone, but inside \code{chipenrich()}.
}
\examples{

# Example with just chr, start, end
peaks_df = data.frame(
chr = c('chr1','chr2','chr3'),
start = c(35,74,235),
end = c(46,83,421),
stringsAsFactors = FALSE)
peaks = load_peaks(peaks_df)

# Example with extra columns
peaks_df = data.frame(
chr = c('chr1','chr2','chr3'),
start = c(35,74,235),
end = c(46,83,421),
strand = c('+','-','+'),
score = c(36, 747, 13),
stringsAsFactors = FALSE)
peaks = load_peaks(peaks_df, keep.extra.columns = TRUE)

}
